/**
 * @file
 * @brief %Classifier interface
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-03-02 14:02:00 +0100 (Die, 02 Mär 2004) $
 * $Revision: 677 $
 *
 * @since    Wed Mar 19 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: Classifier.h 677 2004-03-02 13:02:00Z rene-cvs $
 */

#ifndef _CLASSIFIER_H
#define _CLASSIFIER_H

#ifdef _CLASSIFIER_EXPORT
# define _FEATURE_EXPORT
#endif
#ifdef _CLASSIFIER_EXPORT_CLASS
# define _FEATURE_EXPORT_CLASS
#endif

#include "../FeatureContainer/FeatureContainer.h"

//@{
/** compiler specific flags to export functions and classes */
#define CLASSIFIER_EXPORT FEATURE_EXPORT
#define CLASSIFIER_EXPORT_CLASS FEATURE_EXPORT_CLASS
//@}

/**
 * Export a classifier
 *
 * This macro has to be used in global scope and gets substituted with the
 * implementation for exporting a ClassifierAlgorithm singleton instance to the
 * framework.
 *
 * @param classifier The Class that implements the ClassifierAlgorithm interface
 *                 and should be exported to the framework.
 */
#define EXPORT_CLASSIFIER(classifier) \
	/** Singleton implementing the ClassifierAlgorithm interface */ \
	static classifier *clf; \
	void INIT_EXPORT library_initialize() { clf = NULL; } \
	void FINI_EXPORT library_finalize() { if (clf != NULL) delete clf; } \
	/** \
     * Returns an instance of a class implementing the ClassifierAlgorithm interface \
	 * \
	 * @param params A map of parameters to initialize the instance \
	 * @return A ClassifierAlgorithm implementation \
	 */ \
	extern "C" CLASSIFIER_EXPORT ClassifierAlgorithm* getClassifier(classifierparams &params) { if (clf == NULL) clf = new classifier(params); return clf; }

/**
 * Parameters passed to classifiers
 *
 * When initializing a classifier, an arbitrary number of named (string)
 * parameters can be passed from the application (e.g. read from a config file).
 * These key-value tuples can then be used by the classifier as
 * configuration or initialization values. They are completely classifier
 * specific.
 *
 * @see Classifier
 * @see ClassifierAlgorithm
 */
typedef parametermap classifierparams;

/**
 * ClassifierAlgorithm Interface
 *
 * This class defines the abstract interface to classification/clustering
 * algorithms. Any algorithm which is to be used for the classification step
 * in this framework has to implement these methods and be derived from this
 * class.
 */
class CLASSIFIER_EXPORT_CLASS ClassifierAlgorithm {
	public:
		/** Destructor */
		virtual ~ClassifierAlgorithm() {}

		/**
		 * Initializer
		 *
		 * Initializes internal data structures of the
		 * classification/clustering algorithm. This usually involves
		 * constructing initial prototypes for cluster centers; thus,
		 * randomized feature vectors are necessary for the
		 * initialization. The getFeatureVector method of the passed
		 * FeatureContainer object may be called multiple times by the
		 * initialization.
		 * The given FeatureContainer is also used when nextSample()
		 * is called.
		 *
		 * @param fc The FeatureContainer from which the randomized
		 *           feature vectors should be retrieved.
		 * @see FeatureContainer::getFeatureVector
		 * @see nextSample()
		 */
		virtual void init(FeatureContainer *fc) = 0;
		/**
		 * Get membership vector
		 *
		 * This method returns the data for the next steps, which is a
		 * list of context classes found by the algorithm with
		 * associated membership values. These values specify the
		 * membership of the current feature values to each of the
		 * context classes in the interval [0; 1]. They can thus by seen
		 * as the probabilities that the context classes are currently
		 * active.
		 *
		 * @param sample The sample vector for which the membershiplist
		 *               should be calculated
		 */
		virtual membershiplist getClusterMembership(const featurevector *sample) = 0;
		/**
		 * Fetch next sample vector
		 *
		 * Gets the next sample vector from the configured feature
		 * source. This method calls nextSample and getSampleVector on
		 * the FeatureContainer given to the init method.
		 *
		 * @see FeatureConteiner::nextSample
		 * @see FeatureContainer::getSampleVector
		 */
		virtual unsigned long nextSample() = 0;
		/**
		 * Fetch next sample vector
		 * Gets the next sample vector passed as parameter.
		 *
		 * @param sample samplevector
		 */
		virtual unsigned long nextSample(const featurevector* sample) = 0;
		
		/**
		 * Serialize a samples data to a string
		 *
		 * @return String representation of the samples data.
		 */
		virtual string serialize() const = 0;

		/**
		 * Unserialize a samples data from a string
		 *
		 * @param data String representation of the samples data. 
		 */
		virtual void unserialize(string data) = 0;

#if _DEBUG_CLASSIFIERS
		/** This is only for testing. */
		virtual string toString() const = 0;
#endif
};

/**
 * Classifier Wrapper
 *
 * The Classifier Class is a wrapper class that loads
 * a dll containing the actual implementation during
 * runtime.
 */
class Classifier : public ClassifierAlgorithm {
	private:
		/** Instance */
		FeatureContainer *f;
		/** Implementation */
		ClassifierAlgorithm *ca;

#if defined(_WIN32_WCE) || defined (_WINDOWS)
		HMODULE dll;
#else
		/** Library handle */
		void* dlHandle;
#endif

	public:
		/**
		 * Creates the classifier
		 *
		 * @param name The name of the requested classifier implementation.
		 * @param params A map of parameters to initialize the classifier.
		 */
		Classifier(const string &name, const classifierparams &params);
		/** Destructor */
		virtual ~Classifier();

		virtual void init(FeatureContainer *fc) { f = fc; ca->init(fc); }
		virtual membershiplist getClusterMembership(const featurevector *sample) { return ca->getClusterMembership(sample); }
		virtual unsigned long nextSample() { return ca->nextSample(); }
		virtual unsigned long nextSample(const featurevector* sample) { return ca->nextSample(sample); }
		virtual string serialize() const { return ca->serialize(); }
		virtual void unserialize(string data) { ca->unserialize(data); }

		/** Return actual ClassifierAlgorithm instance */
		ClassifierAlgorithm* getCa() const { return ca; }

#if _DEBUG_CLASSIFIERS
		/** This is only for testing. */
		virtual string toString() const { return ca->toString(); }
#endif
};

/**
 * Type of the getClassifier function
 *
 * This type definition is needed by the Classifier Wrapper to query a
 * ClassifierAlgorithm implementation for an instance of its class.
 */
typedef ClassifierAlgorithm* (*getClassifier_t) (const classifierparams &params);

#endif
