/**
 * @file
 * @brief Growing neural gas
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-03-19 14:19:26 +0100 (Fre, 19 Mär 2004) $
 * $Revision: 710 $
 *
 * @since    Mon Mar 3 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: GNG.cpp 710 2004-03-19 13:19:26Z rene $
 */

#include "GNG.h"
// for strsplit
#include "../../util/featurevector.h"

EXPORT_CLASSIFIER(GNG)

// insertion criteria
unsigned GNG::lambda = 10;            // insert a new unit after (lambda * <number of units>) samples
// deletion criterias
double GNG::minimal_deletion_age = 0.5;   // only delete units which are older than that (their youth must be lower than this value)
double GNG::deletion_threshold = 0.5;	  // the threshold for the deletion criterion (it must be lower than this for a node to get deleted)
double GNG::deletion_threshold_lq = 0.15;  // the threshold for the quality measure for learning when deleting a node (it must be lower than this for a node to get deleted)

// only used in unit.cpp
unsigned long GNG::edge_age_max = 30;     // maximum age of an edge
double GNG::T_short = 6;                 // time constant for exponential decrease of the short term error of a winning unit
double GNG::T_long = 170;                 // time constant for exponential decrease of the long term error of a winning unit
double GNG::T_youth = 50;                // time constant for exponential decrease of the youth
double GNG::T_insert_threshold = 55;
double GNG::adaptation_threshold = 0.0005;
double GNG::insertion_tolerance = 0.02;
double GNG::lr_winner = 0.15;              // the learning rate of the winning unit
double GNG::lr_neighbor = 0.01;           // the learning rate of all neighbors
double GNG::minimum_lr = 0.2;             // the absolute minimum learning rate
double GNG::lr_insert_threshold = 0.2;    // the learning rate for the insertion threshold

// helper variables for better performance
double GNG::helper_T_short = -1;
double GNG::helper_T_long = -1;
double GNG::helper_T_youth = -1;
double GNG::helper_T_insertion_threshold = -1;

GNG::GNG(classifierparams &params): params(params)
{
	algo = params["persist"].c_str();
}

void GNG::init(FeatureContainer *fc)
{
//	plog("GNG::init setting helper variables \n");
	// check some constant
	assert(lr_winner > 0 && lr_winner <= 1);
	assert(lr_neighbor >= 0 && lr_neighbor <= 1);
	assert(minimum_lr >= 0 && minimum_lr <= 1);
	assert(lr_insert_threshold > 0 && lr_insert_threshold <= 1);

	if (helper_T_short == -1)
		helper_T_short = exp(-1 / T_short);
	if (helper_T_long == -1)
		helper_T_long = exp(-1 / T_long);
	if (helper_T_youth == -1)
		helper_T_youth = exp(-1 / T_youth);
	if (helper_T_insertion_threshold == -1)
		helper_T_insertion_threshold = exp(-1 / T_insert_threshold);

	assert(helper_T_short > 0);
	assert(helper_T_long > 0);
	assert(helper_T_youth > 0);
	assert(helper_T_insertion_threshold > 0);

	Unit* u0, *u1;
	container = fc;

	units.clear();
	feature_mask.clear();

//	plog("GNG::init setting generating feature mask\n");
	featurevector f1 = fc->getFeatureVector(), f2 = fc->getFeatureVector();
	feature_mask = bit_vector(f1.size());
	for (unsigned int i = 0; i < f1.size(); i++)
		if (params.count((*f1[i]).getName()) == 0 ||
		    (*params.find((*f1[i]).getName())).second == "1") { // ugly, but [] on the map won't work if the map is const
			feature_mask[i] = true;
		}
		else {
			feature_mask[i] = false;
//			plog("    Not using feature %s\n", (*f1[i]).getName().c_str());
		}

//	plog("GNG::init creating random sample from feature\n");
	// create the first two units with the random feature vectors
	// the initialization of the local error counters is done in the constructor
	u0 = new Unit(&units, maskFeatureVector(&f1, true), 1);
	u1 = new Unit(&units, maskFeatureVector(&f2, true), 1);
	Unit::reserveMetaId(1);

	units.push_back(u0);
	units.push_back(u1);
	u0->addEdge(u1);

	// unserialize feature data, if available
	FILE *gngfile = fopen(algo, "r");
	if (gngfile != NULL) {
		char buffer[4096];
		string gngdata;

		plog("Found GNG data file, unserializing GNG structure\n");

		while(fgets(buffer, 4096, gngfile) != NULL) {
			gngdata += buffer;
		}

		fclose(gngfile);

		unserialize(gngdata);
	}
}

GNG::~GNG()
{
	writeToDisk();

	for (unit_list::iterator iter = units.begin(); iter != units.end(); iter++) {
		delete (*iter);
	}
	units.clear();
}

void GNG::writeToDisk()
{
	plog("Saving GNG data file\n");
	
	FILE *gngfile = fopen(algo, "w");
	if (gngfile != NULL) {
		fputs(serialize().c_str(), gngfile);
		fclose(gngfile);
	}
}

unsigned long GNG::nextSample() {
	container->nextSample();
	const featurevector* sample = container->getSampleVector();
	unsigned long ret = nextSample(sample);
	
	writeToDisk();

	return ret;
}

/// @todo cope with increasing dimension of the feature vector ?
unsigned long GNG::nextSample(const featurevector* inSample)
{
	static unsigned int num_sample = 0;

	// find the nearest (best matching) and second nearest unit
	Unit *b = NULL, *s = NULL,
		*i = NULL,	// the unit with the highest value of the insertion criterion
		*d = NULL;	// the unit with the low
	double bDist = 1.1, sDist = 1.1, iCrit = 0, dCrit = 1, averageSimilarity = 0;
	featurevector maskSample = maskFeatureVector(inSample, false);
	
	for (unit_list::iterator iter = units.begin(); iter != units.end(); iter++) {
		assert(*iter != NULL);
		// the units matching the new sample best
		double curDist = (*iter)->getDistance(&maskSample);
		if (curDist < sDist) {
			if (curDist < bDist) {
				sDist = bDist;
				s = b;
				bDist = curDist;
				b = *iter;
			}
			else {
				sDist = curDist;
				s = *iter;
			}
		}

		// find the highest insertion criterion (for inserting a new one)
		double curCritIns = (*iter)->getQualityInsertion() - (*iter)->getYouth();
		if (i == NULL || curCritIns > iCrit) {
			i = *iter;
			iCrit = curCritIns;
		}

		// also calculate the average similarity of the input "weights"
		averageSimilarity += (*iter)->getLocalSimilarity();
	}
	averageSimilarity /= units.size();
	assert(b != NULL && s != NULL);
	assert(b != s);
	assert(i != NULL);

	// first update the error of the winning unit
	b->updateErrors(&maskSample);

	// update the weights of the winning unit (and all its neighbors)
	b->moveTowards(&maskSample);

	// update the edges to its neighbors
	b->ageEdges();

	// connect the two "winners" with an edge with age 0
	b->addEdge(s);

	// now remove nodes that have lost all edges
	/// @todo this should be done in a better way
	unit_list u;
	for (unit_list::iterator del_iter = units.begin(); del_iter != units.end(); del_iter++) {
		if ((*del_iter)->getEdgeCount() == 0) {
			if (*del_iter == i) {
				i = NULL;
			}

			plog("_");

			// here we also have to release the meta ID, which will have been assigned to this node after the meta split
			if ((*del_iter)->getMetaId() > 0)
				Unit::releaseMetaId((*del_iter)->getMetaId());

			delete *del_iter;
			u.push_back(*del_iter);
        }
	}
	for (unit_list::iterator foo = u.begin(); foo != u.end(); foo++) {
		units.remove(*foo);
	}

	// after enough steps, insert a new node
	if (num_sample >= lambda * units.size()) {
		num_sample = 0;

		if (i != NULL && iCrit > 0) {
			Unit* newUnit = i->splitUnit();
			if (newUnit != NULL) {
				units.push_back(newUnit);

				plog("+");
			} else {
				plog("x");
			}
		} else {
			plog("X");
		}


		for (unit_list::iterator iter = units.begin(); iter != units.end(); iter++) {
			// the unit with the lowest value of the deletion criterion (if there is one that fulfills it)
			double curCritDel = (*iter)->getLocalSimilarity() / averageSimilarity;
			if (d == NULL ||
			    (curCritDel < dCrit && (*iter)->getEdgeCount() >= 2 &&
			     (*iter)->getYouth() < minimal_deletion_age &&
			     (*iter)->getQualityLearning() < deletion_threshold_lq)) {
				d = *iter;
				dCrit = curCritDel;
			}
		}
		assert(d != NULL);
		// and delete the node (if one has been found)
		if (d != NULL && d != b && units.size() > 2 && dCrit < deletion_threshold) {
			plog("-");

			// the edges to this node are removed in the destructor
			// but if there are no edges here, then also release the meta ID
			if (d->getEdgeCount() == 0)
				Unit::releaseMetaId(d->getMetaId());
			delete d;
#ifdef NO_OPTIMIZATION_SPLAYTREE
			units.erase(find(units.begin(), units.end(), d));
#else
			units.remove(d);
#endif
		} else {
			if (d == b)
				plog("?1");
			if (units.size() <= 2)
				plog("?2");
			if (dCrit >= deletion_threshold)
				plog("?3");
		}
	}
	num_sample++;

	// and update the internal counters
	//b->decreaseErrors();
	b->decreaseYouth();
	b->decreaseInsertionThreshold();

	units.sort();

	return b->getMetaId();
}

double GNG::getWinnerDistance(const featurevector *inSample, unsigned int &cluster_id, unsigned int &meta_cluster_id) const
{
	// find the nearest (best matching) and second nearest unit
	Unit *b = NULL;
	double bDist = 1;
	unsigned int i=0;
	featurevector maskSample = maskFeatureVector(inSample, false);

	for (unit_list::const_iterator iter = units.begin(); iter != units.end(); iter++) {
		assert(*iter != NULL);
		double curDist = (*iter)->getDistance(&maskSample);
		if (b == NULL || curDist < bDist) {
			b = *iter;
			bDist = curDist;
			cluster_id = i;
		}
		i++;
	}
	meta_cluster_id = b->getMetaId();
	return bDist;
}

unit_list::size_type GNG::getNumberNodes() const
{
    return units.size();
}

meta_cluster_list GNG::getMetaClusters()
{
	// fill an empty structure for easy access
	vector<meta_cluster> tmp(Unit::getMaxMetaId()+1);
	for (vector<meta_cluster>::iterator iter = tmp.begin(); iter != tmp.end(); iter++)
		(*iter).cluster_id = 0;

	// just traverse over all clusters and add them to the respective meta cluster
	for (unit_list::const_iterator iter2 = units.begin(); iter2 != units.end(); iter2++) {
		assert((*iter2)->getMetaId() < tmp.size());
		tmp[(*iter2)->getMetaId()].cluster_id = (*iter2)->getMetaId();
		tmp[(*iter2)->getMetaId()].units.push_back(*iter2);
		tmp[(*iter2)->getMetaId()].winner_hits = 0;
	}

	meta_cluster_list ret;
	for (vector<meta_cluster>::iterator iter3 = tmp.begin(); iter3 != tmp.end(); iter3++)
		if ((*iter3).cluster_id != 0)
			ret.push_back(*iter3);

	/// @todo THIS MUST WORK !!!!
	//assert(ret.size() == Unit::getNumMetaIds());

	return ret;
}

string GNG::serialize() const
{
	string ret;
	vector<Unit*> unit_vec;
	map<Unit*, unsigned int> unit_ind;
	unit_list::const_iterator iter = units.begin();
	unsigned int i = 0;

	while(iter != units.end()) {
		unit_vec.push_back(*iter);
		// remember their IDs for fast matching of the pointers
		unit_ind[*iter] = i;
		iter++; i++;
	}

	// now serialize the units with their ID
	for (i=0; i<unit_vec.size(); i++) {
		// their data
		ret += unit_vec[i]->serialize();
		// their edges
		for (edge_map::const_iterator iter2 = unit_vec[i]->getEdges()->begin();
				iter2 != unit_vec[i]->getEdges()->end(); iter2++) {
			char buffer [sizeof(unsigned)*8+sizeof(unsigned long)*8+2];
			sprintf(buffer, "%d,%ld,", unit_ind[(*iter2).first], (*iter2).second);
			ret += buffer;
		}
		ret += ";\n";
	}

	return ret;
}

void GNG::unserialize(string data)
{
	vector<Unit*> unit_vec;
	featurevector inFeatures = container->getFeatureVector();
	featurevector maskFeatures = maskFeatureVector(&inFeatures, false);

	// don't forget to clean up the meta id list - it will be recreated by Unit::unserialize
	Unit::clearMetaIds();

	size_t pos = 0, pos2, num_unit = 0;
	while ((pos2 = data.find('\n', pos)) < data.length()) {
		string line = data.substr(pos, pos2-pos);
		pos = pos2+1;

		if (unit_vec.size() <= num_unit) {
			unit_vec.push_back(new Unit(&units, featurevector(), 0));
		}

		// the unit data
		size_t edges_pos = unit_vec[num_unit]->unserialize(maskFeatures, line);
//		size_t edges_pos = unit_vec[num_unit]->unserialize(inFeatures, line);

		// the edges
		char* buf = strdup(line.substr(edges_pos).c_str());
		char* tok = strsplit(buf, ',');
		do {
			unsigned int ind = atoi(tok);
			while (unit_vec.size() <= ind)
				unit_vec.push_back(new Unit(&units, featurevector(), 0));
			unit_vec[num_unit]->addEdge(unit_vec[ind]);
			tok = strsplit(NULL, ',');
			unit_vec[num_unit]->setEdgeAge(unit_vec[ind], atol(tok));
		} while ((tok = strsplit(NULL, ',')) != NULL);
		free(buf);
		num_unit++;
	}

	// free the featurevector, that was used to reconstruct the units
	for (featurevector::iterator iter = inFeatures.begin(); iter != inFeatures.end(); iter++) {
		delete *iter;
	}

	// OK, the units are ready, take them instead of the old ones
	for (unit_list::iterator iter2 = units.begin(); iter2 != units.end(); iter2++) {
		delete *iter2;
	}
	units.clear();

	for (vector<Unit*>::const_iterator iter3 = unit_vec.begin(); iter3 != unit_vec.end(); iter3++) {
		units.push_back(*iter3);
	}
}

membershiplist GNG::getClusterMembership(const featurevector *inSample)
{
	membershiplist list;
	featurevector maskSample = maskFeatureVector(inSample, false);

	for (unit_list::iterator iter = units.begin(); iter != units.end(); iter++) {
		double membership = 1.0 - (*iter)->getDistance(&maskSample);

		if (list.count((*iter)->getMetaId()) == 0)
			list[(*iter)->getMetaId()] = 0,0;

		if (list[(*iter)->getMetaId()] < membership) {
			list[(*iter)->getMetaId()] = membership;
		}
	}

	return list;
}

/*cluster_table *GNG::getCluster()
{
	cluster_table *table = new cluster_table();
	cluster_info info;

	info.param.push_back(0.0);
	info.param.push_back(1.0);

	info.pos.push_back(0.0);
	info.pos.push_back(1.0);
	info.pos.push_back(2.0);

	table->push_back(info);
	table->push_back(info);
	table->push_back(info);

	return table;
}*/

/*connection_table *GNG::getConnections()
{
	connection_table *table = new connection_table();
	cluster_connection conn;

	conn.a = 1;
	conn.b = 2;

	table->push_back(conn);

	conn.b = 3;

	table->push_back(conn);

	conn.a = 2;

	table->push_back(conn);

	return table;
}*/

featurevector GNG::maskFeatureVector(const featurevector* sample, bool freeUnused) const
{
	featurevector ret;

	for (unsigned int i=0; i<sample->size(); i++) {
		if (feature_mask[i]) {
			ret.push_back((*sample)[i]);
		} else {
			if (freeUnused && ((*sample)[i] != NULL)) {
				delete (*sample)[i];
			}
		}
	}

	return ret;
}

string GNG::toString() const 
{
	return "";
}
