/**
 * @file
 * @brief Growing neural gas
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-03-02 13:37:23 +0100 (Die, 02 Mär 2004) $
 * $Revision: 671 $
 *
 * @since    Mon Mar 3 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: GNG.h 671 2004-03-02 12:37:23Z rene-cvs $
 */

#ifndef _GNG_H
#define _GNG_H

#include "../Classifier.h"
#include "Unit.h"

/**
 * @todo documentation
 */
struct meta_cluster {
	//@{
	/** @todo documentation */
	unsigned int cluster_id;
	list<Unit*> units;
	unsigned int winner_hits;
	//@}
};

/**
 * @todo documentation
 */
typedef list<meta_cluster> meta_cluster_list;

/**
 * This is lifelong GNG. It is basically an implementation of Fred H. Hamker:
 * "Life-long learning Cell Structures---continuously learning without catastrophic interference"
 * Neural Networks 12 (2001) 551--573
 *
 * @todo documentation
 */
class CLASSIFIER_EXPORT_CLASS GNG : public ClassifierAlgorithm
{
	public:
		//@{
		/**
		* GNG parameter
		* @todo documentation
		*/
		// these are needed in the GNG class
		static unsigned lambda;          ///< insert a new unit after (lambda * \<number of units\>) samples
		static double minimal_deletion_age;  ///< only delete units which are older than that (their youth must be lower than this value)
		static double deletion_threshold;    ///< the threshold for the deletion criterion (it must be lower than this for a node to get deleted)
		static double deletion_threshold_lq; ///< the threshold for the quality measure for learning when deleting a node (it must be lower than this for a node to get deleted)

		// these are only needed in the Unit class, but defined here for consistency
		static unsigned long edge_age_max;   ///< maximum age of an edge
		static double T_short;               ///< time constant for exponential decrease of the short term error of a winning unit
		static double T_long;                ///< time constant for exponential decrease of the long term error of a winning unit
		static double T_youth;               ///< time constant for exponential decrease of the youth
		static double T_insert_threshold;
		static double adaptation_threshold;
		static double insertion_tolerance;
		static double lr_winner;             ///< the learning rate of the winning unit
		static double lr_neighbor;           ///< the learning rate of all neighbors
		static double minimum_lr;            ///< the absolute minimum learning rate
		static double lr_insert_threshold;   ///< the learning rate for the insertion threshold

		// these helper functions are used for better performance
		static double helper_T_short;              // helper variable for decreasing short term error, defined es helper_T = e^(-1/T)
		static double helper_T_long;               // helper variable for decreasing long term error, defined es helper_T = e^(-1/T)
		static double helper_T_youth;              // helper variable for decreasing youth, defined es helper_T = e^(-1/T)
		static double helper_T_insertion_threshold;
		//@}


		/*static double e_b(double dist);*/			/* returns the distance by which the current cluster is moved
								* into the direction of the input sample based on the euclidean
								* distance to it
								*/
		//static double e_n(double dist);			// same as above for direct neighbours

		/**
		 * Only stores the parameters for later use in init().
		 *
		 * @param params Initialization parameters
		 * @see init
		 */
		GNG(classifierparams &params);

		/** Destructor */
		virtual ~GNG();

		/**
		 * This method initializes the helper variables, creates two initial
		 * cluster center prototypes by calling getFeatureVector on the given
		 * FeatureContainer and stored the FeatureContainer reference for calls
		 * to nextSample(). Additionally, it initializes the bit vector for
		 * masking feature dimensions by using the given algorithm parameters
		 * and the feature dimensions from the retrieved feature vectors.
		 * @see nextSample()
		 */
		virtual void init(FeatureContainer *fc);
		virtual membershiplist getClusterMembership(const featurevector *sample);
		virtual unsigned long nextSample();
		virtual unsigned long nextSample(const featurevector* sample);
		virtual string serialize() const;
		virtual void unserialize(string data);
		//virtual cluster_table* getCluster();
		//virtual connection_table* getConnections();

		// these are for testing and for quantitative evaluations
		/** @todo documentation */
		virtual double getWinnerDistance(const featurevector *sample, unsigned int &cluster_id, unsigned int &meta_cluster_id) const;
		/** @todo documentation */
		virtual unit_list::size_type getNumberNodes() const;

		virtual meta_cluster_list getMetaClusters(); // the connected meta clusters in the graph

#if _DEBUG_CLASSIFIERS
		/** This is only for testing. */
		virtual string toString() const;
#endif

	private:
		const char *algo;

		/**
		 * The featurecontainer from which the parameter-less nextSample()
		 * implementation will fetch samples whenever called.
		 *
		 * @see nextSample()
		 */
		FeatureContainer *container;

		/**
		 * The list of units in this instance of GNG. This is the authorative
		 * list of all units used by GNG, other lists (e.g. the all_units
		 * pointer in Unit or the edge map in Unit) may only contain pointers
		 * to units stored in this list.
		 *
		 * @see Unit::edge
		 * @see Unit::all_units
		 */
		unit_list units;

		/**
		 * This bitvector has the same length as the feature vectors and
		 * defines for each dimension if it is used for classification. By
		 * setting it to 0, those dimensions are effectively ignored by GNG
		 * and are not even passed to the methods of the Unit class. This
		 * vector is initialized by the init method (when the feature
		 * dimensions are known).
		 * The feature mask is honored by all methods which accepts a feature
		 * vector as input parameter namely nextSample, getClusterMembership
		 * and the helper methods getWinnerDistance and getMetaClusters.
		 */
		bit_vector feature_mask;

		/**
		 * This is just a copy of the algorithm parameters passed to the
		 * constructor and used by init for constructing the feature mask.
		 *
		 * @see GNG
		 * @see init
		 * @see feature_mask
		 */
		const classifierparams params;

		/**
		 * This is a helper variable to generate a feature vector that only
		 * contains those dimensions which are set to 1 in the feature mask.
		 * The pointers to used feature dimensions are copied to the returned
		 * feature vector, the unused feature dimensions (those which are set
		 * to 0 in the feature mask) are freed if the second parameter is true.
		 *
		 * @param sample The input feature vector.
		 * @param freeUnused If true, then unused feature dimensions will be
		 *                   freed. Use with care.
		 * @return The reduced feature vector conforming to the mask.
		 * @see feature_mask
		 */
		featurevector maskFeatureVector(const featurevector* sample, bool freeUnused = false) const;

		void writeToDisk();
};

#endif
