/**
 * @file
 * @brief Growing neural gas
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-03-12 11:57:21 +0100 (Fre, 12 Mär 2004) $
 * $Revision: 694 $
 *
 * @since    Mon Mar 3 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: Unit.cpp 694 2004-03-12 10:57:21Z rene $
 */

#include "GNG.h"
#include "../../util/featurevector.h"

list<unsigned long> Unit::used_meta_ids;

Unit::Unit(unit_list const*const units, const featurevector pos) : all_units(units) {
	plog("Warning: creating Unit without giving the meta id!\n");
	Unit(units, pos, getNextMetaId());
}

Unit::Unit(unit_list const*const units, const featurevector pos, unsigned long id) : all_units(units) {
	position = pos;
	meta_id = id;

	// taken directly from the article
	youth = 1;
	error_long = error_short = insertion_threshold = 0;
	error_insert = 1;
	helper_local_similarity = -1;

	traversing = false;

	updateRadius();
}

Unit::~Unit() {
	// free all the features in the position
	for (featurevector::iterator iter = position.begin(); iter != position.end(); iter++) {
		if(*iter != NULL) {
			delete *iter;
		}
	}

	// and remove all edges to this node
	for (edge_map::iterator iter2 = edge.begin(); iter2 != edge.end(); iter2++) {
		assert((*iter2).first != NULL);
		(*iter2).first->eraseEdge(this);
	}
}

/*double Unit::e_b(double dist) {
	return 0.2;
}

double Unit::e_n(double dist) {
	return 0.1;
}*/

// beware: this method can return NULL if no suitable neighor could be found for splitting
Unit *Unit::splitUnit() {
	Unit *b = NULL;	// the neighbor with the highest "quality measure for insertion"
	Unit *r = NULL;	// the new unit
	double quality = 0.0;
	featurevector r_pos;

	// find the neighbor to use
	for (edge_map::iterator iter = edge.begin(); iter != edge.end(); iter++) {
		double cur_qual = (*iter).first->getQualityInsertion();
		/**
		 * @todo According to the paper, the quality should be _larger_ then zero.
		 *       Check why getQualityInsertion can return a zero quality for the
		 *       neighbors of the winner (!!!) and revert this check to cur_qual
		 *       > quality !
		 *       UPDATE: getQualityInsertion can of course return zero for the first
		 *       two units, when error_short and error_long are both zero. Hmm,
		 *       check the paper again.
		 */
		if (cur_qual >= quality) {
			b = (*iter).first;
			quality = cur_qual;
		}
	}
	if (b == NULL || quality < 0) {
		// no suitable neighbor could be found
		return NULL;
	}

	// removeEdge splits the metacluster, we don't want this here
	eraseEdge(b);
	b->eraseEdge(this);

	// determine the new unit's position as exactly between the two units
	featurevector::const_iterator a_iter = position.begin();
	featurevector::const_iterator b_iter = b->getPosition()->begin();
	while ((a_iter != position.end()) && (b_iter != b->getPosition()->end())) {
		assert((*a_iter) != NULL);
		assert((*b_iter) != NULL);
		Feature* f = (*a_iter)->clone();

		/* this should set the new sample exactly between the two old ones */
		f->moveTowards(*b_iter, 0.5);

		r_pos.push_back(f);
		a_iter++; b_iter++;
	}

	assert(meta_id != 0);
	r = new Unit(all_units, r_pos, meta_id);
	assert(r != NULL);

	addEdge(r);
	b->addEdge(r);
	assert(r->getEdgeCount() == 2);

	// set the threshold and error counters to the arithmetical average
	r->setShortTermError((this->getShortTermError() + b->getShortTermError()) / 2);
	r->setLongTermError((this->getLongTermError() + b->getLongTermError()) / 2);
	r->setInsertionThreshold((this->getInsertionThreshold() + b->getInsertionThreshold()) / 2);
	/// @todo is this OK ?????
	r->setYouth(1);

	this->updateInsertionThreshold();
	b->updateInsertionThreshold();
	r->updateInsertionThreshold();

	// and finally set the new inherited insertion errors
	this->setInsertionError(this->getLongTermError());
	b->setInsertionError(b->getLongTermError());
	r->setInsertionError(r->getLongTermError());

	r->updateRadius();

	return r;
}

double Unit::getShortTermError() const {
	assert(error_short >= 0 && error_short <= 1);

	return this->error_short;
}

void Unit::setShortTermError(double error) {
	this->error_short = error;

	assert(error_short >= 0 && error_short <= 1);
}

double Unit::getLongTermError() const {
	assert(error_long >= 0 && error_long <= 1);

	return this->error_long;
}

unsigned long Unit::getMetaId() const
{
	return meta_id;
}

void Unit::setLongTermError(double error) {
	this->error_long = error;

	assert(error_long >= 0 && error_long <= 1);
}

/*void Unit::decreaseErrors() {
	assert(GNG::helper_T_short > 0 && GNG::helper_T_short <= 1);
	assert(GNG::helper_T_long > 0 && GNG::helper_T_long <= 1);

	error_short *= GNG::helper_T_short;
	error_long *= GNG::helper_T_long;

	assert(error_short >= 0 && error_short <= 1);
	assert(error_long >= 0 && error_long <= 1);
}*/

void Unit::updateErrors(const featurevector *sample) {
	assert(GNG::helper_T_short > 0 && GNG::helper_T_short <= 1);
	assert(GNG::helper_T_long > 0 && GNG::helper_T_long <= 1);
	assert(error_short >= 0 && error_short <= 1);
	assert(error_long >= 0 && error_long <= 1);

	double dist = getDistance(sample);
	assert(dist >= 0 && dist <= 1);

	error_short = error_short * GNG::helper_T_short + (1 - GNG::helper_T_short) * dist;
	error_long = error_long * GNG::helper_T_long + (1 - GNG::helper_T_long) * dist;

	assert(error_short >= 0 && error_short <= 1);
	assert(error_long >= 0 && error_long <= 1);
}

double Unit::getInsertionError() const {
	assert(error_insert >= 0 && error_insert <= 1);

	return this->error_insert;
}

void Unit::setInsertionError(double error) {
	this->error_insert = error;

	assert(error_insert >= 0 && error_insert <= 1);
}

double Unit::getInsertionThreshold() const {
	assert(insertion_threshold >= 0 && insertion_threshold <= 1);

	return this->insertion_threshold;
}

void Unit::setInsertionThreshold(double threshold) {
	this->insertion_threshold = threshold;

	assert(insertion_threshold >= 0 && insertion_threshold <= 1);
}

void Unit::updateInsertionThreshold() {
	assert(error_long >= 0 && error_long <= 1);
	assert(error_insert >= 0 && error_insert <= 1);
	assert(insertion_threshold >= 0 && insertion_threshold <= 1);
	assert(GNG::insertion_tolerance >= 0 && GNG::insertion_tolerance <= 1);
	assert(GNG::lr_insert_threshold >= 0 && GNG::lr_insert_threshold <= 1);

	if (error_long > error_insert * (1 - GNG::insertion_tolerance))
		// the last insertion was not successful (the error at the time
		// of the last insertion was lower than the current long term
		// error) --> adapt the insertion threshold
		insertion_threshold += GNG::lr_insert_threshold * (error_long - insertion_threshold * (1 - GNG::insertion_tolerance));

	assert(insertion_threshold >= 0 && insertion_threshold <= 1);
}

double Unit::getYouth() const {
	assert(youth >= 0 && youth <= 1);

	return this->youth;
}

void Unit::setYouth(double youth) {
	this->youth = youth;

	assert(youth >= 0 && youth <= 1);
}

void Unit::decreaseYouth() {
	assert(GNG::helper_T_youth > 0 && GNG::helper_T_youth <= 1);
	assert(youth >= 0 && youth <= 1);

	youth *= GNG::helper_T_youth;

	assert(youth >= 0 && youth <= 1);
}

void Unit::decreaseInsertionThreshold() {
	assert(insertion_threshold >= 0 && insertion_threshold <= 1);
	assert(GNG::helper_T_insertion_threshold > 0 && GNG::helper_T_insertion_threshold <= 1);
	assert(GNG::adaptation_threshold > 0 && GNG::adaptation_threshold <= 1);

	double qual = getQualityLearning();
	assert(qual >= 0 && qual <= 2);

	double a = (1 + fabs(qual - 1)) / (1 + GNG::adaptation_threshold) -1;
	if (a < 0)
		a = 0;
	else if (a > 1)
		a = 1;

	insertion_threshold *= GNG::helper_T_insertion_threshold * (1 - a);

	assert(insertion_threshold >= 0 && insertion_threshold <= 1);
}

const featurevector* Unit::getPosition() const {
	return &position;
}

bool Unit::operator< (Unit *u)
{
	return radius < u->radius;
}

double Unit::getDistance(const featurevector *sample) const {
	/// @todo implement the distance from heterogenous kohonen !

	double dist = 0.0, r;
	int dims = 0;
	featurevector::const_iterator pos_iter = position.begin();
	featurevector::const_iterator sam_iter = sample->begin();

	while ((pos_iter != position.end()) &&
		(sam_iter != sample->end())) {
		if ((*pos_iter) != NULL && (*sam_iter) != NULL) {
		        double curDist = (*pos_iter)->getDistance(*sam_iter);
			// check if it is normalized
#if _DEBUG_FEATURES
/*			if (curDist >= 0.5) {
				plog("%s --- %s === %f\n", (*pos_iter)->toString().c_str(), (*sam_iter)->toString().c_str(), curDist);
			}*/
			if (curDist < 0 || curDist > 1)
				plog("%s --- %s === %f\n", (*pos_iter)->toString().c_str(), (*sam_iter)->toString().c_str(), curDist);
#endif
			assert(curDist >= 0 && curDist <= 1);
			// also check the symmetry here, it's not elegant but works
			// the reason why this is disabled is the following: IT DOES NOT WORK ON ABSTRACTSTRING WITH LEVENSHTEIN ! LEAVE IT DISABLED !
			//assert(fabs((*pos_iter)->getDistance(*sam_iter) - (*sam_iter)->getDistance(*pos_iter)) < 0.0001);
			// and the identity
			//assert(fabs((*pos_iter)->getDistance(*pos_iter)) < 0.0001);
			//assert(fabs((*sam_iter)->getDistance(*sam_iter)) < 0.0001);
			dist += curDist * curDist;
		}

		pos_iter++; sam_iter++; dims++;
	}

	if (dist == 0) {
		return 0.0;
	} else {
		r = sqrt(dist) / sqrt((double) dims);
		assert(r >= 0 && r <= 1);
		return r;
	}
}

// move the unit itself and alls its neighbors towards the sample
void Unit::moveTowards(const featurevector *sample) {
	// first the unit itself
	movePositionTowards(sample, true);

	// and now all neighbors
	for (edge_map::iterator iter = edge.begin(); iter != edge.end(); iter++) {
		assert((*iter).first != NULL);
		//if ((*iter).first->getDistance(sample) < epsilon) {
			(*iter).first->movePositionTowards(sample, false);
		//}
	}

	// now, since this unit's position (and the position of its neighbors)
	// changed, we have to recalculate our local similarity measure
	calculateLocalSimilarity();
	for (edge_map::const_iterator iter2 = edge.begin(); iter2 != edge.end(); iter2++) {
		(*iter2).first->calculateLocalSimilarity();
	}
}

// this just moves the single position
void Unit::movePositionTowards(const featurevector *sample, bool isWinner) {
	featurevector::iterator pos_iter = position.begin();
	featurevector::const_iterator sam_iter = sample->begin();

	while ((pos_iter != position.end()) &&
		   (sam_iter != sample->end())) {
		if ((*pos_iter) != NULL && (*sam_iter) != NULL) {
			(*pos_iter)->moveTowards(*sam_iter, getLearningRate(isWinner));
		} else if (*pos_iter == NULL) {
			plog("Warning: Unit::movePositionTowards encountered a NULL value in the position!\n");
		} // else we simple have encountered a missing sample....

		pos_iter++; sam_iter++;
	}

	updateRadius();
}

double Unit::getQualityLearning() const {
	assert(error_short >= 0 && error_short <= 1);
	assert(error_long >= 0 && error_long <= 1);

	double r = (error_short + 1) / (error_long + 1);
	assert(r >= 0 && r <= 2);

	return r;
}

double Unit::getQualityInsertion() const {
	assert(error_long >= 0 && error_long <= 1);
	assert(insertion_threshold >= 0 && insertion_threshold <= 1);
	assert(GNG::insertion_tolerance >= 0 && GNG::insertion_tolerance <= 1);

	return error_long - insertion_threshold * (1 + GNG::insertion_tolerance);
}

double Unit::getLearningRate(bool isWinner) const {
	double a = getQualityLearning() / (1 + GNG::adaptation_threshold) + youth - 1;
	double n = max(0.0, min(1.0, a)) * (isWinner ? GNG::lr_winner : GNG::lr_neighbor);

	assert(n >= 0 && n <= 1);

	// now check if this is smaller than the minimum learning rate
	/// @todo implement getActivation()
	//nm = n * (1 - getActivation()) * minimum_lr;
	double nm = 0;
	if (n >= nm)
		return n;
	else
		return nm;
}

#if OPTIMIZATIONS_CACHE_SIMILARITY
void Unit::calculateLocalSimilarity() const {
#else
double Unit::getLocalSimilarity() const {
#endif
	double s = 0;
	for (edge_map::const_iterator iter = edge.begin(); iter != edge.end(); iter++) {
		s += getDistance((*iter).first->getPosition());
	}
	// a unit with no edges should not happen.....
#if OPTIMIZATIONS_CACHE_SIMILARITY
	helper_local_similarity = (getEdgeCount() > 0) ? s / getEdgeCount() : 0;
	assert(helper_local_similarity >= 0 && helper_local_similarity <= 1);
}

double Unit::getLocalSimilarity() const {
	if (helper_local_similarity == -1)
		// unitialized
		calculateLocalSimilarity();
	assert(helper_local_similarity >= 0 && helper_local_similarity <= 1);
	return helper_local_similarity;
}
#else
	double ret = (getEdgeCount() > 0) ? s / getEdgeCount() : 0;
	assert(ret >= 0 && ret <= 1);
	return ret;
}

void Unit::calculateLocalSimilarity() const {}
#endif

void Unit::ageEdge(Unit* u) {
	edge[u]++;
}

void Unit::ageEdges() {
	unit_list u;

	for (edge_map::iterator iter = edge.begin(); iter != edge.end(); iter++) {
		// always check for symmetry...
		assert((*iter).second == (*iter).first->edge[this]);
		(*iter).second++;						// increment the age
		(*iter).first->ageEdge(this);					// but do it symmetrically
		assert((*iter).second == (*iter).first->edge[this]);
		if ((*iter).second > GNG::edge_age_max) {
			u.push_back((*iter).first);				// remove all edges with age > maxage
		}
	}

	/**
	 * @todo optimize here ! we will have splits with new meta IDs just before
	 *       removing those newly split units again (in GNG::nextSample) !
	 */
	for (unit_list::iterator iter2 = u.begin(); iter2 != u.end(); iter2++) {
		removeEdge(*iter2);
	}
}

// this is always symmetric, i.e. an edge from u to this will also be inserted
void Unit::addEdge(Unit* u) {
	assert(meta_id != 0);
	if ((meta_id != u->meta_id) && (u->meta_id != 0)) {
		// this already prevents merging way too early ! find some better way to do it, maybe just stop merging metas-clusters which have already been named
		// don't merge meta clusters
		//plog("!");

		plog("m");
		// the meta id of the new merged cluster can be reused
		releaseMetaId(u->getMetaId());
		u->setMetaId(meta_id);

		insertEdge(u);
		u->insertEdge(this);
	} else {
		u->meta_id = meta_id;

		insertEdge(u);
		u->insertEdge(this);
	}
}

// this can only be called from addEdge, never directly
void Unit::insertEdge(Unit* u) {
	edge[u] = 0;
}

// this is always symmetric, i.e. the edge from u to this will also be removed
void Unit::removeEdge(Unit* u) {
	eraseEdge(u);
	u->eraseEdge(this);

	// optimization: if one of the two nodes has now lost its last edge, don't to a full split because this node will be removed anyways (as soon as this calls returns) by GNG::nextSample
#if OPTIMIZATIONS_SPLIT_MERGE
	if (getEdgeCount() == 0) {
		meta_id = 0;
	} else if (u->getEdgeCount() == 0) {
		u->meta_id = 0;
	} else
#endif
	if (!isConnected(u)) {
		plog("s");
		u->updateMetaId();
	}
}

// this can only be called from removeEdge, never directly
void Unit::eraseEdge(Unit* u) {
	edge.erase(edge.find(u));
}

void Unit::setEdgeAge(Unit* u, unsigned long age) {
	edge[u] = age;
	u->edge[this] = age;
}

edge_map::size_type Unit::getEdgeCount() const {
	return edge.size();
}

edge_map* Unit::getEdges() {
	return &edge;
}

string Unit::serialize() const {
	string ret;
	char buffer [sizeof(double)*8+1];

	// first the position
	ret += serializeFeatureVector(position);

	// and now the unit variables
	sprintf(buffer, "%.17f", error_short);
	ret += buffer;
	ret += ";";
	sprintf(buffer, "%.17f", error_long);
	ret += buffer;
	ret += ";";
	sprintf(buffer, "%.17f", error_insert);
	ret += buffer;
	ret += ";";
	sprintf(buffer, "%.17f", insertion_threshold);
	ret += buffer;
	ret += ";";
	sprintf(buffer, "%.17f", youth);
	ret += buffer;
	ret += ";";
	sprintf(buffer, "%li", meta_id);
	ret += buffer;
	ret += ";";

	// the edges are done by GNG::serialize
	return ret;
}

size_t Unit::unserialize(const featurevector& features, const string data) {
	// the same as in the constructor
	helper_local_similarity = -1;

	// first we get the position
	char* tmpdata = strdup(data.c_str());
	size_t pos = unserializeFeatureVector(features, position, tmpdata), len;

	// now the unit variables
	char* tok = strsplit(tmpdata+pos+1, ';');
	assert(tok != NULL);
	error_short = atof(tok);
	tok = strsplit(NULL, ';');
	assert(tok != NULL);
	error_long = atof(tok);
	tok = strsplit(NULL, ';');
	assert(tok != NULL);
	error_insert = atof(tok);
	tok = strsplit(NULL, ';');
	assert(tok != NULL);
	insertion_threshold = atof(tok);
	tok = strsplit(NULL, ';');
	assert(tok != NULL);
	youth = atof(tok);
	tok = strsplit(NULL, ';');
	assert(tok != NULL);
	meta_id = atol(tok);

	bool found = false;
	for (list<unsigned long>::iterator iter = used_meta_ids.begin(); !found && iter != used_meta_ids.end(); iter++) {
		if (*iter >= meta_id)
			found = true;
		if (*iter > meta_id)
			used_meta_ids.insert(iter, meta_id);
	}
	if (!found) {
		used_meta_ids.push_back(meta_id);
	}

	updateRadius();

	len = tok + strlen(tok) - tmpdata + 1;
	free(tmpdata);
	return len;
}

void Unit::updateMetaId()
{
	setMetaId(getNextMetaId());
}

void Unit::setMetaId(unsigned long id)
{
	meta_id = id;

	for (edge_map::iterator iter = edge.begin(); iter != edge.end(); iter++) {
		if ((*iter).first->meta_id != id) {
			(*iter).first->setMetaId(id);
		}
	}
}

bool Unit::isConnectedHelper(Unit* u) const
{
	/* unit already visited ? */
	if (!traversing) {
		traversing = true;

		for (edge_map::const_iterator iter = edge.begin(); iter != edge.end(); iter++) {
			if ((*iter).first == u) {
				return true;
			} else {
				if ((*iter).first->isConnectedHelper(u)) {
					return true;
				}
			}
		}
	}

	return false;
}

bool Unit::isConnected(Unit* u) const
{
	// reset all traversing flags
	for (unit_list::const_iterator iter = all_units->begin(); iter != all_units->end(); iter++) {
		(*iter)->traversing = false;
	}

	return isConnectedHelper(u);
}

unsigned long Unit::getNextMetaId()
{
	// return the smallest available meta id that has not been used
	unsigned long id = 1;
	bool found = false;

	// the list is always sorted, so just start from the beginning
	for (list<unsigned long>::iterator iter = used_meta_ids.begin(); !found && iter != used_meta_ids.end(); iter++) {
		if (*iter != id) {
			// ok, we found a hole
			found = true;
			// insert it before this position into the list
			used_meta_ids.insert(iter, id);
		}
		else
			id++;
	}
	// id will be the highest id in the list + 1 here (which can also be 1 if the list is empty)
	if (!found) {
		used_meta_ids.push_back(id);
	}
	return id;
}

void Unit::reserveMetaId(unsigned long meta_id)
{
	bool found = false;

	// the list is always sorted, so just start from the beginning
	for (list<unsigned long>::iterator iter = used_meta_ids.begin(); !found && iter != used_meta_ids.end(); iter++) {
		if (*iter != meta_id) {
			// ok, currently not reserved, just do it
			found = true;
			// insert it before this position into the list
			used_meta_ids.insert(iter, meta_id);
		}
	}
	if (!found)
		used_meta_ids.push_back(meta_id);
}

void Unit::releaseMetaId(unsigned long meta_id)
{
	// simply remove the given id from the list
	used_meta_ids.remove(meta_id);
}

size_t Unit::getNumMetaIds()
{
	return used_meta_ids.size();
}

unsigned long Unit::getMaxMetaId()
{
	if (used_meta_ids.size() == 0)
		return 0;
	else
		return used_meta_ids.back();
}

void Unit::clearMetaIds()
{
	used_meta_ids.clear();
}

void Unit::updateRadius()
{
#if OPTIMIZATIONS_SPLAYTREE
	double pos;
	int dims = 0;
	radius = 0.0;

	for (featurevector::iterator iter = position.begin(); iter != position.end(); iter++) {
		if(*iter != NULL) {
			pos = (*iter)->getPosition();
			radius += pos * pos;
			dims++;
		}
	}

	if (dims > 0) {
		radius = sqrt(radius) / sqrt((double) dims);
	}
	assert((radius >= 0.0) && (radius <= 1.0));
#endif
}
