/**
 * @file
 * @brief Configuration parser
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-08-29 11:15:58 +0200 (Son, 29 Aug 2004) $
 * $Revision: 904 $
 *
 * @since    Mon Aug 11 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: ConfigReader.cpp 904 2004-08-29 09:15:58Z rene $
 */

#include <scew.h>
#include "ConfigReader.h"

ConfigReader::ConfigReader(const char *file)
{
	scew_element *item;
	unsigned cnt;

	// check if the file exists....
	FILE* tmp;
	if ((tmp=fopen(file, "r")) == NULL) {
		plog("Error: config file '%s' does not exist!\n", file);
		throw 1;
	} else {
		fclose(tmp);
	}

	scew_parser* parser = scew_parser_create();
	scew_parser_load_file(parser, file);
	
	scew_tree *tree =  scew_parser_tree(parser);
	scew_element *root = scew_tree_root(tree);

	item = scew_element_by_name(root, (XML_Char *) "classifier");
	const XML_Char *algo = scew_attribute_value(scew_attribute_by_name(item, (XML_Char*) "id"));
	if (algo != NULL) {
		classifier.first = string((char*) algo);

		unsigned params = scew_element_count(item);
		for (unsigned k=0; k < params; k++) {
			const scew_element *param = scew_element_by_index(item, k);

			const XML_Char *attr = scew_attribute_value(scew_attribute_by_name(param, (XML_Char*) "id"));
			const XML_Char *val = scew_element_contents(param);

			if ((attr != NULL) && (val != NULL)) {
				classifier.second[string((char*) attr)] = string((char*) val);
			}
		}
	}

	item = scew_element_by_name(root, (XML_Char*) "predictor");
	algo = scew_attribute_value(scew_attribute_by_name(item, (XML_Char*) "id"));
	if (algo != NULL) {
		predictor.first = string((char*) algo);

		unsigned params = scew_element_count(item);
		for (unsigned k=0; k < params; k++) {
			const scew_element *param = scew_element_by_index(item, k);

			const XML_Char *attr = scew_attribute_value(scew_attribute_by_name(param, (XML_Char*) "id"));
			const XML_Char *val = scew_element_contents(param);

			if ((attr != NULL) && (val != NULL)) {
				predictor.second[string((char*) attr)] = string((char*) val);
			}
		}
	}

	item = scew_element_by_name(root, (XML_Char*) "featureproviders");
	cnt = scew_element_count(item);
	for (unsigned j=0; j < cnt; j++) {
		scew_element *node = scew_element_by_index(item, j);
		const XML_Char *prov = scew_attribute_value(scew_attribute_by_name(node, (XML_Char*) "id"));
		// this is used when multiple feature providers of the same type are used
		const XML_Char *label = scew_attribute_value(scew_attribute_by_name(node, (XML_Char*) "label"));

		if (prov != NULL) {
			unsigned params = scew_element_count(node);
			for (unsigned i=0; i < params; i++) {
				const scew_element *param = scew_element_by_index(node, i);

				const XML_Char *attr = scew_attribute_value(scew_attribute_by_name(param, (XML_Char*) "id"));
				const XML_Char *val = scew_element_contents(param);

				if ((attr != NULL) && (val != NULL)) {
					if (label == NULL)
						provider[string((char*) prov)][string((char*) attr)] = string((char*) val);
					else {
						// if we have a label, use it instead of the id
						provider[string((char*) label)][string((char*) attr)] = string((char*) val);
						// but remember the real name
						provider[string((char*) label)]["REAL_ID"] = string((char*) prov);
					}
				}
			}
		}
	}

	item = scew_element_by_name(root, (XML_Char*) "features");
	cnt = scew_element_count(item);
	for (unsigned k=0; k < cnt; k++) {
		scew_element *node = scew_element_by_index(item, k);
		const XML_Char *attr = scew_attribute_value(scew_attribute_by_name(node, (XML_Char*) "id"));
		const XML_Char *aggr = scew_attribute_value(scew_attribute_by_name(node, (XML_Char*) "aggregate"));

		if (attr != NULL) {
			features.push_back(string((char*) attr));

			if (aggr != NULL) {
				aggregates.push_back(string((char*) attr));
			}
		}
	}

	item = scew_element_by_name(root, (XML_Char*) "protocols");
	cnt = scew_element_count(item);
	for (unsigned k2=0; k2 < cnt; k2++) {
		scew_element *node = scew_element_by_index(item, k2);
		const XML_Char *attr = scew_attribute_value(scew_attribute_by_name(node, (XML_Char*) "id"));

		if (attr != NULL) {
			protocols.push_back(string((char*) attr));
		}
	}

	item = scew_element_by_name(root, (XML_Char*) "globals");
	cnt = scew_element_count(item);
	for (unsigned k3=0; k3 < cnt; k3++) {
		scew_element *node = scew_element_by_index(item, k3);
		const XML_Char *attr = scew_attribute_value(scew_attribute_by_name(node, (XML_Char*) "id"));
		const XML_Char *val = scew_element_contents(node);

		if (attr != NULL && val != NULL) {
			globals[string((char*) attr)] = string((char*) val);
		}
	}

	scew_tree_free(tree);
	scew_parser_free(parser);
}

const pair<string, map<string, string> > *ConfigReader::getClassifier()
{
	return &classifier;
}

const vector<string> *ConfigReader::getFeatures()
{
	return &features;
}

const vector<string> *ConfigReader::getAggregates()
{
	return &aggregates;
}

const vector<string> *ConfigReader::getProtocols()
{
	return &protocols;
}

const map<string, map<string, string> > *ConfigReader::getFeatureProvider()
{
	return &provider;
}

const map<string, string> *ConfigReader::getGlobals()
{
	return &globals;
}
