/**
 * @file
 * @brief Configuration parser
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-02-12 21:08:39 +0100 (Don, 12 Feb 2004) $
 * $Revision: 612 $
 *
 * @since    Mon Aug 11 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: ConfigReader.h 612 2004-02-12 20:08:39Z harald $
 */

#ifndef _CONFIGREADER_H
#define _CONFIGREADER_H

#include "../Features/Feature.h"
#include "../Classifiers/Classifier.h"
#include "../Predictors/Predictor.h"

/**
 * Configuration parser
 *
 * This class parses the configuration file and provides
 * the settings as hashtables.
 */
class ConfigReader {
	private:
		/**
		 * Classifier settings
		 *
		 * This variable stores the name of the classifier
		 * algorithm and a map of parameters for the
		 * classifier. The data structure represents the
		 * \<classifier\> section of the configuration
		 * file.
		 */
		pair<string, classifierparams> classifier;

		/**
		 * Predictor settings
		 *
		 * This variable stores the name of the predictor
		 * algorithm and a map of parameters for the
		 * predictor. The data structure represents the
		 * \<predictor\> section of the configuration
		 * file.
		 */
		pair<string, providerparams> predictor;

		/**
		 * Provider settings
		 *
		 * This variable stores additional configuration
		 * parameters for various feature providers. The
		 * data structure represents the
		 * \<featureproviders\> section of the
		 * configuration file.
		 */
		map<string, providerparams> provider;

		/**
		 * Global settings
		 *
		 * This variable stores global settings to
		 * control the framework. The data structure
		 * represents the \<globals\> section of
		 * the configuration file.
		 */
		parametermap globals;

		/**
		 * Features
		 */
		stringvector features;

		/**
		 * Aggregates
		 */
		stringvector aggregates;

		/**
		 * Protocols
		 */
		stringvector protocols;

	public:
		/**
		 * Constructor
		 *
		 * @param file Path to the configuration file
		 */
		ConfigReader(const char *file);
		
		//@{
		/**
		 * @todo documentation
		 * @todo source cleanup (don't copy maps)
		 */

		const pair<string, map<string, string> > *getClassifier();
		const pair<string, map<string, string> > *getPredictor();
		const map<string, map<string, string> > *getFeatureProvider();
		const vector<string> *getFeatures();
		const vector<string> *getAggregates();
		const vector<string> *getProtocols();
		const map<string, string> *getGlobals();
		//@}
};

#endif // _CONFIGREADER_H

