/**
 * @file
 * @brief %Feature container
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2005-02-16 17:00:50 +0100 (Mit, 16 Feb 2005) $
 * $Revision: 920 $
 *
 * @since    Sun Apr 6 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: FeatureContainer.cpp 920 2005-02-16 16:00:50Z rene $
 */

#include "FeatureContainer.h"
#include "../ConfigReader/ConfigReader.h"
#include "../ConfigReader/PersistantReader.h"
#include "../ConfigReader/PersistantWriter.h"

FeatureContainer::FeatureContainer(ConfigReader *config)
{
	map<string, string> globals = (*config->getGlobals());
	persist = globals["persist"];

	plog("FeatureContainer constructor: loading persistant file %s\n", persist.c_str());

	PersistantReader* persistant = new PersistantReader(persist.c_str());

	const map<string, map<string, string> > *provider = config->getFeatureProvider();
	for (map<string, map<string, string> >::const_iterator iter = provider->begin(); iter != provider->end(); iter++) {
		string f_prov;
		featureparams::const_iterator real = (*iter).second.find("REAL_ID");
		if (real != (*iter).second.end()) {
			assert((*real).first == "REAL_ID");
			f_prov = (*real).second;
			plog("FeatureContainer constructor: loading feature provider %s (label %s)", f_prov.c_str(), 
				(*iter).first.c_str());
		}
		else {
			f_prov = (*iter).first;
			plog("FeatureContainer constructor: loading feature provider %s", f_prov.c_str());
		}

		if (loadProvider(f_prov.c_str(), (*iter).second) != NULL) {
			plog(" successful\n");
		} else {
			plog(" failed\n");
		}
	}

	const vector<string> *bar = config->getFeatures();
	for (unsigned int i=0; i<bar->size(); i++) {
		plog("FeatureContainer constructor: loading feature %s", (*bar)[i].c_str());

		if (loadFeature((*bar)[i].c_str())) {
			plog(" successful\n");
		} else {
			plog(" failed\n");
		}
	}

	for (featuremap::iterator iter2 = features.begin(); iter2 != features.end(); iter2++) {
		const char *p, *feat = (*iter2).first.c_str();
		featureparams *data;
		for (p = feat; *p != '.'; p++); p++;
		Feature *f = (*iter2).second->getFeature(p);

		if (f != NULL) {
			if (f->isExternalizable() && ((data = persistant->getPersistantData(feat)) != NULL)) {
				plog("FeatureContainer constructor: initializing feature %s\n", feat);

				((PersistantFeature *) f)->read(data);
			}
			delete f;
		}
	}

#ifndef _WIN32_WCE
	tzset();
#endif

	delete persistant;
}

FeatureContainer::~FeatureContainer()
{
	/* free sample vector */
	for (featurevector::iterator iter = feature.begin(); iter != feature.end(); iter++) {
		if (*iter != NULL) {
			delete (*iter);
		}
	}

	freeProvider();
}

bool FeatureContainer::loadFeature(const char *id)
{
	const map<string, string> dummy;

	char *conf = strdup(id);
	char *prov = strtok(conf, ".");
	char *feat = strtok(NULL, "");
	bool found = false;

	FeatureProvider *fp = loadProvider(prov, dummy);

	if (fp == NULL) {
		//error
		free(conf);
		return false;
	}

	provider[prov] = fp;
	stringvector *fl = fp->getFeatureList();

	for (stringvector::iterator iter = fl->begin(); iter != fl->end(); iter++) {
		if (*iter == feat) {
			found = true;
			break;
		}
	}

	if (!found || (feat == NULL)) {
		// error
		freeProvider(handle.back());
		handle.pop_back();
	} else {
		features.push_back(featurepair(id, fp));
	}

	free(conf);
	return found;
}

void FeatureContainer::nextSample()
{
	map<string, featureparams> persistdata;
	clock_t now = clock();
	bool v = true;

	/* aquire a new sample for each feature */
	for (providermap::iterator iter = provider.begin(); iter != provider.end(); iter++) {
		plog("FeatureContainer::nextSample calling nextSample on feature %s\n", (*iter).second->getName().c_str());
		(*iter).second->nextSample(now);
		plog("done\n");
	}

	/* free old features */
	for (featurevector::iterator iter1 = feature.begin(); iter1 != feature.end(); iter1++) {
		if (*iter1 != NULL) {
			plog("FeatureContainer::nextSample freeing feature %s\n", (*iter1)->getName().c_str());
			delete (*iter1);
		}
	}
	feature.clear();

	for (featuremap::const_iterator iter2 = features.begin(); iter2 != features.end(); iter2++) {
		const char *p, *feat = (*iter2).first.c_str();
		for (p = feat; *p != '.'; p++); p++;
		
		plog("FeatureContainer::nextSample calling getSample on feature %s\n", (*iter2).second->getName().c_str());
		Feature *f = (*iter2).second->getSample(p);
		feature.push_back(f);

		if ((f != NULL) && (f->isExternalizable())) {
			PersistantFeature *p = (PersistantFeature *) f;
			if (!p->isValid()) {
				v = false;
				persistdata[feat] = p->write();
				p->validate();
			}
		}
	}

	if (!v) {
		PersistantWriter::write(persist.c_str(), persistdata);
	}
}

featurevector FeatureContainer::getFeatureVector() const
{
	featurevector v;

	for (featuremap::const_iterator iter = features.begin(); iter != features.end(); iter++) {
		const char *p, *feat = (*iter).first.c_str();
		for (p = feat; *p != '.'; p++); p++;

		v.push_back((*iter).second->getFeature(p));
	}

	return v;
}

const featurevector* FeatureContainer::getSampleVector() const
{
	return &feature;
}

void FeatureContainer::freeProvider()
{
	plog("FeatureContainer::freeProvider called\n");

	for (vector<void *>::iterator iter = handle.begin(); iter != handle.end(); iter++) {
		plog("FeatureContainer::freeProvider\n");

		freeProvider(*iter);
	}
}
