/**
 * @file
 * @brief %Feature container
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-03-02 13:36:47 +0100 (Die, 02 Mär 2004) $
 * $Revision: 670 $
 *
 * @since    Sun Apr 6 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: FeatureContainer.h 670 2004-03-02 12:36:47Z rene-cvs $
 */

#ifndef _FEATURECONTAINER_H
#define _FEATURECONTAINER_H

#include <stdio.h>
#include "../Features/Feature.h"

class ConfigReader;

/**
 * Container class
 *
 * Loads and initializes the feature providers as defined
 * in the configuration file. It provides methods to commonly
 * operate on all loaded providers and their requested features.
 */
class FeatureContainer {
	private:
		/** Map of FeatureProvider handles */
		typedef map<string, FeatureProvider*> providermap;

		//@{
		/**
		 * Private property
		 * @todo documentation
		 */
		string persist;
		vector<void*> handle;
		providermap provider;
		featurevector feature;
		FILE *out;
		//@}

		/**
		 * Load a feature
		 *
		 * A feature is identified by a unique name of the form
		 * \<providername\>.\<featurename\>. loadFeature
		 * calls loadProvider with the \<providername\> part
		 * to load the correct feature provider and requests the
		 * feature \<featurename\> from it.
		 *
		 * @param id Unique name of the feature
		 * @return <var>true</var> on success, otherwise <var>false</var>
		 */
		bool loadFeature(const char *id);
		/**
		 * Load a provider
		 *
		 * Like a feature a provider must also have a unique name
		 * (this constraint is also given by the fact that a feature
		 * provider is a shared library). loadProvider locates the
		 * feature provider in the file system and loads it into the
		 * framework passing a providerparams datastructure to it for
		 * initialization.
		 *
		 * @param provider Unique name of the feature provider
		 * @param args A providerparams datastructure containing additional
		 *             parameters from the configuration file.
		 * @return An instance of a FeatureProvider class
		 */
		FeatureProvider* loadProvider(const char *provider, const providerparams &args);
		/**
		 * Unload a provider
		 *
		 * Cleans up all used memory and unloads the library.
		 *
		 * @param provider Handle of the provider library
		 */
		void freeProvider(void *provider);
		/**
		 * Cleanup container
		 *
		 * This function calls freeProvider for every loaded provider
		 * in the container. It is intended to be called on shutdown
		 * of the framework.
		 */
		void freeProvider();

	protected:
		//@{
		/**
		 * Protected property
		 * @todo documentation
		 */

		typedef pair<string, FeatureProvider*> featurepair;
		typedef vector<featurepair> featuremap;

		featuremap features;
		//@}

	public:
		/**
		 * Constructor
		 *
		 * Loads all feature providers listed in the configuration
		 * file and initializes them with the according parameters.
		 * It then loads all requested features and restores persistant
		 * information for them.
		 *
		 * @param config A ConfigReader instance
		 */
		FeatureContainer(ConfigReader *config);
		/** Destructor */
		virtual ~FeatureContainer();

		/**
		 * Retrieve next sample
		 *
		 * Retrieve next samples from all features at once. This
		 * method should be called at the beginning of each time step.
		 * This method calls nextSample for each FeatureContainer and
		 * then allocates the global feature vector which can be
		 * returned by successive calls to getSampleVector. The specific
		 * feature objects are created by the FeatureContainer
		 * implementations, as this method calls getSample for every
		 * dimension.
		 *
		 * @see FeatureContainer::nextSample
		 * @see FeatureContainer::getSample
		 * @see getSampleVector
		 */
		virtual void nextSample();

		/**
		 * Get random feature vector
		 *
		 * A feature vector contains pointers to implementations of the
		 * specific features, but initialized with random values. These
		 * random positions in the feature space are usually used for
		 * initializing classification algorithms. The feature objects
		 * returned by this method are allocated automatically for the
		 * caller (since the caller can not know the specific
		 * implementations, this method acts as a fectors), but
		 * <b>must</b> be freed by the caller. The specific feature
		 * objects are created by the FeatureContainer implementations,
		 * as this method calls getFeature for every dimension.
		 *
		 * @return A vector of Feature implementations. Every element
		 *         of the returned vector must be freed after use.
		 * @see Feature
		 * @see FeatureContainer::getFeature
		 */
		virtual featurevector getFeatureVector() const;

		/**
		 * Get sample vector
		 *
		 * A sample vector contains pointers to implementations of the
		 * specific features, with values representing the current
		 * sensor values. For performance reasons (since this method can
		 * be called by multiple callers in the same time step for
		 * retrieving the current feature values), this method returns
		 * a pointer to a globally allocated object. Callers
		 * <b>must not</b> free or modify either the returned vector
		 * reference or the feature implementations contained therein.
		 *
		 * @return A reference to a globally allocated vector containing
		 *         features that represent the current sensor values.
		 */
		virtual const featurevector* getSampleVector() const;
};

#endif // _FEATURECONTAINER_H
