/**
 * @file
 * @brief Abstract string feature class implementation
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-06-04 13:37:10 +0200 (Fre, 04 Jun 2004) $
 * $Revision: 798 $
 *
 * @since    Wed Mar 19 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: AbstractString.cpp 798 2004-06-04 11:37:10Z rene $
 */

// for sprintf
#include <stdio.h>
#include "AbstractString.h"

/**
 * Modified Levenshtein algorithm
 *
 * This function implements a slightly modified version
 * of the Levenshtein algorithm to not only calculate the
 * distance of two strings but also modify the string <var>x</var>
 * towards the string <var>t</var> with the propability
 * <var>factor</var>. In other words, every transformation
 * operation found by the algorithm is performed on the string
 * <var>x</var> with the propability <var>factor</var>.
 *
 * @param x Input string.
 * @param t String to compare the input string to.
 * @param factor Propability with witch transformations are performed.
 *               Has to be a value out of the intervall <var>[0; 1]</var>.
 * @return The levenshtein distance between <var>x</var> and <var>t</var>.
 * @relates AbstractStringFeature
 */
static long levenshtein (char **x, const char *t, double *factor)
{
	size_t o;
	size_t n; // length of s
	size_t m; // length of t
	unsigned *d; // pounsigneder to matrix
	unsigned i; // iterates through s
	unsigned j; // iterates through t
	unsigned above; // contents of cell immediately above
	unsigned left; // contents of cell immediately to left
	unsigned diag; // contents of cell immediately above and to left
	unsigned current; // result
	unsigned result; // result
	char *s = *x;
	assert(s != NULL);

	n = strlen(s);
	m = strlen(t);

	// exit if one of the strings is empty and we don't have to modify the string
	if (((m == 0) || (n == 0)) && (factor == NULL)) {
#if 0
		if(m != n)
			plog("Levenshtein: NULL string encountered\n");
#endif
		return (long) max(m, n);
	}

	// alloc matrix
	d = (unsigned *) malloc((n+1) * (m+1) * sizeof(unsigned));
	assert(d != NULL);

	// init row
	for (i = 0; i <= n; i++) {
		*(d + i) = i;
	}

	// init col
	for (j = 0; j <= m; j++) {
		*(d + (j * (n + 1))) = j;
	}

	// calculate matrix
	for (i = 1; i <= n; i++) {
		for (j = 1; j <= m; j++) {
			above = *(d + i - 1 + (j * (n + 1))) + 1;
			left = *(d + i + ((j - 1) * (n + 1))) + 1;
			diag = *(d + i - 1 + ((j - 1) * (n + 1)));

			if (s[i-1] != t[j-1]) {
				diag++;
			}

			*(d + i + (j * (n + 1))) = min(above, min(left, diag));
		}
	}

	// store result
	result = *(d + n + (m * (n + 1)));

	// adapt string
	if (result && (factor != NULL)) {
		o = m + n;
		
		/// @todo alloc once
		char *z = (char *) malloc(o + 1); // alloc new string
		assert (z != NULL);

		i--; j--;
		z[o--] = '\0';
		// walk back the matrix
		while ((i > 0) && (j > 0)) {
			bool match = true;
			int last = 0;

			diag = *(d + i - 1 + ((j - 1) * (n + 1)));
			current = *(d + i + (j * (n + 1)));

			if (s[i-1] != t[j-1]) {
				match = false;
				diag++;
			}

			if (current == diag) {
				// change
				if (!match && (rand_double() < *factor)) {
					z[o--] = t[j-1];
				} else {
					z[o--] = s[i-1];
				}
				i--; j--;
				last = 0;
			} else if (current == *(d + i - 1 + (j * (n + 1)))) {
				// delete
				assert(last != 2);
				if (rand_double() > *factor) {
					z[o--] = s[i-1];
				}
				i--;
				last = 1;
			} else {
				// insert
				assert(last != 1);
				if (rand_double() < *factor) {
					z[o--] = t[j-1];
				}
				j--;
				last = 2;
			}
		}
		assert((int) o >= -1);
		while (i > 0) {
			// delete
			if (rand_double() > *factor) {
				z[o--] = s[i-1];
			}
			i--;
		}
		while (j > 0) {
			// insert
			if (rand_double() < *factor) {
				z[o--] = t[j-1];
			}
			j--;
		}
		assert((int) o >= -1);

		*x = strdup(&z[o+1]);
		assert(*x != NULL);
		free(s);
		free(z);
	}

	// free matrix
	free (d);

	if (factor == NULL)
		return result;
	else
		// it's not used anyway when adapting
		return -1;
}

AbstractStringFeature::AbstractStringFeature(ComparatorType comp, stringcode *code, long *maxlen)
{
	comparator = comp;
	this->code = code;
	maxlength = maxlen;
	codeval = (long) (rand_double() * code->size());
	position = NULL;

	// this is a unit, not a sample. init the position member
	for (stringcode::iterator iter = code->begin(); iter != code->end(); iter++) {
		if ((*iter).second == codeval) {
			position = strdup((*iter).first.c_str());
			break;
		}
	}

	if (position == NULL)
		// this means that until now, no single code has been put into our code list
		// --> we are at initialization stage without any previous sample
		// simply initialize as an empty string here
		position = strdup("");
}

AbstractStringFeature::AbstractStringFeature(ComparatorType comp, stringcode *code, long *maxlen, const string &name)
{
	comparator = comp;
	this->code = code;
	maxlength = maxlen;
	codeval = getCodeForName(name);
	position = NULL;
}

AbstractStringFeature::~AbstractStringFeature()
{
	if (position != NULL) {
		free(position);
	}
}

string AbstractStringFeature::serialize() const
{
	char buffer[sizeof(long)*3+1];

	switch (comparator) {
		case Levenshtein:
			if (position != NULL) {
				string ret = "#";
				ret += position;

				return ret;
			}
			break;
		case None:
			// just do nothing here, only shut up the compiler warning
			break;
	}

	sprintf(buffer, "%ld", codeval);

	return buffer;
}

void AbstractStringFeature::unserialize(string value)
{
	const char *s = value.c_str();

	if (*maxlength == 0) {
		for (stringcode::iterator iter = code->begin(); iter != code->end(); iter++) {
			long l = (long)(*iter).first.size();
			if (l > *maxlength) {
				*maxlength = l;
			}
		}
	}

	switch (comparator) {
		case Levenshtein:
			if (position != NULL) {
				free(position);
				position = NULL;
			}
			if (s[0] == '#') {
				position = strdup(&s[1]);
				return;
			}
			break;
		case None:
			// just do nothing here, only shut up the compiler warning
			break;
	}

	codeval = atol(s);
}

featureparams AbstractStringFeature::write() const
{
	featureparams param;

	for (stringcode::iterator iter = code->begin(); iter != code->end(); iter++) {
		char buffer[sizeof(long)*3+1];
		sprintf(buffer, "%ld", (*iter).second);

		param[buffer] = (*iter).first;
	}
	return param;
}

void AbstractStringFeature::read(featureparams *param)
{
	for (featureparams::iterator iter = param->begin(); iter != param->end(); iter++) {
		(*this->code)[(*iter).second] = atol((*iter).first.c_str());
	}
}

unsigned long AbstractStringFeature::getCodeForName(const string &name) {
	this->name = name;

	if (code->count(name) == 0) {
		/* separate the two lines because the line below implicitely increments size */
		unsigned long nextfree = (unsigned long) this->code->size();
		(*code)[name] = nextfree;

		if ((long) name.size() > *maxlength) {
			*maxlength = (long) name.size();
		}

		invalidate();

		return nextfree;
	} else {
		return (*code)[name];
	}
}

double AbstractStringFeature::getPosition() const
{
	double r;

	if ((position == NULL) || (*maxlength == 0)) {
		r = 0.0;
	} else {
		// distance from "" string
		r = (double) strlen(position) / *maxlength;
	}
	assert(r >= 0.0 && r <= 1.0);
	return r;
}

double AbstractStringFeature::getDistance(Feature *f) const
{
	switch (comparator) {
		case None:
			if (codeval == ((AbstractStringFeature*) f)->codeval) {
				return 0.0;
			} else {
				return 1.0;
			}

		case Levenshtein: {
			string s;

			/// @todo implement a better 'get string for code'
			// codevals are different, calculate distance
			for (stringcode::iterator iter = code->begin(); iter != code->end(); iter++) {
				if ((*iter).second == ((AbstractStringFeature*) f)->codeval) {
					s = (*iter).first;
					break;
				}
			}

			assert(position != NULL);
			if (*maxlength > 0) {
				return (double) levenshtein(const_cast<char**> (&position), s.c_str(), NULL) / *maxlength;
			} else {
				return 0.0;
			}
		}
	}

	return 1.0;
}

void AbstractStringFeature::moveTowards(Feature *f, double factor)
{
    assert(factor >= 0.0 && factor <= 1.0);
	switch (comparator) {
		case None:
			if (rand_double() < factor) {
				codeval = ((AbstractStringFeature*) f)->codeval;
			}
			break;

		case Levenshtein: {
			string s;

			/// @todo implement a better 'get string for code'
			// codevals are different, calculate distance
			for (stringcode::iterator iter = code->begin(); iter != code->end(); iter++) {
				if ((*iter).second == ((AbstractStringFeature*) f)->codeval) {
					s = (*iter).first;
					break;
				}
			}

			assert(position != NULL);
			factor *= 100;
			if (factor > 1.0)
				factor = 1.0;
			levenshtein(&position, s.c_str(), &factor);
			break;
		}
	}
}

Feature* AbstractStringFeature::clone() const
{
	AbstractStringFeature* f = new AbstractStringFeature(comparator, code, maxlength);
	f->codeval = codeval;

	if (f->position != NULL) {
		free(f->position);
		f->position = NULL;
	}

	if (position != NULL) {
		f->position = strdup(position);
	}

	f->name = name;
	return f;
}

void AbstractStringFeature::aggregate(aggregatelist )
{
}

#if _DEBUG_FEATURES
string AbstractStringFeature::toString() const
{
	char buffer[256];
	sprintf(buffer, "general AbstractStringFeature: '%s', has code %s", name.c_str(), serialize().c_str());
	return buffer;
}
#endif


AbstractStringListFeature::AbstractStringListFeature(stringcode *code, long *maxlen) : AbstractStringFeature(AbstractStringFeature::None, code, maxlen)
{
	listval.resize(code->size());
	for (bit_vector::iterator iter = listval.begin(); iter != listval.end(); iter++) {
		(*iter) = (rand_double() > 0.5) ? true : false;
	}
}

AbstractStringListFeature::AbstractStringListFeature(stringcode *code, long *maxlen, const stringvector *names) : AbstractStringFeature(AbstractStringFeature::None, code, maxlen)
{
	if (names != NULL) {
		this->listval = getCodeForList(names);
	}
}

bit_vector AbstractStringListFeature::getCodeForList(const stringvector *names)
{
	// so that getNames will work, a bit of ineffeciency here...
	if (names != NULL) {
		for (stringvector::const_iterator iter = names->begin(); iter != names->end(); iter++) {
			this->names.push_back(*iter);
		}
	}

	// this is the maximum length that the vector can grow here
	bit_vector ret(code->size() + names->size(), false);
	if (names != NULL) {
		for (stringvector::const_iterator iter2 = names->begin(); iter2 != names->end(); iter2++) {
			ret[getCodeForName(*iter2)] = true;
		}
	}

	return ret;
}

string AbstractStringListFeature::serialize() const
{
	string ret;
	for (bit_vector::const_iterator iter = listval.begin(); iter != listval.end(); iter++) {
		ret += (*iter) ? '1' : '0';
	}

	return ret;
}

void AbstractStringListFeature::unserialize(string value)
{
	// here we know the length for the bit vector....
	listval.resize(value.length(), false);

	for (unsigned int i=0; i<value.length(); i++) {
		if (value[i] == '1') {
			listval[i] = true;
		} else if (value[i] == '0') {
			listval[i] = false;
		} else {
			plog("Error: unkown string on unserialize, string was '%s'!\n", value.c_str());

			throw 0;
		}
	}
}

double AbstractStringListFeature::getPosition() const
{
	long i = 0;
	// distance from the empty bitvector
	for (bit_vector::const_iterator iter = listval.begin(); iter != listval.end(); iter++) {
		if (*iter) {
			i++;
		}
	}

	if (code->size() == 0) {
		// empty lists are always the same
		return 0.0;
	} else {
		double r = (double) i / code->size();
		assert(r >= 0 && r <= 1);
		return r;
	}
}

double AbstractStringListFeature::getDistance(Feature *f) const
{
	size_t dist = 0;
	AbstractStringListFeature* s = (AbstractStringListFeature*) f;

	for (unsigned int i=0; i<listval.size() || i<s->listval.size(); i++) {
		if ((i >= listval.size() && s->listval[i])	||
			(i >= s->listval.size()  && listval[i]) ||
			(i < listval.size() && i < s->listval.size() && listval[i] != s->listval[i])) {
			dist++;
		}
	}

	if (dist > code->size()) {
		/// @todo this should not happen, find the reason for that
		dist = code->size();
	}

	if (code->size() == 0) {
		// empty lists are always the same
		return 0.0;
	} else {
		double r = (double) dist / code->size();
		assert(r >= 0.0 && r <= 1.0);
		return r;
	}
}

void AbstractStringListFeature::moveTowards(Feature *f, double factor)
{
    assert(factor >= 0.0 && factor <= 1.0);
	// change, according to factor, a fraction of the listval bits to those from s
	AbstractStringListFeature* s = (AbstractStringListFeature*) f;

	listval.resize(max(listval.size(), s->listval.size()), false);
	s->listval.resize(listval.size(), false);

	// count the number of different bits
	unsigned int numbits = 0;
	for (unsigned int i=0; i < listval.size(); i++) {
		if (s->listval[i] != listval[i]) {
			numbits++;
		}
	}

	// thus, each bit should be changed with a "force factor" of ...
	double force = ((double) max(listval.size(), s->listval.size()) / numbits);

	// We simply change each bit that is different with a probability
	// dependent on numbits. Even if this is not completely correct, it
	// is more performant than trying to find a random selection of exactly
	// numbits different bits to change.
	for (unsigned int i2=0; i2 < listval.size(); i2++) {
		if (s->listval[i2] != listval[i2]) {
			// this bit is different (code taken from getDistance)
			// change it with the given probability
			if (rand_double() < factor * force) {
				listval[i2] = s->listval[i2];
			}
		}
	}
}

Feature* AbstractStringListFeature::clone() const
{
	AbstractStringListFeature* f = new AbstractStringListFeature(code, maxlength);
	f->listval = listval;
        // so that getNames will work
	for (stringvector::const_iterator iter = names.begin(); iter != names.end(); iter++) {
		f->names.push_back(*iter);
	}
	return f;
}

#if _DEBUG_FEATURES
string AbstractStringListFeature::toString() const
{
	string ret = "general AbstractStringListFeature: '";
	for (unsigned int i = 0; i < getListValues().size(); i++) {
		ret += getListValues()[i];
		ret += ", ";
	}
	ret += "', has code ";
	ret += serialize();

	return ret;
}
#endif
