/**
 * @file
 * @brief Abstract string feature class declaration
 *
 * This file contains the abstract class definitions that can be used
 * by sensors that return values that can be represented as strings.
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-03-01 12:15:32 +0100 (Mon, 01 Mär 2004) $
 * $Revision: 664 $
 *
 * @since    Wed Mar 19 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: AbstractString.h 664 2004-03-01 11:15:32Z rene $
 */

#ifndef _ABSTRACTSTRING_H
#define _ABSTRACTSTRING_H

#include "Feature.h"

/** An association of seen strings and their assigned numerical code */
typedef map<string, unsigned long> stringcode;

/**
 * Baseclass for features providing string values
 *
 * A feature of type nominal which has character strings as values. For
 * comparing values of this feature, either the normal string comparion with
 * the number of different characters (Manhattan distance) or the Levenshtein
 * distance can be used.
 */
class AbstractStringFeature: public PersistantFeature
{
	public:
		/** The distance metric to use */
		typedef enum ComparatorType { None, Levenshtein };

	protected:
		/**
		 * %Feature constructor
		 *
		 * This constructor initializes the feature as prototypes
		 * value. For creating a specific sample of a feature, the
		 * sample constructor should be used.
		 *
		 * @param comp The distance metric that should be used.
		 * @param code A reference to a static variable (it must be
		 *             same for all created objects of _one_ feature)
		 *             for persistantly storing the code table.
		 * @param maxlen A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum string length.
		 */
		AbstractStringFeature(ComparatorType comp, stringcode *code, long *maxlen);

		/**
		 * Sample constructor
		 *
		 * This constructor initializes the feature with a specific
		 * value and should thus be used for samples.
		 *
		 * @param comp The distance metric that should be used.
		 * @param code A reference to a static variable (it must be
		 *             same for all created objects of _one_ feature)
		 *             for persistantly storing the code table.
		 * @param maxlen A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum string length.
		 * @param name The string representation of the sample value.
		 */
		AbstractStringFeature(ComparatorType comp, stringcode *code, long *maxlen, const string &name);

	public:
		virtual ~AbstractStringFeature();

		virtual double getPosition() const;
		virtual double getDistance(Feature *f) const;
		virtual void moveTowards(Feature *f, double factor);

		virtual string serialize() const;
		virtual void unserialize(string value);

		virtual featureparams write() const;
		virtual void read(featureparams *param);

		virtual Feature* clone() const;
		virtual void aggregate(aggregatelist samples);

		virtual FeatureType getType() const { return nominal; }
		virtual const string getName() const { 	return "AbstractStringFeature"; }

		/**
		 * Query the string values
		 *
		 * @return Values list
		 */
		const string& getVal() const { return name; }

		/// @TODO
		const unsigned long getCodeVal() const { return codeval; }

		/// @TODO
		const stringcode* getCode() { return code; }

#if _DEBUG_FEATURES
		/**
	     * Get feature as string
		 *
		 * @note This is only for testing.
		 * @return Feature as string.
		 */
		virtual string toString() const;
#endif

	private:
		/** The feature value */
		string name;

	protected:
		/** A reference to the static, persistant maximum length of strings seen so far. */
		long *maxlength;
		/** A reference to the static, persistant code table of strings seen so far. */
		stringcode *code;
		/** Only for internal usage in moveTowards and getDistance. */
		char *position;

		/** 
		 * The coded value of the feature
		 *
		 * @see name
		 * @see code
		 */
		unsigned long codeval;

		/**
	     * Get a code value for a given string
	     *
		 * A helper function to return the respective code for a given string.
		 * If the string is encountered the first time a new code is allocated.
		 *
		 * @param name The string to code
		 * @return A code value
		 */
		unsigned long getCodeForName(const string &name);

	private:
		/** The selected distance metric. */
		ComparatorType comparator;
};

/**
* Baseclass for features providing lists of string values
*
* A feature of type nominal which has lists of strings as values.
*/
class AbstractStringListFeature: public AbstractStringFeature
{
	private:
		/** The feature value */
		stringvector names;

	protected:
		/**
		 * The coded value of the feature.
		 *
		 * @see names
		 * @see code
		 */
		bit_vector listval;

		/**
		 * %Feature constructor
		 *
		 * This constructor initializes the feature as prototypes
		 * value. For creating a specific sample of a feature, the
		 * sample constructor should be used.
		 *
		 * @param code A reference to a static variable (it must be
		 *             same for all created objects of _one_ feature)
		 *             for persistantly storing the code table.
		 * @param maxlen A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum string length.
		 */
		AbstractStringListFeature(stringcode *code, long *maxlen);

		/**
		 * Sample constructor
		 *
		 * This constructor initializes the feature with a specific
		 * value and should thus be used for samples.
		 *
		 * @param code A reference to a static variable (it must be
		 *             same for all created objects of _one_ feature)
		 *             for persistantly storing the code table.
		 * @param maxlen A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum string length.
		 * @param names The string representation of the sample value.
		 */
		AbstractStringListFeature(stringcode *code, long *maxlen, const stringvector *names);

		/**
	     * Get a code value for a given stringvector
	     *
	     * The function composes a bit_vector wherein for each string
	     * in the stringvector the bit on the position of the strings
	     * code value is set to <var>true</var>. The remaining bits
	     * are initialized to <var>false</var>.
		 *
		 * @param names The stringvector to code
		 * @return A code value as bit_vector
		 */
		bit_vector getCodeForList(const stringvector *names);

	public:
		virtual string serialize() const;
		virtual void unserialize(string value);

		virtual double getPosition() const;
		virtual double getDistance(Feature *f) const;
		virtual void moveTowards(Feature *f, double factor);

		virtual Feature* clone() const;

		virtual const string getName() const { return "AbstractStringListFeature"; }

		/**
		 * Query the list of values
		 *
		 * @return Values list
		 */
		const bit_vector& getListVal() const {
			return listval;
		}

		/**
		 * Query the list of values
		 *
		 * @return Values list
		 */
		const stringvector& getListValues() const {
			return names;
		}

#if _DEBUG_FEATURES
		virtual string toString() const;
#endif
};

#endif // _ABSTRACTSTRING_H
