/**
 * @file
 * @brief Active window feature
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-02-12 21:08:39 +0100 (Don, 12 Feb 2004) $
 * $Revision: 612 $
 *
 * @since    Wed Apr 9 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: ActiveWindow.h 612 2004-02-12 20:08:39Z harald $
 */

#ifndef _ACTIVEWINDOWFEATURE_H
#define _ACTIVEWINDOWFEATURE_H

#include "../AbstractString.h"

/**
 * Returns the application with focus
 *
 * This feature tracks the application with focus on the users desktop and
 * returns the name active application. It is based on the AbstractStringFeature and uses the
 * Levenshtein distance.
 */
class FEATURE_EXPORT_CLASS ActiveWindowFeature : public AbstractStringFeature
{
	friend class ActiveWindowFeatureProvider;

	protected:
		/**
		 * %Feature constructor
		 *
		 * This constructor initializes the feature as prototypes
		 * value. For creating a specific sample of a feature, the
		 * sample constructor should be used.
		 *
		 * @param code A reference to a static variable (it must be
		 *             same for all created objects of _one_ feature)
		 *             for persistantly storing the code table.
		 * @param maxlen A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum string length.
		 */
		ActiveWindowFeature(stringcode* code, long* maxlen) : AbstractStringFeature(AbstractStringFeature::Levenshtein, code, maxlen) {}

		/**
		 * Sample constructor
		 *
		 * This constructor initializes the feature with a specific
		 * value and should thus be used for samples.
		 *
		 * @param code A reference to a static variable (it must be
		 *             same for all created objects of _one_ feature)
		 *             for persistantly storing the code table.
		 * @param maxlen A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum string length.
		 * @param window The string representation of the sample value.
		 */
		ActiveWindowFeature(stringcode* code, long* maxlen, string window) : AbstractStringFeature(AbstractStringFeature::Levenshtein, code, maxlen, window) {}

	public:
		virtual ~ActiveWindowFeature() {}

		virtual const string getName() const { return "ActiveWindow.ActiveWindow"; }

#if _DEBUG_FEATURES
		virtual string toString() const;
#endif
};

/**
 * Provider for the ActiveWindowFeature
 */
class FEATURE_EXPORT_CLASS ActiveWindowFeatureProvider : public FeatureProvider
{
	private:
		/** List of features */
		stringvector features;
		/** Map of code values */
		stringcode activeWindowFeature_code;
		/** Length of longest string */
		long activeWindowFeature_maxlen;

		/** Name of the X display */
		string display;
		/** Window name */
		string window;
		/** Indicator whether a sample is valid or not */
		bool windowValid;

	public:
		/**
		 * Constructor
		 *
		 * @param params Map of initialization parameters
		 */
		ActiveWindowFeatureProvider(providerparams &params);

		virtual Feature* getFeature(string name) const;
		virtual Feature* getSample(string name) const;
		virtual void nextSample(clock_t checkpoint);

#if _DEBUG_FEATURES
		string _test_getWindow() {
			return window;
		}
#endif
};

#endif // _ACTIVEWINDOWFEATURE_H
