/**
 * @file
 * @brief Audio feature
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-06-08 11:38:41 +0200 (Die, 08 Jun 2004) $
 * $Revision: 809 $
 *
 * @since    Wed Apr 9 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: Audio.cpp 809 2004-06-08 09:38:41Z rene $
 */

// for sprintf
#include <stdio.h>
#include "Audio.h"
#ifndef _WIN32_WCE
#include "../../thirdparty/transform/transform.h"
#endif

EXPORT_FEATURE(AudioFeatureProvider);

#if _DEBUG_FEATURES
#ifndef _WIN32_WCE
string AudioBandFeature::toString() const
{
	char buffer[64];
	sprintf(buffer, "AudioBandFeature: %.17f", getVal());
	return buffer;
}
#endif

string AudioMeanFeature::toString() const
{
	char buffer[64];
	sprintf(buffer, "AudioMeanFeature: %.17f", getVal());
	return buffer;
}

string AudioPeakFeature::toString() const
{
	char buffer[64];
	sprintf(buffer, "AudioPeakFeature: %ld", getVal());
	return buffer;
}
#endif

AudioFeatureProvider::AudioFeatureProvider(providerparams &params) : FeatureProvider("Audio", &features)
{
    PaError err;
    unsigned frames;
	frames = atol(params["frames"].c_str());
	bands = atol(params["bands"].c_str());

	if (frames <= 128) {
		plog ("Wrong framesize given, setting to default of 1024.\n");
		frames = 1024;
	}

	if (bands <= 0) {
		plog ("FFT disabled.\n");
	} else if (frames % bands) {
		frames = bands * (frames / bands);
		plog ("bands doesn't fit into frames, setting frames to %u.\n", frames);
	}

	for (unsigned i = 0; i < bands; i++) {
		char buf[64];
		sprintf(buf, "Band.%u", i);
		features.push_back(buf);
	}

	features.push_back("Mean");
	features.push_back("Peaks");

/// @todo symbian stl fix
#ifndef _SYMBIAN
	audioBandFeature_minval = numeric_limits<double>::infinity();
	audioBandFeature_maxval = -numeric_limits<double>::infinity();
	audioMeanFeature_minval = numeric_limits<double>::infinity();
	audioMeanFeature_maxval = -numeric_limits<double>::infinity();
	audioPeaksFeature_minval = numeric_limits<long>::max();
	audioPeaksFeature_maxval = numeric_limits<long>::min();
#else
	audioBandFeature_minval = KMaxReal;
	audioBandFeature_maxval = KMinReal;
	audioMeanFeature_minval = KMaxReal;
	audioMeanFeature_maxval = KMinReal;
	audioPeaksFeature_minval = KMaxTInt;
	audioPeaksFeature_maxval = KMinTInt;
#endif

	// init spectrum with the right amount of bands
	current = SampleData(bands);

	if (Pa_Initialize() != paNoError) {
		log_audio("Unable to initialize portaudio!\n");
		return;
	}

	if (params.count("device_id") == 0) {
		if ((device = Fixed_Pa_GetDefaultInputDeviceID()) == paNoDevice) {
			log_audio("Error opening audio device: no default device is available !\n");
			return;
		}
	} else {
		device = atoi(params["device_id"].c_str());
	}

	log_audio("Using audio device number %i\n", device);

	if ((err = Pa_OpenStream(
			&stream, 
			device, // input device
			2, // stereo
			PA_SAMPLE_TYPE, 
			NULL,
			paNoDevice, // output device
			0, // no output
			PA_SAMPLE_TYPE,
			NULL,
			SAMPLING_RATE,
			frames, // frames per buffer
			0, // number of buffers, if zero then use default minimum
			paClipOff, // we won't output out of range samples so don't bother clipping them
			audioCallback, 
			this)) != paNoError) {
		log_audio("Unable to open audio stream: %s\n", Pa_GetErrorText(err));
		return;
	}

	if (Pa_StartStream(stream) != paNoError) {
		log_audio("Unable to start audio stream!\n");
		return;
	}

	log_audio("Audio initialized successfully\n");
}

AudioFeatureProvider::~AudioFeatureProvider()
{
	plog("AudioFeatureProvider destructor called, terminating portaudion thread\n");
	if (Pa_StopStream(stream) != paNoError) {
		log_audio("Unable to stop audio stream!\n");
		return;
	}

	if (Pa_Terminate() != paNoError) {
		log_audio("Unable to close portaudio!\n");
		return;
	}

	log_audio("Audio closed successfully\n");
}

Feature* AudioFeatureProvider::getFeature(string name) const
{
	if (name.substr(0, 5) == "Band.") {
#ifndef _WIN32_WCE
		return new AudioBandFeature(const_cast<double*> (&audioBandFeature_minval), const_cast<double*> (&audioBandFeature_maxval));
#else	
		return NULL;
#endif
	} else if (name == "Mean") {
		return new AudioMeanFeature(const_cast<double*> (&audioMeanFeature_minval), const_cast<double*> (&audioMeanFeature_maxval));
	} else if (name == "Peaks") {
		return new AudioPeakFeature(const_cast<long*> (&audioPeaksFeature_minval), const_cast<long*> (&audioPeaksFeature_maxval));
	}

	return NULL;
}

Feature* AudioFeatureProvider::getSample(string name) const
{
	// always take a finished buffer for the features !
	if (name.substr(0, 5) == "Band.") {
#ifndef _WIN32_WCE
		unsigned band = atol(name.substr(5).c_str());
		if (band < bands) {
			return new AudioBandFeature(const_cast<double*> (&audioBandFeature_minval), const_cast<double*> (&audioBandFeature_maxval), processed.spectrum[band]);
		}
#else
		return NULL;
#endif
	} else if (name == "Mean") {
		return new AudioMeanFeature(const_cast<double*> (&audioMeanFeature_minval), const_cast<double*> (&audioMeanFeature_maxval), processed.average);
	} else if (name == "Peaks") {
		return new AudioPeakFeature(const_cast<long*> (&audioPeaksFeature_minval), const_cast<long*> (&audioPeaksFeature_maxval), processed.peaks);
	}

	return NULL;
}

void AudioFeatureProvider::nextSample(clock_t checkpoint)
{
	(void) checkpoint;	// to get rid of unused variable warnings

	// free unsused spectrum
	processed = current;					// copy processed data from current
	current = SampleData(bands);
}

int AudioFeatureProvider::audioCallback(void *inputBuffer, void *outputBuffer, unsigned long framesPerBuffer, PaTimestamp outTime, void *userData) {
	SAMPLE *in = (SAMPLE *) inputBuffer;
	AudioFeatureProvider *This = (AudioFeatureProvider *) userData;

	(void) outputBuffer;	// to get rid of unused variable warnings
	(void) outTime;	// to get rid of unused variable warnings

	log_audio("audioCallback called with %d frames\n", framesPerBuffer);

	/* FFT */
#ifndef _WIN32_WCE
	if (This->bands > 0) {
		WFastFT<SAMPLE> fft(This->bands);
		complex<double> *spectrum;

		for (unsigned long i = 0; i < (framesPerBuffer / This->bands); i++) {
			fft.setData(in + (i * This->bands));
			spectrum = fft.getSpectrum();

			This->current.spectrum_counter++;

			for (unsigned j = 0; j < This->bands; j++) {
				This->current.spectrum[j] = This->current.spectrum[j] * (1.0 - 1.0 / This->current.spectrum_counter) + spectrum[j].real() / This->current.spectrum_counter;
			}
		}
	}
#endif

	for (unsigned long j = 0; j < framesPerBuffer; j++) {
		This->current.average_counter++;
		This->current.average2_counter++;

		/* average samples */
		This->current.average = This->current.average * (1.0 - 1.0 / This->current.average_counter) + abs(in[j]) / This->current.average_counter;
		This->current.average2 = This->current.average2 * (1.0 - 1.0 / This->current.average2_counter) + abs(in[j]) / This->current.average2_counter;

		/* trigger peaks */
		if (This->current.average2_counter > (SAMPLING_RATE / 100)) {
			double diff = abs(This->current.average2 - This->current.average2_old);

			This->current.average2_old = This->current.average2;
			This->current.average2 = 0.0;
			This->current.average2_counter = 0;

			if (diff > This->current.treshold) {
				This->current.treshold *= 1.0 + ((diff - This->current.treshold) / This->current.treshold) / 100.0;
				if (This->current.peak) {
					This->current.peaks++;
					This->current.peak = false;
				}
			}

			if (diff < This->current.treshold) {
				This->current.treshold *= 1.0 - ((This->current.treshold - diff) / This->current.treshold) / 300.0;
				This->current.peak = true;
			}
		}
	}

	// we never finish from here
	return 0;
}
