/**
 * @file
 * @brief Audio feature
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-03-02 13:36:18 +0100 (Die, 02 Mär 2004) $
 * $Revision: 669 $
 *
 * @since    Wed Apr 9 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: Audio.h 669 2004-03-02 12:36:18Z rene-cvs $
 */

#ifndef _AUDIO_H
#define _AUDIO_H

#include "../Numerical.h"
#include "PaWrapper.h"

#ifndef _WIN32_WCE
/**
 * Represents a subband
 *
 * This feature represents a subband of the frequency spectrum. The calculation
 * is performed using the FFT algorithm. The feature has to be referenced as
 * "Audio.Band.[n]" where n is a number in the intervall [0; bands]. The maximum
 * number of bands is a global variable defined in the AudioFeatureProvider.
 */
class FEATURE_EXPORT_CLASS AudioBandFeature : public NumericalContinuousFeature
{
	friend class AudioFeatureProvider;

	protected:
		/**
		 * %Feature constructor
		 *
		 * This constructor initializes the feature with a random value
		 * and should thus only be used for creating e.g. prototypes
		 * of points in a clustering space. For creating a specific
		 * sample of a feature, the second constructor should be used.
		 *
		 * @param minval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the minimum value.
		 * @param maxval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum value.
		 */
		AudioBandFeature(double* minval, double* maxval) : NumericalContinuousFeature(minval, maxval) {}

		/** 
		 * Sample constructor
		 *
		 * This constructor initializes the feature with a specific
		 * value and should thus be used for samples.
		 *
		 * @param minval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the minimum value.
		 * @param maxval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum value.
		 * @param val The sample value.
		 */
		AudioBandFeature(double* minval, double* maxval, double val) : NumericalContinuousFeature(minval, maxval, val) {}

	public:
		virtual const string getName() const { return "Audio.Band"; }

#if _DEBUG_FEATURES
		virtual string toString() const;
#endif
};
#endif

/**
 * Calculates the mean level of the input signal
 *
 * This feature extracts the mean signal level out of the input signal.
 * After each call to nextSample the time window taken for calculating the
 * mean value is cleared.
 */
class FEATURE_EXPORT_CLASS AudioMeanFeature : public NumericalContinuousFeature
{
	friend class AudioFeatureProvider;

	protected:
		/**
		 * %Feature constructor
		 *
		 * This constructor initializes the feature with a random value
		 * and should thus only be used for creating e.g. prototypes
		 * of points in a clustering space. For creating a specific
		 * sample of a feature, the second constructor should be used.
		 *
		 * @param minval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the minimum value.
		 * @param maxval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum value.
		 */
		AudioMeanFeature(double* minval, double* maxval) : NumericalContinuousFeature(minval, maxval) {}

		/** 
		 * Sample constructor
		 *
		 * This constructor initializes the feature with a specific
		 * value and should thus be used for samples.
		 *
		 * @param minval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the minimum value.
		 * @param maxval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum value.
		 * @param mean The sample value.
		 */
		AudioMeanFeature(double* minval, double* maxval, double mean) : NumericalContinuousFeature(minval, maxval, mean) {}

	public:
		virtual const string getName() const { return "Audio.Mean"; }

#if _DEBUG_FEATURES
		virtual string toString() const;
#endif
};

/**
 * Counts the peaks in the input signal
 *
 * This feature counts the peaks in the input signal. If there are high
 * peaks the extractor becomes less sensible to lower peaks, if there
 * are only a few low peaks the extractor becomes more sensible to them.
 * When nextSample is called, the peak counter is reset to 0.
 */
class FEATURE_EXPORT_CLASS AudioPeakFeature : public NumericalDiscreteFeature 
{
	friend class AudioFeatureProvider;

	protected:
		/**
		 * %Feature constructor
		 *
		 * This constructor initializes the feature with a random value
		 * and should thus only be used for creating e.g. prototypes
		 * of points in a clustering space. For creating a specific
		 * sample of a feature, the second constructor should be used.
		 *
		 * @param minval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the minimum value.
		 * @param maxval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum value.
		 */
		AudioPeakFeature(long* minval, long* maxval) : NumericalDiscreteFeature(minval, maxval) {}

		/** 
		 * Sample constructor
		 *
		 * This constructor initializes the feature with a specific
		 * value and should thus be used for samples.
		 *
		 * @param minval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the minimum value.
		 * @param maxval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum value.
		 * @param peaks The sample value.
		 */
		AudioPeakFeature(long* minval, long* maxval, unsigned peaks) : NumericalDiscreteFeature(minval, maxval, peaks) {}

	public:
		virtual const string getName() const { return "Audio.Peaks"; }

#if _DEBUG_FEATURES
		virtual string toString() const;
#endif
};

/**
 * Provider for the audio features
 */
class FEATURE_EXPORT_CLASS AudioFeatureProvider : public FeatureProvider
{
	private:
		/**
		 * @todo documentation
		 */
		struct SampleData {
			//@{
			/** Sample property */

			double *spectrum;
			unsigned long spectrum_counter;

			double average;
			unsigned long average_counter;

			double average2;
			double average2_old;
			unsigned long average2_counter;

			double treshold;
			unsigned long peaks;
			bool peak;
			//@}

			/**
			 * Constructor
			 *
			 * @param bands Number of subbands
			 */
			SampleData(unsigned bands=0) {
				spectrum_counter = 0;
				if (bands != 0)
					spectrum = new double[bands];
				else
					spectrum = NULL;
				for (unsigned i = 0; i < bands; i++) {
					spectrum[i] = 0.0;
				}

				average_counter = 0;
				average = 0.0;

				average2_counter = 0;
				average2 = 0.0;
				average2_old = 0.0;

				treshold = 10000.0;		// init 10000.0
				peaks = 0;
				peak = true;
			}

			/** Destructor */
			~SampleData() {
				if (spectrum != NULL) {
					delete spectrum;
				}
			}
		};

		/** List of features */
		stringvector features;
		/** Number of subbands to calculate */
		unsigned bands;

		//@{
		/** Feature properties */

		double audioBandFeature_minval;
		double audioBandFeature_maxval;
		double audioMeanFeature_minval;
		double audioMeanFeature_maxval;
		long audioPeaksFeature_minval;
		long audioPeaksFeature_maxval;

		PaDeviceID device;
		PortAudioStream *stream;
		SampleData current;
		SampleData processed;
		//@}

		/** Portaudio callback function */
		static int audioCallback(void *inputBuffer, void *outputBuffer, unsigned long framesPerBuffer, PaTimestamp outTime, void *userData);

	public:
		AudioFeatureProvider(providerparams &params);
		virtual ~AudioFeatureProvider();

		virtual Feature* getFeature(string name) const;
		virtual Feature* getSample(string name) const;
		virtual void nextSample(clock_t checkpoint);
};

#endif // _AUDIO_H
