/**
 * @file
 * @brief Bluetooth feature
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-02-12 21:08:39 +0100 (Don, 12 Feb 2004) $
 * $Revision: 612 $
 *
 * @since    Tue Mar 18 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: Bluetooth.h 612 2004-02-12 20:08:39Z harald $
 */

#ifndef _BLUETOOTH_H
#define _BLUETOOTH_H

#include "../AbstractString.h"
#include "../Numerical.h"

/**
 * List of peers in range
 *
 * This feature returns a list of hardware addresses representing the Bluetooth peers
 * in reachable by the calling device.
 */
class FEATURE_EXPORT_CLASS BluetoothPeersFeature : public AbstractStringListFeature
{
	friend class BluetoothFeatureProvider;

	protected:
		/**
		 * %Feature constructor
		 *
		 * This constructor initializes the feature as prototypes
		 * value. For creating a specific sample of a feature, the
		 * sample constructor should be used.
		 *
		 * @param code A reference to a static variable (it must be
		 *             same for all created objects of _one_ feature)
		 *             for persistantly storing the code table.
		 * @param maxlen A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum string length.
		 */
		BluetoothPeersFeature(stringcode* code, long* maxlen) : AbstractStringListFeature(code, maxlen) {}

		/**
		 * Sample constructor
		 *
		 * This constructor initializes the feature with a specific
		 * value and should thus be used for samples.
		 *
		 * @param code A reference to a static variable (it must be
		 *             same for all created objects of _one_ feature)
		 *             for persistantly storing the code table.
		 * @param maxlen A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum string length.
		 * @param strList The string representation of the sample value.
		 */
		BluetoothPeersFeature(stringcode* code, long* maxlen, const stringvector* strList) : AbstractStringListFeature(code, maxlen, strList) {}

	public:
		virtual const string getName() const { return "Bluetooth.Peers"; }

#if _DEBUG_FEATURES
		virtual string toString() const;
#endif
};

/**
 * Number of peers in range
 *
 * This feature returns the number of Bluetooth peers reachable by the calling device.
 */
class FEATURE_EXPORT_CLASS BluetoothNumPeersFeature : public NumericalDiscreteFeature
{
	friend class BluetoothFeatureProvider;

	protected:
		/**
		 * %Feature constructor
		 *
		 * This constructor initializes the feature with a random value
		 * and should thus only be used for creating e.g. prototypes
		 * of points in a clustering space. For creating a specific
		 * sample of a feature, the second constructor should be used.
		 *
		 * @param minval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the minimum value.
		 * @param maxval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum value.
		 */
		BluetoothNumPeersFeature(long* minval, long* maxval) : NumericalDiscreteFeature(minval, maxval) {}

		/** 
		 * Sample constructor
		 *
		 * This constructor initializes the feature with a specific
		 * value and should thus be used for samples.
		 *
		 * @param minval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the minimum value.
		 * @param maxval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum value.
		 * @param numPeers The sample value.
		 */
		BluetoothNumPeersFeature(long* minval, long* maxval, unsigned numPeers) : NumericalDiscreteFeature(minval, maxval, numPeers) {}

	public:
		virtual const string getName() const { return "Bluetooth.NumPeers"; }

#if _DEBUG_FEATURES
		virtual string toString() const;
#endif
};

/**
 * Provider for the Bluetooth features
 */
class FEATURE_EXPORT_CLASS BluetoothFeatureProvider : public FeatureProvider
{
	private:
		/** List of features */
		stringvector features;

		//@{
		/** Feature properties
		@todo move to features
		*/

		stringcode peersFeature_code;
		long peersFeature_maxlen;
		long numPeersFeature_minval;
		long numPeersFeature_maxval;
		//@}

	protected:
		//@{
		/** Feature properties
		@todo move to features
		*/

		stringvector peerList;
		bool peerListValid;
		//@}

	public:
		/**
		 * Constructor
		 *
		 * @param params Map of initialization parameters
		 */
		BluetoothFeatureProvider(providerparams &params);
		/** Destructor */
		virtual ~BluetoothFeatureProvider();

		virtual Feature* getFeature(string name) const;
		virtual Feature* getSample(string name) const;
		virtual void nextSample(clock_t checkpoint);
};

#endif // _BLUETOOTH_H
