/**
 * @file
 * @brief Bluetooth feature
 *
 * Some parts of this code have been taken from hcitool.c of the
 * bluez-utils disribution and were modified accordingly. The original
 * license is:
 *
 *
 * BlueZ - Bluetooth protocol stack for Linux
 * Copyright (C) 2000-2001 Qualcomm Incorporated
 *
 * Written 2000,2001 by Maxim Krasnyansky <maxk@qualcomm.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) AND AUTHOR(S) BE LIABLE FOR ANY CLAIM,
 * OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * ALL LIABILITY, INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PATENTS, COPYRIGHTS,
 * TRADEMARKS OR OTHER RIGHTS, RELATING TO USE OF THIS SOFTWARE IS DISCLAIMED.
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-08-29 11:13:55 +0200 (Son, 29 Aug 2004) $
 * $Revision: 903 $
 *
 * @since    Tue Mar 18 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: BluetoothLinux.cpp 903 2004-08-29 09:13:55Z rene $
 */

extern "C" {
#include <errno.h>

// these are for the layer two connection handling
#include <unistd.h>
#include <sys/time.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>

// general Bluetooth API includes
#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/l2cap.h>

// these are for figuring out the own BT address
#include <sys/ioctl.h>
#include <sys/socket.h>
}

#include "BluetoothLinux.h"

/// Constants needed to perform inquiries on the Linux platform
//@{
#define BLUETOOTH_DEFAULT_DEVICE -1         ///< default device id of the Bluetooth adapter to use: -1 means the first available one
#define BLUETOOTH_SCANDELAY 5000            ///< milliseconds between inquiries / scans
#define BLUETOOTH_INQUIRYLENGTH 5           ///< the length of the Bluetooth inquiry in 1.28 second intervals
#define BLUETOOTH_FLUSH_CACHE true          ///< if the Bluetooth cache should be flushed between inquiries
#define BLUETOOTH_USE_L2_CONNECTIONS false  ///< if an L2 connection should be started to each found peer - this allows to measure the ping time and link quality
#define BLUETOOTH_L2_PINGDELAY 500          ///< milliseconds between L2 pings (if used)
#define BLUETOOTH_MAX_PING_TIME 60          ///< milliseconds L2PING time
#define BLUETOOTH_MIN_LINK_QUALITY 210      ///< minimum L2 link quality (range 0 -255)
//@}

EXPORT_FEATURE(BluetoothLinuxFeatureProvider);

void BluetoothFeatureProvider::nextSample(clock_t checkpoint)
{
	(void) checkpoint;	// to get rid of unused variable warnings
  // do nothing by now
}

BluetoothLinuxFeatureProvider::BluetoothLinuxFeatureProvider(featureparams& params) : BluetoothFeatureProvider(params)
{
	// first of all, the bluetooth adapter to use if there are multiple
	if (params.count("device") == 0) {
		// try the first available one
		device_id = hci_get_route(NULL);
		if (device_id < 0)
			device_id = BLUETOOTH_DEFAULT_DEVICE;
	}
	else {
		/// @todo check parameter !!
		const char* device_name = params["device"].c_str();
		device_id = hci_devid(device_name);
		if (device_id < 0) {
			plog("Error: Bluetooth device name unknown, using default device id !\n");
			device_id = BLUETOOTH_DEFAULT_DEVICE;
		}
	}

#if 1
	hci_dev_info dev_inf;
	if (hci_devinfo(device_id, &dev_inf) == 0) {
		bdaddr_t bdaddr;
		assert(device_id == -1 || device_id == dev_inf.dev_id);
		baswap(&bdaddr, &dev_inf.bdaddr);
		plog("Using Bluetooth device '%s' with id %d: MAC address %s\n",
			dev_inf.name, dev_inf.dev_id, batostr(&bdaddr));
	}
	else
		plog("Error: unable to get device information from device id %d\n", device_id);
#endif

	if (params.count("delay") == 0)
		delay = BLUETOOTH_SCANDELAY;
	else
		/// @todo check parameter !!
		delay = atoi(params["delay"].c_str());

	if (params.count("inquiry_length") == 0)
		inquiryLength = BLUETOOTH_INQUIRYLENGTH;
	else
		/// @todo check parameter !!
		inquiryLength = atoi(params["inquiry_length"].c_str());
	if (params.count("flush_cache") == 0)
		flushInquiryCache = BLUETOOTH_FLUSH_CACHE;
	else
		flushInquiryCache = (params["flush_cache"] == "true");

	// additional features
	if (params.count("use_l2_connections") == 0)
		use_l2_connections = BLUETOOTH_USE_L2_CONNECTIONS;
	else
		use_l2_connections = (params["use_l2_connections"] == "true");
	if (params.count("l2_pingdelay") == 0)
		l2_pingdelay = BLUETOOTH_L2_PINGDELAY;
	else
		/// @todo check parameter !!
		l2_pingdelay = atoi(params["l2_pingdelay"].c_str());
	if (params.count("maxpingtime") == 0)
		maxpingtime = BLUETOOTH_MAX_PING_TIME;
	else
		/// @todo check parameter !!
		maxpingtime = atoi(params["maxpingtime"].c_str());
	if (params.count("minlinkquality") == 0)
		minlinkquality = BLUETOOTH_MIN_LINK_QUALITY;
	else
		/// @todo check parameter !!
		minlinkquality = atoi(params["minlinkquality"].c_str());

	// don't allow the list to be read by now
	peerListValid = false;

	// start the scanning thread
	this->start();
}

BluetoothLinuxFeatureProvider::~BluetoothLinuxFeatureProvider() throw()
{
//    printf("BluetoothLinuxFeatureProvider destructor called\n");
	// stop the scanning thread and wait for it to finish
	this->cancel();
	this->join();

	this->BluetoothFeatureProvider::~BluetoothFeatureProvider();
}

void BluetoothLinuxFeatureProvider::run() throw()
{
	int num_rsp;

	inquiry_info *info = NULL;
	bdaddr_t bdaddr;

	while (!this->isCancelled()) {
		/* hci_inquiry will malloc() the info structure for the found number
		   of devices */
		num_rsp = hci_inquiry(device_id, inquiryLength, 0, NULL, &info, flushInquiryCache ? IREQ_CACHE_FLUSH : 0);
		if (num_rsp >= 0) {
			if (!use_l2_connections) {
				// keep the locked block to a minimum - the delay is caused by hci_inqiry
				Guard g(lockPeerList);
				peerList.clear();
			}

			for (int i = 0; i < num_rsp; i++) {
				char* macStr = NULL;
				baswap(&bdaddr, &(info+i)->bdaddr);
				macStr = batostr(&bdaddr);
				if (macStr != NULL) {
					if (!use_l2_connections) {
						Guard g(lockPeerList);
						peerList.push_back(string(macStr));
					}
					else {
						// check if there is already a thread running for this peer
						string mac(macStr);
						if (find(l2connections.begin(), l2connections.end(), mac) == l2connections.end()) {
							// no thread, start one
							l2connections.push_back(mac);
							new L2Connection(this, mac);
						}
						// the MAC addresses will be inserted into peerlist by the threads
					}

					free(macStr);
				}
			}
			peerListValid = true;

			if (info != NULL)
				free(info);
			// we have to set it to 0 so that hci_inquire will again malloc !
			info = NULL;
		}

		// HACK, HACK, HACK
		unsigned int sleep, num;
		if (delay > 1000) {
			num = delay/100;
			sleep = 100;
		}
		else {
			num = 1;
			sleep = delay;
		}
		for (unsigned int i=0; i<num && !this->isCancelled(); i++)
			this->sleep(sleep);
	}
}

Feature* BluetoothLinuxFeatureProvider::getSample(string name) const
{
	Guard g(lockPeerList);
	return BluetoothFeatureProvider::getSample(name);
}


/// a helper function for caclulating the time difference
static float tv2fl(struct timeval tv)
{
	return (float)(tv.tv_sec*1000.0) + (float)(tv.tv_usec/1000.0);
}

/// a helper function copied from hcitool.c
static int find_conn(int s, int dev_id, long arg)
{
	struct hci_conn_list_req *cl;
	struct hci_conn_info *ci;
	int i;

	if (!(cl = (hci_conn_list_req*) malloc(10 * sizeof(*ci) + sizeof(*cl)))) {
		perror("Can't allocate memory");
		exit(1);
	}
	cl->dev_id = dev_id;
	cl->conn_num = 10;
	ci = cl->conn_info;

	if (ioctl(s, HCIGETCONNLIST, (void*)cl)) {
		perror("Can't get connection list");
		exit(1);
	}

	for (i=0; i < cl->conn_num; i++, ci++)
		if (!bacmp((bdaddr_t *)arg, &ci->bdaddr))
			return 1;
	return 0;
}


BluetoothLinuxFeatureProvider::L2Connection::L2Connection(BluetoothLinuxFeatureProvider* prov, string mac) :
		prov(prov), mac(mac) {
	plog("----------------------------------------------------------------------------------------\n");
	plog("Creating new L2Connection object for MAC %s, starting ping thread now\n", mac.c_str());
	this->start();
}

void BluetoothLinuxFeatureProvider::L2Connection::run() throw() {
	bool running = true, started = false;
	bdaddr_t bdaddr;
	int id;
	struct sockaddr_l2 addr;
	char buf[128];
	struct hci_conn_info_req *cr;
	struct hci_request rq;
	get_link_quality_rp rp;

	baswap(&bdaddr, strtoba(mac.c_str()));

	// intialize ping packet buffer
	for(unsigned int i = L2CAP_CMD_HDR_SIZE; i < sizeof(buf); i++)
		buf[i]=(i%40)+'A';

	memset(&addr, 0, sizeof(addr));
	addr.l2_family = AF_BLUETOOTH;

	// also prepare the hci structures for querying the link quality
	cr = (hci_conn_info_req*) malloc(sizeof(*cr) + sizeof(struct hci_conn_info));
	if (!cr)
		running = false;
	bacpy(&cr->bdaddr, &bdaddr);
	cr->type = ACL_LINK;

	memset(&rq, 0, sizeof(rq));
	rq.ogf = OGF_STATUS_PARAM;
	rq.ocf = OCF_GET_LINK_QUALITY;
	rq.clen = 2;
	rq.rparam = &rp;
	rq.rlen = GET_LINK_QUALITY_RP_SIZE;

	id = ident;
	while (running) {
		struct timeval tv_send, tv_recv, tv_diff;
		l2cap_cmd_hdr *cmd = (l2cap_cmd_hdr *) buf;
		int s, dd, dev_id;

		// first we need to create a Bluetooth socket
		if ((s = socket(PF_BLUETOOTH, SOCK_RAW, BTPROTO_L2CAP)) < 0) {
			plog("BluetoothLinuxFeatureProvider::L2Connection::run: Can't create socket: %s\n", strerror(errno));
			running = false;
			break;
		}
		plog("BluetoothLinuxFeatureProvider::L2Connection::run (%s): opened socket %d\n", mac.c_str(), s);

		bacpy(&(addr.l2_bdaddr), BDADDR_ANY);
        	if (bind(s, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
			plog("BluetoothLinuxFeatureProvider::L2Connection::run: Can't bind socket.\n");
			running = false;
			close(s);
			break;
        	}

		// connect to the remote host
		bacpy(&addr.l2_bdaddr, &bdaddr);
		if (connect(s, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
			plog("BluetoothLinuxFeatureProvider::L2Connection::run: Can't connect: %s (%d)\n", strerror(errno), errno);
			running = false;
			close(s);
			break;
		}

		/* Build command header */
		cmd->code  = L2CAP_ECHO_REQ;
		cmd->ident = id;
		// do we need this ??
		//cmd->len   = __cpu_to_le16(size);
		cmd->len = size;

		gettimeofday(&tv_send, NULL);

		/* Send Echo Request */
		if( send(s, buf, size + L2CAP_CMD_HDR_SIZE, 0) <= 0 ){
			plog("BluetoothLinuxFeatureProvider::L2Connection::run: Send failed\n");
			running = false;
			close(s);
			break;
		}

		/* Wait for Echo Response */
		bool lost = false;
		while( 1 ) {
			struct pollfd pf[1];
			register int err;

			pf[0].fd = s; pf[0].events = POLLIN;
			// wait for a maximum of 1 second for a reply
			if( (err = poll(pf, 1, 1000)) < 0 ) {
				plog("BluetoothLinuxFeatureProvider::L2Connection::run: Poll failed\n");
				running = false;
				lost = true;
				break;
			}

			if( !err ){
				lost = true;
				break;
			}

			if( (err = recv(s, buf, sizeof(buf), 0)) < 0 ) {
				plog("BluetoothLinuxFeatureProvider::L2Connection::run: Recv failed\n");
				running = false;
				lost = true;
				break;
			}

			if( !err ){
				printf("BluetoothLinuxFeatureProvider::L2Connection::run: Disconnected\n");
				running = false;
				lost = true;
				break;
			}

			// again, do we need this ???
			//cmd->len = __le16_to_cpu(cmd->len);

			/* Check for our id */
			if( cmd->ident != id )
				continue;

			/* Check type */
			if( cmd->code == L2CAP_ECHO_RSP )
				break;
			if( cmd->code == L2CAP_COMMAND_REJ ){
				printf("BluetoothLinuxFeatureProvider::L2Connection::run: Peer doesn't support Echo packets\n");
				running = false;
				lost = true;
				break;
			}

		}

		if( lost) {
			printf("no response from %s: id %d\n", mac.c_str(), id);
			running = false;
			break;
		}

		gettimeofday(&tv_recv, NULL);
		timersub(&tv_recv, &tv_send, &tv_diff);

		// also open the Bluetooth HCI device to query the link quality
		dev_id = hci_for_each_dev(HCI_UP, find_conn, (long) &bdaddr);
		if (dev_id < 0) {
			plog("BluetoothLinuxFeatureProvider::L2Connection::run: Not connected. This really shouldn't happen - we got a reply\n");
			running = false;
			break;
		}

		dd = hci_open_dev(dev_id);
		if (dd < 0) {
			plog("BluetoothLinuxFeatureProvider::L2Connection::run: HCI device open failed\n");
			running = false;
			break;
		}

		if (ioctl(dd, HCIGETCONNINFO, (unsigned long) cr) < 0) {
			plog("BluetoothLinuxFeatureProvider::L2Connection::run: Get connection info failed\n");
			running = false;
			close(dd);
			break;
		}

		rq.cparam = &cr->conn_info->handle;

		if (hci_send_req(dd, &rq, 100) < 0) {
			plog("BluetoothLinuxFeatureProvider::L2Connection::run: HCI get_link_quality request failed\n");
			running = false;
			close(dd);
			break;
		}

		if (rp.status) {
			plog("BluetoothLinuxFeatureProvider::L2Connection::run: HCI get_link_quality cmd failed (0x%2.2X)\n",
					rp.status);
			running = false;
			close(dd);
			break;
		}
		// and close the Bluetooth L2 connection again so that the device is freed
		hci_disconnect(dd, cr->conn_info->handle, 0x13, 100);
		close(s);
		close(dd);

		if( ++id > 254 ) id = ident;


/*		printf("%d bytes from %s id %d time %.2fms\n", cmd->len, mac.c_str(), id, tv2fl(tv_diff));
		printf("Link quality: %d\n", rp.link_quality);*/
		if (tv2fl(tv_diff) <= prov->maxpingtime && rp.link_quality >= prov->minlinkquality) {
			// yes, we have a good link
			if (!started) {
				// this is the first ping reply --> add the MAC to the peer list
				Guard g(prov->lockPeerList);
				prov->peerList.push_back(mac);
			}
			started = true;

			this->sleep(prov->l2_pingdelay);
		}
		else {
			// no, link is too bad (but we still have a connection, so continue with pinging but only remove from the list)
			if (started) {
				Guard g(prov->lockPeerList);
				stringvector::iterator iter;
				if ((iter=find(prov->peerList.begin(), prov->peerList.end(), mac)) != prov->peerList.end())
					prov->peerList.erase(iter);
			}
			started = false;
		}
	}

	// lost the connection --> remove the MAC from the peer list
	{
		Guard g(prov->lockPeerList);
		stringvector::iterator iter;
		if ((iter=find(prov->peerList.begin(), prov->peerList.end(), mac)) != prov->peerList.end())
			prov->peerList.erase(iter);
	}

	free(cr);

	plog("Deleting L2Connection object for MAC %s, ping thread stopped\n", mac.c_str());

	// the last thing to do
	stringvector::iterator iter;
	if ((iter=find(prov->l2connections.begin(), prov->l2connections.end(), mac)) != prov->l2connections.end())
		prov->l2connections.erase(iter);
	delete this;
}
