/**
 * @file
 * @brief %Feature interface declaration
 *
 * This file contains the interfaces that have to be implemented
 * to enable a sensor to be used by the framework.
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-02-12 21:08:39 +0100 (Don, 12 Feb 2004) $
 * $Revision: 612 $
 *
 * @since    Sun Mar 30 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: Feature.h 612 2004-02-12 20:08:39Z harald $
 */

#ifndef _FEATURE_H
#define _FEATURE_H

#include "../util/defines.h"

#include <time.h>
#include <assert.h>
#ifndef _SYMBIAN
#endif

//@{
/** Compiler specific flags to export functions and classes in shared libraries */

#if defined(_WINDOWS) || defined(_WIN32_WCE)
# ifdef _FEATURE_EXPORT
#  define FEATURE_EXPORT __declspec(dllexport)
#  ifdef _FEATURE_EXPORT_CLASS
#   define FEATURE_EXPORT_CLASS __declspec(dllexport)
#  endif // _FEATURE_EXPORT_CLASS
# else // _FEATURE_EXPORT
#  define FEATURE_EXPORT __declspec(dllimport)
#  ifdef _FEATURE_EXPORT_CLASS
#   define FEATURE_EXPORT_CLASS __declspec(dllimport)
#  endif // _FEATURE_EXPORT_CLASS
# endif // _FEATURE_EXPORT
#elif defined(_LINUX) // _WINDOWS or _WIN32_WCE
# define FEATURE_EXPORT
# define FEATURE_EXPORT_CLASS
#elif defined(_SYMBIAN)
# ifdef _FEATURE_EXPORT
#  define FEATURE_EXPORT EXPORT_C
#  ifdef _FEATURE_EXPORT_CLASS
#   define FEATURE_EXPORT_CLASS EXPORT_C
#  endif // _FEATURE_EXPORT_CLASS
# else // _FEATURE_EXPORT
#  define FEATURE_EXPORT IMPORT_C
#  ifdef _FEATURE_EXPORT_CLASS
#   define FEATURE_EXPORT_CLASS IMPORT_C
#  endif // _FEATURE_EXPORT_CLASS
# endif // _FEATURE_EXPORT
#endif // _WINDOWS or _WIN32_WCE

#ifndef FEATURE_EXPORT_CLASS
# define FEATURE_EXPORT_CLASS
#endif
//@}

#ifdef _SYMBIAN
// for Symbian OS, static variables are not supported in DLLs, we have to use Thread Local Storage for that
/*
#define EXPORT_FEATURE(provider) \
	void INIT_EXPORT library_initialize() { Dll::SetTls(NULL); } \
	void FINI_EXPORT library_finalize() { if (Dll::Tls() != NULL) delete (provider*) (Dll::Tls()); } \
	extern "C" FEATURE_EXPORT FeatureProvider* getProvider(providerparams &params) { if (Dll::Tls() == NULL) Dll::SetTls((TAny*) new provider(params)); return (provider*) Dll::Tls(); }*/
// there is no support for loading DLLs by now anyways
#define EXPORT_FEATURE(provider)
#else // _SYMBIAN
/**
 * Export a feature provider
 *
 * This macro has to be used in global scope and gets substituted with the
 * implementation for exporting a FeatureProvider singleton instance to the
 * framework.
 *
 * @param provider The Class that implements the FeatureProvider interface
 *                 and should be exported to the framework.
 */
#define EXPORT_FEATURE(provider) \
	/** Singleton implementing the FeatureProvider interface */ \
	static provider *fp; \
	void INIT_EXPORT library_initialize() { fp = NULL; } \
	void FINI_EXPORT library_finalize() { if (fp != NULL) delete fp; } \
	/** \
     * Returns an instance of a class implementing the FeatureProvider interface \
	 * \
	 * @param params A map of parameters to initialize the instance \
	 * @return A FeatureProvider implementation \
	 */ \
	extern "C" FEATURE_EXPORT FeatureProvider* getProvider(providerparams &params) { if (fp == NULL) fp = new provider(params); return fp; }
#endif // _SYMBIAN

class Feature;

/**
 * Parameters passed to feature providers
 *
 * When initializing a feature provider, an arbitrary number of named (string)
 * parameters can be passed from the application (e.g. read from a config file).
 * These key-value tuples can then be used by the feature provider as
 * configuration or initialization values. They are completely FeatureProvider
 * specific.
 *
 * The same structure is also used to store and read persistant feature data.
 *
 * @see FeatureProvider
 * @see PersistantFeature
 */
typedef parametermap providerparams;

/** @todo documentation */
typedef parametermap featureparams;

/** A <timestamp, Feature> tuple */
typedef pair<time_t, Feature*> aggregatefeature;

/** A list of <timestamp, Feature> tuples */
typedef list<aggregatefeature> aggregatelist;

/** A map type for returning membership values for each cluster. */
typedef map<unsigned long, double> membershiplist;

/**
 * @name DLL initialization functions
 *
 * These functions will be called when a library is loaded
 * or unloaded by the framework.
 */
//@{    
extern "C" {
	/** Load library */
	void library_initialize();
	/** Unload library */
	void library_finalize();
}
//@}

/**
 * Feature interface
 *
 * A feature represents a single dimension in the cluster space and
 * implements the necessary funcionality to compare features of the
 * same kind and measure distances between them.
 * A feature is identified by a name which has to be unique in the
 * scope of its feature provider. It abstracts the type and structure of the
 * underlying sensor data and can also apply arbitrary data processing and
 * transformation to the raw sensor data. Typically, a sensor (i.e. a
 * FeatureProvider) will provide multiple features which are extracted using
 * domain-specific knowledge.
 *
 * A feature will also be referenced as sample. Samples are a more generic
 * from of a feature only containing information about their position and
 * without information about corresponding neighbours, age, error, etc.
 *
 * @see FeatureProvider
 */
class Feature {
	public:
		/** Possible feature types */
		typedef enum { boolean, nominal, ordinal, numerical_discrete, numerical_continuous } FeatureType;

	protected:
		/**
		 * Externalization flag
		 *
		 * If true, then this feature has persistent data which should
		 * be preserved across restarts of the application (e.g. a
		 * list of already seen feature values for nominal and ordinal
		 * types or maximum and minimum values for numerical types).
		 *
		 * For all persistant features, the object should be cast to
		 * PersistentFeature, because only subclasses of this type are
		 * (by policy) allowed to set this variable to true.
		 *
		 * @see PersistantFeature
		 */
		const bool externalize;

		/**
		 * Constructor for usage in PerstantFeature
		 *
		 * Sets the externalization flag to the given value.
		 *
		 * @see extnlze
		 */
		Feature(bool ext) : externalize(ext) {}

		/**
		 * Default constructor
		 *
		 * Sets the externalization flag to <var>false</var>.
		 *
		 * @see extnlze
		 */
		Feature() : externalize(false) {}

	public:
		/**
	     * Destructor
	     *
		 * Empty virtual destructor to have an entry in the implementing
		 * classes method table. This ensures that derived classes get
		 * destructed correctly.
		 */
		virtual ~Feature() {}

		/**
		 * Query a features type
		 *
		 * @return The type of the Feature.
		 */
		virtual FeatureType getType() const = 0;

		/**
		 * Query a features name
		 *
		 * @return Name of the Feature in the format "Featureprovider.Feature"
		 */
		virtual const string getName() const = 0;

		/**
		 * Query a features position
		 *
		 * @return The distance to the origin. Every implementation must take
		 *         care that only values in the interval <var>[0; 1]</var> are returned.
		 *
		 * @remarks This function is only for the purpose of internal
		 *          search operations and visualisation and should
		 *          not be used in any other context.
		 */
		virtual double getPosition() const = 0;

		/**
		 * Query externalization flag
		 *
		 * @return <var>true</var> if the feature is externalizeable (
		 *         i.e. it has persistend data that should be stored
		 *         across restarts), <var>false</var> otherwise.
		 */
		bool isExternalizable() { return externalize; }

		/**
		 * Calculates the distance between two features
		 *
		 * @param f Feature used for distance measurement.
		 * @return Distance between the two samples. The return
		 *         value has to be in the interval <var>[0; 1]</var>
		 *         to guarantee comparability among samples of
		 *         different features.
		 */
		virtual double getDistance(Feature *f) const = 0;

		/**
		 * Move feature
		 *
		 * Moves the feature towards a given sample value by
		 * the specified <var>factor</var> (<var>factor</var> times the distance).
		 * If the <var>factor</var> is <var>0</var>, the old feature value must
		 * not change.
		 * If the <var>factor</var> is <var>1</var>, the new feature value should
		 * be equivalent to the sample s.
		 *
		 * @param f Feature used for distance measurement.
		 * @param factor Distance weight.
		 */
		virtual void moveTowards(Feature *f, double factor) = 0;

		/**
		 * Clone a feature
		 *
		 * A feature must be able to clone itself. This is typically implemented
		 * with a copy constructor, but should return a new object.
		 */
		virtual Feature* clone() const = 0;

		/**
		 * Serialize a samples data to a string
		 *
		 * @return String representation of the samples data.
		 */
		virtual string serialize() const = 0;

		/**
		 * Unserialize a samples data from a string
		 *
		 * @param value String representation of the samples data. 
		 */
		virtual void unserialize(string value) = 0;

		/**
		 * Aggregate a sample values from other sources
		 *
		 * @param samples A list of <timestamp, sample> tuples.
		 */
		virtual void aggregate(aggregatelist samples) = 0;

#if _DEBUG_FEATURES
		/** This is only for testing. */
		virtual string toString() const = 0;
#endif
};

/**
 * PersistantFeature interface
 *
 * This interface only enhances the Feature interface by methods to handle
 * persistant data of a feature. It sets the externalization flag to <var>true</var>.
 *
 * @see externalize
 */
class PersistantFeature : public Feature {
	private:
		/** 
		 * Validation flag
		 *
		 * <var>true</var> when the persitant feature data is valid (i.e.
		 * it has not been modified since the last call to write),
		 * <var>false</var> if it has been modified and needs to be written.
		 */
		bool valid;

	protected:
		/**
		 * Default constructor
		 *
		 * Sets the externalization flag and the validation flag
		 * to <var>true</var>.
		 *
		 * @see externalize
		 * @see valid
		 */
		PersistantFeature() : Feature(true) { valid = true; }

	public:
		/**
		 * Externalize feature
		 *
		 * @returns Persistant feature data.
		 * @see read
		 */
		virtual featureparams write() const = 0;
		
		/**
		 * Load feature from persistant data
		 *
		 * Initializes the persistant feature data from the given representation.
		 *
		 * @param param Persistant feature data.
		 * @see write
		 */
		virtual void read(featureparams *param) = 0;

		/**
	     * Invalidate feature
	     *
	     * Sets the validation flag to <var>flase</var> to signal the
	     * framework that the corresponding persistant feature data
	     * is not up to date and has to be written again.
	     */
		void invalidate() { valid = false; }
		
		/** Set the validation flag to <var>true</var> */
		void validate() { valid = true; }
		
		/**
		 * Query validation flag
		 *
		 * @return <var>true</var> if the features persistant data is up to
		 *         date, <var>false</var> otherwise.
		 */
		bool isValid() { return valid; }
};

/**
 * Ordered list of features
 *
 * This is one of the basic datatypes in this framework. It represents an
 * ordered list of features, which describes a single point in the
 * heterogeneous feature space.
 *
 * @note The elements of this vector are pointers to feature objects.
 *       Therefore, the memory management of its members is not defined
 *       globally but depends on the method which returns or creates a
 *       featurevector.
 */
typedef vector<Feature*> featurevector;

/**
 * FeatureProvider interface
 *
 * Each feature that is used in the cluster space definition is provided
 * by a feature provider. A feature provider can provide more than one
 * feature and will be refered to by a unique name.
 * A feature can then be unambiguously identified by a combination of
 * its own name and the name of its feature provider.
 * Thus the actual cluster space can be defined in a configuration file
 * and the clustering infrastructure can fetch the according features
 * based on that definition.
 *
 * @see Feature
 * @see PersistantFeature
 */
class FeatureProvider {
	private:
		/** The list of feature names provided */
		stringvector *features;
		/** The name of the feature provider itself */
		string name;

	protected:
		/**
		 * Constructor
		 *
		 * Initializes the names of the provider and its features.
		 *
		 * @param name Name of the provider.
		 * @param features List of featurenames provided.
		 */
		FeatureProvider(string name, stringvector *features) {
			this->name = name;
			this->features = features;
		}
		
	public:
		/**
	     * Destructor
	     *
		 * Empty virtual destructor to have an entry in the implementing
		 * classes method table. This ensures that derived classes get
		 * destructed correctly.
		 */
		virtual ~FeatureProvider() {}

		/**
		 * Query feature instance
		 *
		 * Query the feature provider for a specific feature
		 * initialized with random values.
		 *
		 * @param name Name of the requested feature.
		 * @return An instance of the requested feature.
		 */
		virtual Feature* getFeature(string name) const = 0;

		/**
		 * Query sample instance
		 *
		 * Query the feature provider for a sample value
		 * of the specified feature.
		 * Prior to this method you have to invoke nextSample().
		 *
		 * @param name Name of the requested feature.
		 * @return An instance of the sample.
		 *
		 * @see nextSample
		 */
		virtual Feature* getSample(string name) const = 0;

		/**
		 * Prepare sample
		 *
		 * Prepare the feature provider to return the next sample from
		 * the sensor when getSample() is called.
		 *
		 * @param checkpoint This parameter tells the feature extractor
		 *                   that it must not return sensordata sampled
		 *                   earlier than the given clock value.
		 * @see getSample
		 */
		virtual void nextSample(clock_t checkpoint) = 0;

		/**
		 * Query provider name
		 *
		 * @return A global unique name for the feature provider.
		 */
		string getName() const
		{
			return name;
		}

		/**
		 * Query list of provided features
		 *
		 * @return A list of feature names provided by that feature provider.
		 */
		stringvector* getFeatureList() const
		{
			return features;
		}
};

/**
 * Type of the getProvider function
 *
 * This type definition is needed by the FeatureContainer to query a
 * Feature implementation for an instance of its class.
 */
typedef FeatureProvider* (*getProvider_t) (const providerparams &params);

#endif // _FEATURE_H
