/**
 * @file
 * @brief GSM feature
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-03-02 13:36:18 +0100 (Die, 02 Mär 2004) $
 * $Revision: 669 $
 *
 * @since    Mon May 12 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: GSM.cpp 669 2004-03-02 12:36:18Z rene-cvs $
 */

// for sprintg
#include <stdio.h>
#include "GSM.h"
#include <algorithm>
#ifndef _WIN32_WCE
#include <cctype>
#endif

EXPORT_FEATURE(GSMFeatureProvider)

//@{
/** AT string commands for retrieving the GSM cell id */
static string COMMAND_INIT("atz");
static string COMMAND_CREG("at+creg=2");
static string ANSWER_OK("OK");
static string ANSWER_CREG("+CREG: 2,1,");
//@}

/**
 * toupper implementations for string class
 *
 * @param s Lowercase string
 * @return Uppercase string
 */
static inline string toupperstr(string s)
{
        std::transform(s.begin(), s.end(), s.begin(), toupper);
	return s;
}

#if _DEBUG_FEATURES
string GSMCellFeature::toString() const
{
	char buffer[256];
	sprintf(buffer, "GSMCellFeature: '%s', has code %s", getName().c_str(), serialize().c_str());
	return buffer;
}
#endif


GSMFeatureProvider::GSMFeatureProvider(providerparams& params) : FeatureProvider("GSM", &features)
{
	if (params.count("delay") == 0) {
		this->delay = GSM_SCANDELAY;
	} else {
		/// @todo check parameter !!
		this->delay = atoi(params["delay"].c_str());
	}

	this->cellIdValid = false;

	features.push_back("CellID");

	gsmCellFeature_maxlen = 0;

	/* The thread can not be started in this constructor because then the virtual
	   tables are not initialized when the thread starts and thus the overloaded
	   methods (which are pure virtual in this base class) can't be called, resulting
	   in an abort. Start in the base classes */

	init(params);

	// start me
	this->start();
}

GSMFeatureProvider::~GSMFeatureProvider()
{
	plog("GSMFeatureProvider destructor called, waiting for thread to terminate\n");
	// signal the thread to terminate and wait for it
	this->cancel();
	this->join();
}

Feature* GSMFeatureProvider::getSample(string name) const
{
	Feature *ret;
	Guard g(lockCellId);
	
	if (name == "CellID" && cellIdValid) {
		ret = new GSMCellFeature(const_cast<stringcode*> (&gsmCellFeature_code), const_cast<long*> (&gsmCellFeature_maxlen), cellId);
		return ret;
	}

	return NULL;
}

Feature* GSMFeatureProvider::getFeature(string name) const
{
	if (name == "CellID") {
		return new GSMCellFeature(const_cast<stringcode*> (&gsmCellFeature_code), const_cast<long*> (&gsmCellFeature_maxlen));
	}

	return NULL;
}

void GSMFeatureProvider::nextSample(clock_t checkpoint)
{
	(void) checkpoint;	// to get rid of unused variable warnings
  // do nothing by now
}

void GSMFeatureProvider::run()
{
	string result, result2, loc, cell;

	while (! this->isCancelled()) {
		if (openSerialPort(serialPort))
		{
			if (! writeLine("atz\r\n") ||
				readLine () != COMMAND_INIT || 
				toupperstr(readLine()) != ANSWER_OK) 
			{ // guard block
				plog("Error: Phone not responding to ATZ command\n");

				Guard g(lockCellId);
				cellIdValid = false;
			} // guard block
			else { 
				if (! writeLine("at+creg=2\r\n") || 
					readLine () != COMMAND_CREG || 
					toupperstr(readLine()) != ANSWER_OK) 
				{ // guard block
					Guard g(lockCellId);
					cellIdValid = false;

					plog("Error: Phone not responding to AT+CREG=2 command\n");
				} // guard block
				else {
					writeLine("at+creg?\r\n");
					readLine();
					result = readLine();
					readLine();
					result2 = readLine();
					// check if we understand the result and if it's expected
					if (result.find(ANSWER_CREG) == 0) 
					{ // guard block
						Guard g(lockCellId);

						// yes, now extract the cell ID
						// the first field is the location ID
						loc = result.substr(12, 4);
						// the second is the cell ID
						cell = result.substr(19, 4);
						// for the returned cell ID string, we use both
						cellId = cell + loc;

						if (result2 == ANSWER_OK) 
							cellIdValid = true;
						else {
							cellIdValid = false;

							plog("Error: no OK in response to AT+CREG? command\n");
						}
					} // guard block
				}
			}
			closeSerialPort();
		}
		else
		{ // guard block
			Guard g(lockCellId);
			cellIdValid = false;
		} // guard block
		for (int i=0; i<delay*10 && !this->isCancelled(); i++)
			this->sleep(100);
	}
}
