/**
 * @file
 * @brief GSM feature
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-02-12 21:08:39 +0100 (Don, 12 Feb 2004) $
 * $Revision: 612 $
 *
 * @since    Mon May 12 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: GSMLinux.cpp 612 2004-02-12 20:08:39Z harald $
 */

#include "GSM.h"

#include <fcntl.h>
#include <termios.h>
#include <unistd.h>

/** Default serial port to use */
#define GSM_SERIALPORT "/dev/rfcomm0"

/** Serial port handle */
static int portFd;
/** Serial port attributes */
static struct termios term_old;

void GSMFeatureProvider::init(featureparams& params)
{
	if (params.count("serialport") == 0) {
		this->serialPort = GSM_SERIALPORT;
	} else {
		/// @todo check parameter !!
		this->serialPort = params["serialport"];
	}
}

bool GSMFeatureProvider::openSerialPort(string serialport)
{
	struct termios term_tmp;

	portFd = open(serialport.c_str(), O_RDWR);
	if (portFd < 0) {
		plog("Error: unable to open serial port %s\n", serialport.c_str());
		return false;
	}
	if (!isatty(portFd)) {
		plog("Error: the device is not a serial port: %s\n", serialport.c_str());
		return false;
	}

 	if (tcgetattr(portFd, &term_old) != 0) {
		plog("Error: unable to get attributes from port\n");
		return false;
	}

	 // now set the baud rate , data bits, stop bits and parity
	term_tmp = term_old;
	cfsetospeed(&term_tmp, B115200);
	cfsetispeed(&term_tmp, B115200);
	term_tmp.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
	term_tmp.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP);
	term_tmp.c_iflag |= (IXON | IXOFF | IGNCR);
	term_tmp.c_cflag &= ~(CSIZE | PARENB | CSTOPB);
	term_tmp.c_cflag |= (CS8 | CLOCAL);
	term_tmp.c_oflag &= ~(OPOST);
	term_tmp.c_cc[VMIN] = 1;
	term_tmp.c_cc[VTIME] = 1;
	if (tcsetattr(portFd, TCSANOW, &term_tmp) != 0) {
		plog("Error: unable to set port attributes\n");
		return false;
	}

	return true;
}

void GSMFeatureProvider::closeSerialPort()
{
	if (portFd >= 0)
	{
		tcsetattr(portFd, TCSANOW, &term_old);
		close(portFd);
	}
}

bool GSMFeatureProvider::writeLine(string line)
{
	if (portFd >= 0)
	{
		if (write(portFd, line.c_str(), line.length()) != (int) line.length()) {
			plog("Error: unable to write to serial port: %s\n", line.c_str());
			return false;
		}
		lseek(portFd, line.length(), SEEK_CUR);
	}
	return true;
}

string GSMFeatureProvider::readLine()
{
	char buffer[64], ch = ' ';
	string ret;
	int i = 0, err, waitcount = 0;

	if (portFd >= 0)
	{
		while (ch != '\n') {
			err = read(portFd, &ch, 1);
			if (err == -1) {
				plog("Error: unable to read from serial port\n");
				return "";
				//throw 0;
			}
			else if (err == 1) {
				buffer[i] = ch;
				i++;
				if (i == 64) {
					ret += buffer;
					i=0;
				}
			}
			else { // err == 0 --> no character read from port
				if (waitcount >= 10)
					break;
				waitcount++;
				Thread::sleep(100);
			}
		}
		if (ch == '\n')
			buffer[i-1] = 0;
		ret += buffer;

		return ret;
	}
	else
		return "";
}
