/**
 * @file
 * @brief GSM feature
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-02-12 21:08:39 +0100 (Don, 12 Feb 2004) $
 * $Revision: 612 $
 *
 * @since    Mon May 12 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: GSMWindows.cpp 612 2004-02-12 20:08:39Z harald $
 */

#include "GSM.h"

/** Default serial port to use */
#define GSM_SERIALPORT "COM8:"

/** Serial port handle */
static HANDLE hPort;

void GSMFeatureProvider::init(featureparams& params)
{
	if (params.count("serialport") == 0) {
		this->serialPort = GSM_SERIALPORT;
	} else {
		this->serialPort = params["serialport"];
	}
}

bool GSMFeatureProvider::openSerialPort(string serialport)
{
	DCB PortDCB;
	COMMTIMEOUTS CommTimeouts;

	hPort = INVALID_HANDLE_VALUE;

#ifdef _WIN32_WCE
	wchar_t* port;
	string::size_type len = serialport.length() + 1;
	port = new wchar_t[mbstowcs(NULL, serialport.c_str(), len)];
	mbstowcs(port, serialport.c_str(), len);
#else
	/** name of the serial port */
	const char *port = serialport.c_str();
#endif

	// Open the serial port.
	hPort = CreateFile (port, // Pointer to the name of the port
		GENERIC_READ | GENERIC_WRITE,
		// Access (read/write) mode
		0,            // Share mode
		NULL,         // Pointer to the security attribute
		OPEN_EXISTING,// How to open the serial port
		0,            // Port attributes
		NULL);        // Handle to port with attribute
	// to copy

#ifdef _WIN32_WCE
	delete port;
#endif

	// If it fails to open the port, return FALSE.
	if ( hPort == INVALID_HANDLE_VALUE ) 
	{
		plog("Error: could not open port with error %i\n", GetLastError());

		return false;
	}

	// Get the default port setting information.
	PortDCB.DCBlength = sizeof(DCB);     
	GetCommState(hPort, &PortDCB);

	// Change the DCB structure settings.
	PortDCB.BaudRate = 115200;              // Current baud 
	PortDCB.fBinary = TRUE;               // Binary mode; no EOF check 
	PortDCB.fParity = TRUE;               // Enable parity checking. 
	PortDCB.fOutxCtsFlow = FALSE;         // No CTS output flow control 
	PortDCB.fOutxDsrFlow = FALSE;         // No DSR output flow control 
	PortDCB.fDtrControl = DTR_CONTROL_ENABLE; 
	// DTR flow control type 
	PortDCB.fDsrSensitivity = FALSE;      // DSR sensitivity 
	PortDCB.fTXContinueOnXoff = TRUE;     // XOFF continues Tx 
	PortDCB.fOutX = TRUE;                // XON/XOFF out flow control 
	PortDCB.fInX = TRUE;                 // XON/XOFF in flow control 
	PortDCB.fErrorChar = FALSE;           // Disable error replacement. 
	PortDCB.fNull = FALSE;                // Disable null stripping. 
	PortDCB.fRtsControl = RTS_CONTROL_ENABLE; 
	// RTS flow control 
	PortDCB.fAbortOnError = FALSE;        // Do not abort reads/writes on 
	// error.
	PortDCB.ByteSize = 8;                 // Number of bits/bytes, 4-8 
	PortDCB.Parity = NOPARITY;            // 0-4=no,odd,even,mark,space 
	PortDCB.StopBits = ONESTOPBIT;        // 0,1,2 = 1, 1.5, 2 

	// Configure the port according to the specifications of the DCB 
	// structure.
	if (!SetCommState (hPort, &PortDCB))
	{
		plog("Error: could not set port parameters with error %i\n", GetLastError());

		closeSerialPort();
		return false;
	}

	// Retrieve the time-out parameters for all read and write operations
	// on the port. 
	GetCommTimeouts (hPort, &CommTimeouts);

	// Change the COMMTIMEOUTS structure settings.
	CommTimeouts.ReadIntervalTimeout = MAXDWORD;  
	CommTimeouts.ReadTotalTimeoutMultiplier = 0;  
	CommTimeouts.ReadTotalTimeoutConstant = 0;    
	CommTimeouts.WriteTotalTimeoutMultiplier = 10;  
	CommTimeouts.WriteTotalTimeoutConstant = 1000;    

	// Set the time-out parameters for all read and write operations
	// on the port. 
	if (!SetCommTimeouts (hPort, &CommTimeouts))
	{
		plog("Error: could not set port timeouts with error %i\n", GetLastError());

		closeSerialPort();
		return false;
	}

	// Direct the port to perform extended functions SETDTR and SETRTS.
	// SETDTR: Sends the DTR (data-terminal-ready) signal.
	// SETRTS: Sends the RTS (request-to-send) signal. 
	/*EscapeCommFunction (hPort, SETDTR);
	EscapeCommFunction (hPort, SETRTS);*/

	return true;
}

void GSMFeatureProvider::closeSerialPort()
{
	if (hPort != INVALID_HANDLE_VALUE) {
		// Close the communication port.
		if (!CloseHandle (hPort)) {
			plog("Error: could not close com port with error %i\n", GetLastError());
		} else {
			hPort = INVALID_HANDLE_VALUE;
		}
	}
}


bool GSMFeatureProvider::writeLine(string line)
{
	DWORD dwNumBytesWritten;

	if (!WriteFile (hPort,              // Port handle
		line.c_str(),              // Pointer to the data to write 
		(DWORD) line.length(),                  // Number of bytes to write
		&dwNumBytesWritten, // Pointer to the number of bytes 
		// written
		NULL))              // Must be NULL for Windows CE
	{
		// WriteFile failed. Report error.
		plog("Error: could not write to port with error %i\n", GetLastError());

		return false;
	}
	return true;
}

string GSMFeatureProvider::readLine()
{
	char buffer[64], ch = ' ';
	string ret;
	int i = 0, err, waitcount = 0;
	unsigned long gotBytes;

	while (ch != '\n') {
		err = ReadFile (hPort, &ch, 1, &gotBytes, 0);
		if (err == 1 && gotBytes == 1) {
			if (ch != '\n' && ch != '\r') 
			{
				buffer[i] = ch;
				i++;
				if (i == 64) {
					ret += buffer;
					i=0;
				}
			}
		} else { // no character read from port
			if (waitcount >= 10)
				break;
			waitcount++;
			Thread::sleep(100);
		}
	}
	buffer[i] = 0;
	ret += buffer;

	return ret;
}
