/**
 * @file
 * @brief Abstract string feature class implementation
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-08-29 12:14:03 +0200 (Son, 29 Aug 2004) $
 * $Revision: 905 $
 *
 * @since    Wed May 7 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: Numerical.cpp 905 2004-08-29 10:14:03Z rene $
 */

// for sprintf
#include <stdio.h>
#include "Numerical.h"

NumericalDiscreteFeature::NumericalDiscreteFeature(long *minval, long* maxval)
{
	this->maxval = maxval;
	this->minval = minval;

	if (*maxval > *minval)
		val = *minval + rand_double() * (*maxval - *minval);
	else
		// uninitialized, first sample
		val = 0.0;
}

NumericalDiscreteFeature::NumericalDiscreteFeature(long *minval, long* maxval, long val)
{
	this->val = val;
	this->maxval = maxval;
	this->minval = minval;

	if (val > *(this->maxval)) {
		*(this->maxval) = val;
		invalidate();
	}

	if (val < *(this->minval)) {
		*(this->minval) = val;
		invalidate();
	}
}

double NumericalDiscreteFeature::getPosition() const
{
	double r;
#ifndef _SYMBIAN
	if (*maxval != *minval && *maxval > numeric_limits<long>::min() && *minval < numeric_limits<long>::max()) {
#else
	if (*maxval != *minval && *maxval > KMinTInt && *minval < KMaxTInt) {
#endif
		r = (double) ((long) val - *minval) / (*maxval - *minval);
	} else {
		r = 1.0; // infinite
	}
	assert(r >= 0 && r <= 1);
	return r;
}

double NumericalDiscreteFeature::getDistance(Feature *f) const
{
	assert(*minval == *((NumericalDiscreteFeature*) f)->minval &&
	       *maxval == *((NumericalDiscreteFeature*) f)->maxval);
	return fabs(getPosition() - f->getPosition());
}

void NumericalDiscreteFeature::moveTowards(Feature *f, double factor)
{
	assert(val >= *minval && val <= *maxval);
	assert(factor >= 0.0 && factor <= 1.0);
	assert(((NumericalDiscreteFeature *) f)->val >= *minval &&
	       ((NumericalDiscreteFeature *) f)->val <= *maxval);
	if (val < *minval || val > *maxval) {
		plog("Warning: NumericalDiscreteFeature::moveTowards encountered value out of range, setting directly to target value instead of moving\n");
		val = ((NumericalDiscreteFeature *) f)->val;
	}
	else {
		val += (factor * (((NumericalDiscreteFeature*) f)->val - val));
	}
	assert(val >= *minval && val <= *maxval);
}

string NumericalDiscreteFeature::serialize() const
{
	char buffer [sizeof(long)*8+1];
	sprintf(buffer, "%.17f", val);

	return buffer;
}

void NumericalDiscreteFeature::unserialize(string value)
{
	val = atof(value.c_str());

	if (val > *(this->maxval)) {
		*(this->maxval) = (long) val;
		invalidate();
		plog("Warning: NumericalDiscreteFeature encountered "
		     "value out of range during unserializing, adapting "
		     "maxval to %f. Your persist file may be corrupt or "
		     "incomplete\n", val);
	}

	if (val < *(this->minval)) {
		*(this->minval) = (long) val;
		invalidate();
		plog("Warning: NumericalDescreteFeature encountered "
		     "value out of range during unserializing, adapting "
		     "minval to %f. Your persist file may be corrupt or "
		     "incomplete\n", val);
	}
}

featureparams NumericalDiscreteFeature::write() const
{
	featureparams param;
	char buffer[sizeof(long)*8+1];

	sprintf(buffer, "%ld", *minval);
	param["minval"] = buffer;

	sprintf(buffer, "%ld", *maxval);
	param["maxval"] = buffer;

	return param;
}

void NumericalDiscreteFeature::read(featureparams *param)
{
	featureparams::iterator iter = param->find("minval");
	if (iter != param->end()) {
		*minval = atol((*iter).second.c_str());
	}

	iter = param->find("maxval");
	if (iter != param->end()) {
		*maxval = atol((*iter).second.c_str());
	}
}

Feature* NumericalDiscreteFeature::clone() const
{
	return new NumericalDiscreteFeature(minval, maxval, (long) val);
}

void NumericalDiscreteFeature::aggregate(aggregatelist )
{
}

#if _DEBUG_FEATURES
string NumericalDiscreteFeature::toString() const
{
	char buffer[64];
	sprintf(buffer, "general NumericalDiscreteFeature: %ld", getVal());
	return buffer;
}
#endif




NumericalContinuousFeature::NumericalContinuousFeature(double *minval, double *maxval)
{
	this->maxval = maxval;
	this->minval = minval;

	if (*maxval > *minval)
		val = *minval + rand_double() * (*maxval - *minval);
	else
		// uninitialized, first sample
		val = 0.0;
}

NumericalContinuousFeature::NumericalContinuousFeature(double *minval, double *maxval, double val)
{
	this->val = val;
	this->maxval = maxval;
	this->minval = minval;

	if (val > *(this->maxval)) {
		*(this->maxval) = val;
		invalidate();
	}

	if (val < *(this->minval)) {
		*(this->minval) = val;
		invalidate();
	}
}

double NumericalContinuousFeature::getPosition() const
{
	double r;
#ifndef _SYMBIAN
	if (*maxval != *minval && *maxval > -numeric_limits<double>::infinity() && *minval < numeric_limits<double>::infinity()) {
#else
	if (*maxval != *minval && *maxval > -KMaxTReal && *minval < KMaxTReal) {
#endif
		r = (val - *minval) / (*maxval - *minval);
	} else {
		r = 1.0; // infinite
	}
	assert(r >= 0 && r <= 1);
	return r;
}

double NumericalContinuousFeature::getDistance(Feature *f) const
{
	assert(*minval == *((NumericalContinuousFeature*) f)->minval &&
	       *maxval == *((NumericalContinuousFeature*) f)->maxval);
	return fabs(getPosition() - f->getPosition());
}

void NumericalContinuousFeature::moveTowards(Feature *f, double factor)
{
	assert(val >= *minval && val <= *maxval);
	assert(factor >= 0.0 && factor <= 1.0);
        assert(*((NumericalContinuousFeature *) f)->minval == *minval);
        assert(*((NumericalContinuousFeature *) f)->maxval == *maxval);
	assert(((NumericalContinuousFeature *) f)->val >= *minval &&
	       ((NumericalContinuousFeature *) f)->val <= *maxval);
	if (val < *minval || val > *maxval) {
		plog("Warning: NumericalContinuousFeature::moveTowards encountered value out of range, setting directly to target value instead of moving\n");
		val = ((NumericalContinuousFeature *) f)->val;
	}
	else {
		val += factor * (((NumericalContinuousFeature*) f)->val - val);
	}
	assert(val >= *minval && val <= *maxval);
}


string NumericalContinuousFeature::serialize() const
{
	char buffer [sizeof(double)*8+1];
	sprintf(buffer, "%.17f", val);

	return buffer;
}

void NumericalContinuousFeature::unserialize(string value)
{
	val = atof(value.c_str());

	if (val > *(this->maxval)) {
		*(this->maxval) = val;
		invalidate();
		plog("Warning: NumericalContinuousFeature encountered "
		     "value out of range during unserializing, adapting "
		     "maxval to %f. Your persist file may be corrupt or "
		     "incomplete\n", val);
	}

	if (val < *(this->minval)) {
		*(this->minval) = val;
		invalidate();
		plog("Warning: NumericalContinuousFeature encountered "
		     "value out of range during unserializing, adapting "
		     "minval to %f. Your persist file may be corrupt or "
		     "incomplete\n", val);
	}
}

featureparams NumericalContinuousFeature::write() const
{
	featureparams param;
	char buffer[sizeof(double)*8+1];

	sprintf(buffer, "%.17f", *minval);
	param["minval"] = buffer;

	sprintf(buffer, "%.17f", *maxval);
	param["maxval"] = buffer;

	return param;
}

void NumericalContinuousFeature::read(featureparams *param)
{
	featureparams::iterator iter = param->find("minval");
	if (iter != param->end()) {
		*minval = atof((*iter).second.c_str());
	}

	iter = param->find("maxval");
	if (iter != param->end()) {
		*maxval = atof((*iter).second.c_str());
	}
}

Feature* NumericalContinuousFeature::clone() const
{
	return new NumericalContinuousFeature(minval, maxval, val);
}

void NumericalContinuousFeature::aggregate(aggregatelist )
{
}

#if _DEBUG_FEATURES
string NumericalContinuousFeature::toString() const
{
	char buffer[64];
	sprintf(buffer, "general NumericalContinuousFeature: %.17f", getVal());
	return buffer;
}
#endif

