/**
 * @file
 * @brief Numerical feature class declaration
 *
 * This file contains the abstract class definitions that can be used
 * by sensors that return numerical values.
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-03-01 16:47:51 +0100 (Mon, 01 Mär 2004) $
 * $Revision: 666 $
 *
 * @since    Wed May 7 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: Numerical.h 666 2004-03-01 15:47:51Z rene $
 */

#ifndef _NUMERICAL_H
#define _NUMERICAL_H

#include "Feature.h"

/**
 * Abstract discrete numerical feature
 *
 * A feature of type numerical_discrete. This class imeplements the
 * getType method of the Feature interface and, as a base class for special
 * features, implements the complete handling of numerical discrete values.
 * This includes the implementations of getDistance and moveTowards as well as
 * getPosition, serialize and unserialize. Additionally, minimum and maximum
 * seen values are stored as persistant values of the feature and used for
 * computing the normalized distance and position. Although these values are
 * handled inside this class, the storage for them has to be provided by
 * the subclasses (or generally the factory that creates objects of this
 * class).
 */
class NumericalDiscreteFeature: public PersistantFeature
{
	protected:
		/** The feature value. */
		double val;
		/** A reference to the static, persistant minimum value seen so far. */
		long *minval;
		/** A reference to the static, persistant maximum value seen so far. */
		long *maxval;

		/**
		 * %Feature constructor
		 *
		 * This constructor initializes the feature with a random value
		 * and should thus only be used for creating e.g. prototypes
		 * of points in a clustering space. For creating a specific
		 * sample of a feature, the second constructor should be used.
		 *
		 * @param minval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the minimum value.
		 * @param maxval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum value.
		 */
		NumericalDiscreteFeature(long *minval, long *maxval);

		/** 
		 * Sample constructor
		 *
		 * This constructor initializes the feature with a specific
		 * value and should thus be used for samples.
		 *
		 * @param minval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the minimum value.
		 * @param maxval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum value.
		 * @param val The sample value.
		 */
		NumericalDiscreteFeature(long *minval, long *maxval, long val);

	public:
		virtual double getPosition() const;
		virtual double getDistance(Feature *f) const;
		virtual void moveTowards(Feature *f, double factor);

		virtual string serialize() const;
		virtual void unserialize(string value);

		virtual featureparams write() const;
		virtual void read(featureparams *param);

		virtual Feature* clone() const;
		virtual void aggregate(aggregatelist samples);

		virtual FeatureType getType() const { return numerical_discrete; }
		virtual const string getName() const { return "NumericalDiscreteFeature"; }

		/** @return The value of the feature. */
		const long getVal() const { return (long) val; }

		const long getMin() const { return *minval; }
		const long getMax() const { return *maxval; }

#if _DEBUG_FEATURES
		virtual string toString() const;
#endif
};

/**
 * Abstract continuous numerical feature
 *
 * A feature of type numerical_continuous. Please see
 * NumericalDiscreteFeature for an explanation of the details, which is also
 * valid for this class.
 *
 * @see NumericalDiscreteFeature
 */
class NumericalContinuousFeature: public PersistantFeature
{
	protected:
		/** The feature value. */
		double val;
		/** A reference to the static, persistant minimum value seen so far. */
		double *minval;
		/** A reference to the static, persistant maximum value seen so far. */
		double *maxval;

		/** 
		 * %Feature constructor
		 *
		 * This constructor initializes the feature with a random value
		 * and should thus only be used for creating e.g. prototypes
		 * of points in a clustering space. For creating a specific
		 * sample of a feature, the second constructor should be used.
		 *
		 * @param minval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the minimum value.
		 * @param maxval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum value.
		 */
		NumericalContinuousFeature(double *minval, double *maxval);

		/** 
		 * Sample constructor
		 *
		 * This constructor initializes the feature with a specific
		 * value and should thus be used for samples.
		 *
		 * @param minval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the minimum value.
		 * @param maxval A reference to a static variable (it must be
		 *               same for all created objects of _one_ feature)
		 *               for persistantly storing the maximum value.
		 * @param val The sample value.
		 */
		NumericalContinuousFeature(double *minval, double *maxval, double val);

	public:
		virtual double getPosition() const;
		virtual double getDistance(Feature *f) const;
		virtual void moveTowards(Feature *f, double factor);

		virtual string serialize() const;
		virtual void unserialize(string value);

		virtual featureparams write() const;
		virtual void read(featureparams *param);

		virtual Feature* clone() const;
		virtual void aggregate(aggregatelist samples);

		virtual FeatureType getType() const { return numerical_continuous; }
		virtual const string getName() const { return "NumericalContinuousFeature"; }

		/** @return The value of the feature. */
		const double getVal() const { return val; }

		const double getMin() const { return *minval; }
		const double getMax() const { return *maxval; }

#if _DEBUG_FEATURES
		virtual string toString() const;
#endif
};

#endif
