/**
 * @file
 * @brief System command string feature class implementation
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-06-08 11:37:19 +0200 (Die, 08 Jun 2004) $
 * $Revision: 808 $
 *
 * @since    Mon Jul 7 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: SystemCommandStringList.cpp 808 2004-06-08 09:37:19Z rene $
 */

#include "SystemCommandStringList.h"

/** Default delay (in seconds) between two system calls */
#define SYSTEMCOMMAND_DELAY 15

const string SystemCommandStringListFeature::getName() const {
	string ret = providerName;
	ret += ".";
	ret += featureName;
	return ret;
}

#if _DEBUG_FEATURES
string SystemCommandStringListFeature::toString() const
{
	char buffer[16];
	string ret = featureName + ": '";
	for (stringvector::const_iterator iter = getListValues().begin(); iter != getListValues().end(); iter++) {
		ret += *iter;
		ret += ", ";
	}
	ret += "', has code ";
	sprintf(buffer, "%ld", codeval);
	ret += buffer;
	return ret;
}
#endif

// FeatureProvider
SystemCommandStringListFeatureProvider::SystemCommandStringListFeatureProvider(
	string featureProviderName,
	string featureName,
	string systemCommand,
	stringvector* featureList,
	stringcode* stringCodeList,
	long* maxlen,
	featureparams& params) :
	FeatureProvider(featureProviderName, featureList),
	providerName(featureProviderName), featureName(featureName),
	systemCommand(systemCommand)
{
	this->stringCodeList = stringCodeList;
	this->maxlen = maxlen;
	stringListValid = false;

	if (params.count("delay") == 0)
		delay = SYSTEMCOMMAND_DELAY * 1000;
	else
		/// @todo check parameter
		delay = atoi(params["delay"].c_str());
}

SystemCommandStringListFeatureProvider::~SystemCommandStringListFeatureProvider() throw()
{
	// stop the system command thread and wait for it to finish
	this->cancel();
	this->join();

	this->FeatureProvider::~FeatureProvider();
}


Feature* SystemCommandStringListFeatureProvider::getFeature(string name) const
{
	if (name == featureName) {
		return new SystemCommandStringListFeature(stringCodeList, maxlen, NULL, providerName, featureName);
	}

	return NULL;
}

Feature* SystemCommandStringListFeatureProvider::getSample(string name) const
{
	Guard g(lockStringList);

	if (name == featureName && stringListValid) {
		return new SystemCommandStringListFeature(stringCodeList, maxlen, &stringList, providerName, featureName);
	}

	return NULL;
}

void SystemCommandStringListFeatureProvider::nextSample(clock_t)
{
}

void SystemCommandStringListFeatureProvider::run() throw()
{
	string cmdOutput;

	while (!this->isCancelled())
	{
		try {
		// execute the system command and get the standard output output back
			cmdOutput = executeSystemCommand();

			// keep the locked block to a minimum - the delay is caused by hci_inqiry
			{
				Guard g(lockStringList);
				string::size_type pos = 0, pos1;
				stringList.clear();
				while (pos < cmdOutput.length()) {
					pos1 = cmdOutput.find('\n', pos);
					stringList.push_back(cmdOutput.substr(pos, pos1-pos));
					pos = pos1+1;
				}
				stringListValid = true;
			}
		}
		catch(...)
		{
			stringListValid = false;
		}

		// HACK, HACK, HACK
		unsigned int sleep, num;
		if (delay > 1000) {
			num = delay/100;
			sleep = 100;
		}
		else {
			num = 1;
			sleep = delay;
		}
		for (unsigned int i=0; i<num && !this->isCancelled(); i++)
			this->sleep(sleep);
	}
}

#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>

// execute the system command and get the standard output output back
// keep the locked block to a minimum - the delay is caused by hci_inqiry
string SystemCommandStringListFeatureProvider::executeSystemCommand()
{
	int fds[2];
	pid_t child;
	string cmdOutput;
	char ch;

	pipe(fds);
	child = fork();
	switch (child)
	{
		case -1: // oops, error
			throw 1;
		case 0: // this is the child (the system call)
			close(1); // close normal stdout
			dup(fds[1]); // use the stdout to the pipe
			close(fds[1]);
			close(fds[0]);
			execlp("/bin/sh", "sh", "-c", systemCommand.c_str(), 0);
			// just in case that exit didn't work
			exit(127);
		default: // this is the parent (our main process, i.e. thread)
			close(fds[1]); // we won't write to the pipe
			// wait for the child to terminate
			do {
				int status;
				if (waitpid(child, &status, 0) == -1) {
					if (errno != EINTR)
						throw 1;
		        	} else
					break;
			} while (true);
			// read its standard input
			while (read(fds[0], &ch, 1))
				cmdOutput += ch;
			close(fds[0]);
			return cmdOutput;
	}
}
