/**
 * @file
 * @brief Time feature
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-08-29 12:14:03 +0200 (Son, 29 Aug 2004) $
 * $Revision: 905 $
 *
 * @since    Wed Apr 9 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: Time.cpp 905 2004-08-29 10:14:03Z rene $
 */

// for sprintf
#include <stdio.h>
#include "TimeFeature.h"
#include "../../util/debug.h"

EXPORT_FEATURE(TimeFeatureProvider);

// Feature
TimeFeature::TimeFeature(FeatureKind kind, time_t* minval, time_t* maxval): minval(minval), maxval(maxval)
{
	this->feature = kind;
	if (*maxval >= *minval)
		timestamp = (time_t) (*minval + rand_double() * (*maxval - *minval));
	else
		// for the first sample value, it can happen that the feature (cluster position) has been initialized with default min and maxvals and is therefore invalid
		timestamp = 0; // at least some default....
}

TimeFeature::TimeFeature(FeatureKind kind, time_t* minval, time_t* maxval, time_t timestamp): minval(minval), maxval(maxval)
{
	this->feature = kind;
	this->timestamp = timestamp;

	if (timestamp > *maxval) {
		*(this->maxval) = timestamp;
		invalidate();
	}

	if (timestamp < *minval) {
		*(this->minval) = timestamp;
		invalidate();
	}
}

double TimeFeature::getDistance(Feature *f) const
{
	// this can't happen, but just be sure
	assert(*minval < numeric_limits<time_t>::max() && *maxval > numeric_limits<time_t>::min());
	assert(*maxval >= *minval);

	int d;
	TimeFeature *tf = (TimeFeature *) f;
	time_t t = tf->timestamp;
	tm *a = gmtime(&this->timestamp);
	tm *b = gmtime(&t);
	double sign;

	if (this->timestamp < t) {
		sign = 1.0;
	} else {
		sign = -1.0;
	}

	switch (this->feature) {
		case Timestamp:
			if (*maxval > *minval)
				return fabs((double)(this->timestamp - t) / (*maxval - *minval));
			else
				// not initialized
				return 1.0;
			break;

		case Second:
			d = abs(a->tm_sec - b->tm_sec);
			if (d > 30) {
				d = 60 - d;
			}

			return (double) sign * d / 30;
			break;

		case Minute:
			d = abs(a->tm_min - b->tm_min);
			if (d > 30) {
				d = 60 - d;
			}

			return (double) sign * d / 30;
			break;

		case Hour:
			d = abs(a->tm_hour - b->tm_hour);
			if (d > 12) {
				d = 24 - d;
			}

			return (double) sign * d / 12;
			break;

		case Day:
			d = abs(a->tm_mday - b->tm_mday);
			if (d > 15) {
				d = 31 - d;
			}

			return (double) sign * d / 15;
			break;

		case Weekday:
			d = abs(a->tm_wday - b->tm_wday);
			if (d > 3) {
				d = 7 - d;
			}

			return (double) sign * d / 3;
			break;

		case Yearday:
			d = abs(a->tm_yday - b->tm_yday);
			if (d > 183) {
				d = 356 - d;
			}

			return (double) sign * d / 183;
			break;

		case Month:
			d = abs(a->tm_mon - b->tm_mon);
			if (d > 6) {
				d = 12 - d;
			}

			return (double) sign * d / 6;
			break;

		case Year:
			return (double) a->tm_year - b->tm_year;
			break;
	}

	return -1.0; // error
}

void TimeFeature::read(featureparams *param)
{
	featureparams::iterator iter = param->find("minval");
	if (iter != param->end()) {
#ifdef _INTEGRAL_MAX_BITS
		*minval = (time_t) _atoi64((*iter).second.c_str());
#else
		*minval = atol((*iter).second.c_str());
#endif
	}

	iter = param->find("maxval");
	if (iter != param->end()) {
#ifdef _INTEGRAL_MAX_BITS
		*maxval = (time_t) _atoi64((*iter).second.c_str());
#else
		*maxval = atol((*iter).second.c_str());
#endif
	}
}

featureparams TimeFeature::write() const
{
	featureparams param;
	char buffer[sizeof(time_t)*8+1];

#if defined _INTEGRAL_MAX_BITS && ! defined _WIN32_WCE
	_i64toa(*minval, buffer, 10);
#else
	sprintf(buffer, "%ld", *minval);
#endif
	param["minval"] = buffer;

#if defined _INTEGRAL_MAX_BITS && ! defined _WIN32_WCE
	_i64toa(*maxval, buffer, 10);
#else
	sprintf(buffer, "%ld", *maxval);
#endif
	param["maxval"] = buffer;

	return param;
}

double TimeFeature::getPosition() const
{
	tm *a = gmtime(&this->timestamp);
	double r;

	switch (this->feature) {
		case Timestamp:
		case Year:
			r = (double) (this->timestamp - *minval) / (*maxval - *minval);
			break;

		case Second:
			r = (double) a->tm_sec / 59;
			break;

		case Minute:
			r = (double) a->tm_min / 59;
			break;

		case Hour:
			r = (double) a->tm_hour / 23;
			break;

		case Day:
			r = (double) (a->tm_mday - 1) / 30;
			break;

		case Weekday:
			r = (double) a->tm_wday / 6;
			break;

		case Yearday:
			r = (double) a->tm_year / 365;
			break;

		case Month:
			r = (double) a->tm_mon / 11;
			break;

		default:
			throw 1;
			break;
	}

	assert(r >= 0.0 && r <= 1.0);
	return r;
}

void TimeFeature::moveTowards(Feature *f, double factor)
{
	if (timestamp < *minval || timestamp > *maxval) {
//		plog("Warning: TimeFeature::moveTowards encountered timestamp out of range, setting directly to target value instead of moving\n");
		timestamp = ((TimeFeature *) f)->timestamp;
	}
	else {
		double move = this->getDistance(f) * factor * ((timestamp <= ((TimeFeature *) f)->timestamp) ? 1.0 : -1.0);

		switch (feature) {
			case Second:
				this->timestamp += (time_t) (move * 30);
				break;

			case Minute:
				this->timestamp += (time_t) (move * 60 * 30);
				break;

			case Hour:
				this->timestamp += (time_t) (move * 3600 * 12);
				break;

			case Day:
				this->timestamp += (time_t) (move * 86400 * 15);
				break;

			case Weekday:
				this->timestamp += (time_t) (move * 86400 * 3);
				break;

			case Yearday:
				this->timestamp += (time_t) (move * 86400 * 183);
				break;

			case Month:
				this->timestamp += (time_t) (move * 2635200 * 6);
				break;

			case Year:
			case Timestamp:
				this->timestamp += (time_t) (move * (*maxval - *minval));
				break;
		}
	}

	assert(this->timestamp <= *maxval);
	assert(this->timestamp >= *minval);
}

string TimeFeature::serialize() const
{
	char buffer [sizeof(long)*8+1];
	sprintf(buffer, "%ld", (unsigned long) timestamp);

	return buffer;
}

void TimeFeature::unserialize(string value)
{
	timestamp = (time_t) atol(value.c_str());

	assert(this->timestamp <= *maxval);
	assert(this->timestamp >= *minval);
}

Feature* TimeFeature::clone() const
{
	return new TimeFeature(feature, minval, maxval, timestamp);
}

void TimeFeature::aggregate(aggregatelist )
{
}

const string TimeFeature::getName() const
{
	switch (this->feature) {
		case Timestamp:
			return "Time.Timestamp";

		case Second:
			return "Time.Second";

		case Minute:
			return "Time.Minute";

		case Hour:
			return "Time.Hour";

		case Day:
			return "Time.Day";

		case Weekday:
			return "Time.Weekday";

		case Yearday:
			return "Time.Yearday";

		case Month:
			return "Time.Month";

		case Year:
			return "Time.Year";

		default:
			return "Time.Unkown";
	}
}

#if _DEBUG_FEATURES
string TimeFeature::toString() const
{
	tm *a = gmtime(&this->timestamp);

	char buffer[64];
	switch (this->feature) {
		case Timestamp:
			sprintf(buffer, "TimeFeature/Timestamp: %ld", this->timestamp);
			break;

		case Second:
			sprintf(buffer, "TimeFeature/Second: %d", a->tm_sec);
			break;

		case Minute:
			sprintf(buffer, "TimeFeature/Minute: %d", a->tm_min);
			break;

		case Hour:
			sprintf(buffer, "TimeFeature/Hour: %d", a->tm_hour);
			break;

		case Day:
			sprintf(buffer, "TimeFeature/Day: %d", (a->tm_mday - 1));
			break;

		case Weekday:
			sprintf(buffer, "TimeFeature/Weekday: %d", a->tm_wday);
			break;

		case Yearday:
			sprintf(buffer, "TimeFeature/Yearday: %d", a->tm_yday);
			break;

		case Month:
			sprintf(buffer, "TimeFeature/Month: %d", a->tm_mon);
			break;

		case Year:
			sprintf(buffer, "TimeFeature/Year: %d", (a->tm_year + 1900));
			break;
	}
	return buffer;
}
#endif



TimeFeatureProvider::TimeFeatureProvider(providerparams &) : FeatureProvider("Time", &features) {
	featuremap["Timestamp"] = TimeFeature::Timestamp;
	featuremap["Second"] = TimeFeature::Second;
	featuremap["Minute"] = TimeFeature::Minute;
	featuremap["Hour"] = TimeFeature::Hour;
	featuremap["Day"] = TimeFeature::Day;
	featuremap["Weekday"] = TimeFeature::Weekday;
	featuremap["Yearday"] = TimeFeature::Yearday;
	featuremap["Month"] = TimeFeature::Month;
	featuremap["Year"] = TimeFeature::Year;

	features.push_back("Timestamp");
	features.push_back("Second");
	features.push_back("Minute");
	features.push_back("Hour");
	features.push_back("Day");
	features.push_back("Weekday");
	features.push_back("Yearday");
	features.push_back("Month");
	features.push_back("Year");

#ifndef _SYMBIAN
	timeFeature_minval = numeric_limits<time_t>::max();
	timeFeature_maxval = numeric_limits<time_t>::min();
#else
	timeFeature_minval = KMaxTInt;
	timeFeature_maxval = KMinTInt;
#endif
}

Feature* TimeFeatureProvider::getFeature(string name) const
{
	map<string, TimeFeature::FeatureKind>::const_iterator iter = featuremap.find(name);
	if (iter != featuremap.end()) {
		return new TimeFeature((*iter).second, const_cast<time_t*> (&timeFeature_minval), const_cast<time_t*> (&timeFeature_maxval));
	} else {
		return NULL;
	}
}

Feature* TimeFeatureProvider::getSample(string name) const
{
	map<string, TimeFeature::FeatureKind>::const_iterator iter = featuremap.find(name);
	if (iter != featuremap.end()) {
		return new TimeFeature((*iter).second, const_cast<time_t*> (&timeFeature_minval), const_cast<time_t*> (&timeFeature_maxval), timestamp);
	} else {
		return NULL;
	}
}

void TimeFeatureProvider::nextSample(clock_t)
{
	time(&timestamp);
}
