/**
 * @file
 * @brief Video feature
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-08-19 15:40:27 +0200 (Don, 19 Aug 2004) $
 * $Revision: 899 $
 *
 * @since    Wed Apr 9 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * Quite some of the code in here has been ripped from various files in the 
 * "motion" source code, a tool for detecting motion using webcams.
 *
 * $Id: Video.cpp 899 2004-08-19 13:40:27Z rene $
 */

// for sprintf
#include <stdio.h>
#include "Video.h"

#define DEFAULT_WIDTH 320
#define DEFAULT_HEIGHT 240
#define DEFAULT_FREQUENCY 5


EXPORT_FEATURE(VideoFeatureProvider);

#if _DEBUG_FEATURES
string VideoMotionFeature::toString() const
{
	char buffer[64];
	sprintf(buffer, "VideoMotionFeature: %.17f", getVal());
	return buffer;
}
#endif

Feature* VideoFeatureProvider::getFeature(string name) const
{
	if (name == "Motion") {
		return new VideoMotionFeature(const_cast<double*> (&videoMotionFeature_minval), const_cast<double*> (&videoMotionFeature_maxval));
	}

	return NULL;
}

Feature* VideoFeatureProvider::getSample(string name) const
{
	if (name == "Motion") {
		Guard g(lockCurMotion);
		return new VideoMotionFeature(const_cast<double*> (&videoMotionFeature_minval), const_cast<double*> (&videoMotionFeature_maxval), curMotion);
	}

	return NULL;
}

void VideoFeatureProvider::nextSample(clock_t checkpoint)
{
	(void) checkpoint;	// to get rid of unused variable warnings
}


// Linux specific !
typedef u_int64_t __u64;
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/videodev.h>

#define DEFAULT_DEVICE "/dev/video0"


static int dev_fd, v4l_fmt, v4l_norm, v4l_input, v4l_bufsize, v4l_curbuffer, v4l_maxbuffer;
static unsigned char* img_buf;
static unsigned char *v4l_buffers[2];

VideoFeatureProvider::VideoFeatureProvider(providerparams &params) : FeatureProvider("Video", &features)
{
	features.push_back("Motion");

/// @todo symbian stl fix
#ifndef _SYMBIAN
	videoMotionFeature_minval = numeric_limits<double>::infinity();
	videoMotionFeature_maxval = -numeric_limits<double>::infinity();
#else
	videoMotionFeature_minval = KMaxReal;
	videoMotionFeature_maxval = KMinReal;
#endif

	if (params.count("device") == 0)
		deviceName = DEFAULT_DEVICE;
	else
		/// @todo check parameter !!
		deviceName = params["device"];
	if (params.count("width") == 0)
		width = DEFAULT_WIDTH;
	else
		/// @todo check parameter !!
		width = atoi(params["width"].c_str());
	if (params.count("height") == 0)
		height = DEFAULT_HEIGHT;
	else
		/// @todo check parameter !!
		height = atoi(params["height"].c_str());
	if (params.count("frequency") == 0)
		frequency = DEFAULT_FREQUENCY;
	else
		/// @todo check parameter !!
		frequency = atoi(params["frequency"].c_str());
		
	// open the video device and start the motion analysis thread
	int motionsize, size;
	v4l_fmt=VIDEO_PALETTE_YUV420P;
	v4l_norm = 0; // PAL
	v4l_input = 1; // first input
	
	dev_fd=open(deviceName.c_str(), O_RDWR);
	if (dev_fd <0) {
		perror("failed to open video device");
		throw 1;
	}


	//v4l_start (viddevs[i], width, height, input, norm, cnt->conf.frequency);
	
	switch (v4l_fmt) {
		case VIDEO_PALETTE_RGB24:
			size=width*height*3;
			motionsize=width*height*3;
			break;
		case VIDEO_PALETTE_GREY:
			size=width*height;
			motionsize=width*height;
			break;
		case VIDEO_PALETTE_YUV422:
			v4l_fmt=VIDEO_PALETTE_YUV420P;
		case VIDEO_PALETTE_YUV420P:
			size=(width*height*3)/2;
			motionsize=width*height;
			break;
	}

	this->start();
}

VideoFeatureProvider::~VideoFeatureProvider()
{
	// stop thread and close video device
	this->cancel();
	this->join();
	
	close(dev_fd);
}

void VideoFeatureProvider::run()
{
	while (!this->isCancelled()) {
		{
			Guard g(lockCurMotion);
		}
	}
}


char *VideoFeatureProvider::v4l_start()
{
        struct video_capability vid_caps;
	struct video_channel vid_chnl;
	struct video_tuner vid_tuner;
	struct video_window vid_win;
	struct video_mbuf vid_buf;
	struct video_mmap vid_mmap;
	char *map;
/*
	if (ioctl (dev_fd, VIDIOCGCAP, &vid_caps) == -1) {
		perror ("ioctl (VIDIOCGCAP)");
		return (NULL);
	}
	if (vid_caps.type & VID_TYPE_MONOCHROME) v4l_fmt=VIDEO_PALETTE_GREY;
        if (v4l_input != IN_DEFAULT) {
                vid_chnl.channel = -1;
                if (ioctl (dev,_fd VIDIOCGCHAN, &vid_chnl) == -1) {
                        perror ("ioctl (VIDIOCGCHAN)");
                } else {
                        vid_chnl.channel = v4l_input;
                        vid_chnl.norm    = v4l_norm;
                        if (ioctl (dev_fd, VIDIOCSCHAN, &vid_chnl) == -1) {
                                perror ("ioctl (VIDIOCSCHAN)");
                                return (NULL);
                        }
                }
        }
	if (frequency) {
		if (ioctl (dev_fd, VIDIOCGTUNER, &vid_tuner)==-1) {
			perror("ioctl (VIDIOCGTUNER)");
		} else {
			if (vid_chnl.channel & VIDEO_TUNER_LOW) {
				// steps of 1/16 KHz 
				frequency=frequency*16; 
			} else {
				frequency=(frequency*10)/625;
			}
			if (ioctl(dev_fd, VIDIOCSFREQ, &frequency)==-1) {
				perror("ioctl (VIDIOCSFREQ)");
				return (NULL);
			}
			printf("Frequency set\n");
		}
	}
	if (ioctl (dev_fd, VIDIOCGMBUF, &vid_buf) == -1) {
		fprintf(stderr, "\tno mmap falling back on read\n");
		if (ioctl (dev_fd, VIDIOCGWIN, &vid_win)== -1) {
			perror ("ioctl VIDIOCGWIN");
			return (NULL);
		}
		vid_win.width=width;
		vid_win.height=height;
		vid_win.clipcount=0;
		if (ioctl (dev_fd, VIDIOCSWIN, &vid_win)== -1) {
			perror ("ioctl VIDIOCSWIN");
			return (NULL);
		}
//		viddev->v4l_read_img=1;
		map=malloc(width*height*3);
		v4l_maxbuffer=1;
		v4l_curbuffer=0;
		v4l_buffers[0]=map;
		goto done;
	}
	map=mmap(0, vid_buf.size, PROT_READ|PROT_WRITE, MAP_SHARED, dev_fd, 0);
//	v4l_buffers[0]=map; 
	if (vid_buf.frames>1) {
		v4l_maxbuffer=2;
		v4l_buffers[0]=map;
		v4l_buffers[1]=map+vid_buf.offsets[1];
	} else {
		v4l_buffers[0]=map;
		v4l_maxbuffer=1;
	}

	if ((unsigned char *)-1 == (unsigned char *)map) {
		return (NULL);
	}
	v4l_curbuffer=0;
	vid_mmap.format=v4l_fmt;
	vid_mmap.frame=v4l_curbuffer;
	vid_mmap.width=width;
	vid_mmap.height=height;
	if (ioctl(dev_fd, VIDIOCMCAPTURE, &vid_mmap) == -1) {
		printf("Failed with YUV420P, trying YUV422 palette\n");
		v4l_fmt=VIDEO_PALETTE_YUV422;
		vid_mmap.format=v4l_fmt;
		// Try again... 
		if (ioctl(dev_fd, VIDIOCMCAPTURE, &vid_mmap) == -1) {
			printf("Failed with YUV422, trying RGB24 palette\n");
			v4l_fmt=VIDEO_PALETTE_RGB24;
			vid_mmap.format=v4l_fmt;
			// Try again... 
			if (ioctl(dev_fd, VIDIOCMCAPTURE, &vid_mmap) == -1) {
				perror("ioctl VIDIOCMCAPTURE");
				return (NULL);
			}
		}
	}
done:
	switch (v4l_fmt) {
		case VIDEO_PALETTE_YUV420P:
			v4l_bufsize=(width*height*3)/2;
			break;
		case VIDEO_PALETTE_YUV422:
			4l_bufsize=(width*height*2);
			break;
		case VIDEO_PALETTE_RGB24:
			v4l_bufsize=(width*height*3);
			break;
		case VIDEO_PALETTE_GREY:
			v4l_bufsize=width*height;
			break;
	}
	return map;
*/
}
