/*	alg.c
 *
 *	Detect changes in a video stream.
 *	Copyright 2001 by Jeroen Vreeken (pe1rxq@amsat.org)
 *	This software is distributed under the GNU public license version 2
 *	See also the file 'COPYING'.
 *
 */
#include "motion.h"
#include "video.h"
#include "conf.h"
#include "alg.h"
#include "event.h"

/* locate the center of the movement. */
struct coord alg_locate_center (struct images *imgs, int width, int height)
{
	char *out=imgs->out;
	int x, y, centc=0;
	struct coord cent;

	cent.x=0;
	cent.y=0;
	/* Locate movement */
	for (y=0; y<height; y++) {
		for (x=0; x<width; x++) {
			if (*(out++)) {
				cent.x+=x;
				cent.y+=y;
				centc++;
			}
		}
	}
	if (centc) {
		cent.x=cent.x/centc;
		cent.y=cent.y/centc;
	}
	return (cent);
}

/* draw a box around the movement */
void alg_locate (struct coord *cent, struct images *imgs, int width, int height)
{
	char *out=imgs->out, *new=imgs->new;
	int x, y, maxx=0, maxy=0, minx=width, miny=height;
	int centc=0, xdist=0, ydist=0;

	centc=0;
	for (y=0; y<height; y++) {
		for (x=0; x<width; x++) {
			if (*(out++)) {
				if (x > cent->x) xdist+=x-cent->x;
				if (y > cent->y) ydist+=y-cent->y;
				if (x < cent->x) xdist+=cent->x-x;
				if (y < cent->y) ydist+=cent->y-y;
				centc++;
			}
		}	
	}
	if (centc) {
		minx=cent->x-xdist/centc*2;
		maxx=cent->x+xdist/centc*2;
		/* Make the box a litle bigger in y direction to make sure the
		 * heads fit in */
		miny=cent->y-ydist/centc*3;
		maxy=cent->y+ydist/centc*2;
	}
	if (minx < 0) minx=0;
	if (miny < 0) miny=0;
	if (maxx > width-1) maxx=width-1;
	if (maxy > height-1) maxy=height-1;
	if (maxx < 0) maxx=0;
	if (maxy < 0) maxy=0;
	if (minx > width-1) minx=width-1;
	if (miny > height-1) miny=height-1;
	cent->width=maxx-minx;
	cent->height=maxy-miny;
	out=imgs->out;
	
	/* Draw a box around the movement */
	for (x=minx; x<=maxx; x++) {
		new[x+width*miny]=~new[x+width*miny];
		new[x+width*maxy]=~new[x+width*maxy];
		out[x+width*miny]=~out[x+width*miny];
		out[x+width*maxy]=~out[x+width*maxy];
	}
	for (y=miny; y<=maxy; y++) {
		new[minx+y*width]=~new[minx+y*width];
		new[maxx+y*width]=~new[maxx+y*width];
		out[minx+y*width]=~out[minx+y*width];
		out[maxx+y*width]=~out[maxx+y*width];
	}
}

#define NORM			100
#define ABS(x)			((x)<0 ? -(x) : (x))
#define DIFF(x, y)		(ABS((x)-(y)))
#define NDIFF(x, y)		(ABS(x)*NORM/(ABS(x)+2*DIFF(x,y)))
#define NRES(in, min, max)	((in)<(min)?NDIFF(min,in):((in)>(max)?NDIFF(max,in):100))

int alg_predict_eval_set(struct predict *min, struct predict *max, struct predict *in)
{
	return (
		NRES(in->x, min->x, max->x)+
		NRES(in->y, min->y, max->y)+
		NRES(in->dx, min->dx, max->dx)+
		NRES(in->dy, min->dy, max->dy)+
		NRES(in->width, min->width, max->height)+
		NRES(in->height, min->height, max->height)+
		NRES(in->diffs, min->diffs, max->diffs)+
		NRES(in->length, min->length, max->length)
	)/8;
}

void alg_predict_eval(struct context *cnt)
{
	struct predict_set *set=cnt->predict_sets;
	int high=0, i;
	struct predict_set *highset=set;

	while (set) {
		i=alg_predict_eval_set(&set->min, &set->max, &cnt->predict);
		printf("%s: %d\n", set->name, i);
		if (i>high) {
			high=i;
			highset=set;
		}
		set=set->next;
	}
	if (highset && high>=cnt->conf.predict_threshold) {
		printf("Best guess: %s (%d%%)\n", highset->name, high);
	} else {
		printf("Unknown type of movement\n");
	}
}

void alg_predict(struct coord *box, struct context *cnt, int diffs, int first)
{
	struct predict *avg=&cnt->predict;
	struct coord *lastbox=&cnt->predict_lastbox;

	if (first) {
		memcpy(&cnt->predict_lastbox, box, sizeof(struct coord));
		cnt->last_diffs=diffs;
		memset(avg, 0, sizeof(struct predict));
		avg->length=1;
		avg->x=box->x;
		avg->y=box->y;
		avg->width=box->width;
		avg->height=box->height;
		avg->diffs=diffs;
		return;
	}
	avg->length++;
	avg->x+=box->x;
	avg->y+=box->y;
	avg->width+=box->width;
	avg->height+=box->height;
	avg->dx+=box->x-lastbox->x;
	avg->dy+=box->y-lastbox->y;
	avg->diffs+=diffs;

	memcpy(&cnt->predict_lastbox, box, sizeof(struct coord));
	cnt->last_diffs=diffs;
	return;
}

void alg_predict_result(struct context *cnt, char *filename)
{
	struct predict *avg=&cnt->predict;
	FILE *fddesc;

	if (!avg->length)
		return;
	avg->x/=avg->length;
	avg->y/=avg->length;
	avg->width/=avg->length;
	avg->height/=avg->length;
	avg->diffs/=avg->length;
	printf("length: %d\t", avg->length);
	printf("x: %d\ty: %d\twidth: %d\theight: %d\tdiffs: %d\t", avg->x, avg->y, avg->width, avg->height, avg->diffs);
	printf("dx:%d\tdy:%d\n", avg->dx, avg->dy);
	alg_predict_eval(cnt);
	if ((fddesc=fopen(filename, "w"))==NULL) {
		perror("Error writing motion description file");
		return;
	}
	fprintf(fddesc, "Auto generated description\n");
	fprintf(fddesc, "%d\t", avg->length);
	fprintf(fddesc, "%d\t%d\t%d\t%d\t%d\t", avg->x, avg->y, avg->width, avg->height, avg->diffs);
	fprintf(fddesc, "%d\t%d\n", avg->dx, avg->dy);
	fprintf(fddesc, "%d\t", avg->length);
	fprintf(fddesc, "%d\t%d\t%d\t%d\t%d\t", avg->x, avg->y, avg->width, avg->height, avg->diffs);
	fprintf(fddesc, "%d\t%d\n", avg->dx, avg->dy);
	fclose(fddesc);
}


void alg_noise_tune (struct context *cnt, int motion)
{
	struct images *imgs=&cnt->imgs;
	int i;
	char *ref=imgs->ref;
	char *new=imgs->new;
	int diff, sum=0, count=0;
	unsigned char *mask=imgs->mask;

	i=imgs->motionsize;
	for (; i>0; i--) {
		diff=*ref-*new;
		if (mask)
			diff=((diff**mask++)/255);
		sum+=ABS(diff);
		if (diff)
			count++;
		ref++;
		new++;
	}
	sum/=count/4;
	if (!motion || sum>4*cnt->noise)
		cnt->noise=(cnt->noise+sum)/2;
	printf("New noise level: %d\n", cnt->noise);
}

void alg_threshold_tune (struct context *cnt, int diffs, int motion)
{
	int i;
	int sum=0, top=diffs;

	if (!diffs)
		return;

	printf("%d\n", diffs);
	if (motion)
		diffs=cnt->threshold/4;

	for (i=0; i<THRESHOLD_TUNE_LENGTH-1; i++)
	{
	    sum+=cnt->diffs_last[i];
	    if (cnt->diffs_last[i+1] && !motion)
		cnt->diffs_last[i]=cnt->diffs_last[i+1];
	    else
		cnt->diffs_last[i]=cnt->threshold/4;
	    if (cnt->diffs_last[i]>top)
		top=cnt->diffs_last[i];
	}
	sum+=cnt->diffs_last[i];
	cnt->diffs_last[i]=diffs;
	
	sum/=THRESHOLD_TUNE_LENGTH/4;
	if (sum<top*2)
	    sum=top*2;
	if (sum < cnt->conf.max_changes)
			cnt->threshold=(cnt->threshold+sum)/2;
	printf("New threshold: %d\n", cnt->threshold);
}

/*	The heart of the motion detector
 *	Basicly: ref-new and count the pixels that are over a noise threshold
 */
int alg_diff_standard (struct context *cnt)
{
	struct images *imgs=&cnt->imgs;
	int i, diffs=0;
	long int level=0;
	int noise=cnt->noise;
	char *ref=imgs->ref;
	char *new=imgs->new;
	char *out=imgs->out;
	unsigned char *mask=imgs->mask;


	/* If the average level of the picture is to low, compensate by 
	 * lowering the noise threshold
	 */
	if (cnt->conf.nightcomp) {
		i=imgs->motionsize;
		for (i--; i>=0; i--) {
			level+=(unsigned char)new[i];
		}
		level/=imgs->motionsize;
		if (level < noise*2)
			noise*=2;
	}

	i=imgs->motionsize;
	for (; i>0; i--) {
		*out=*ref-*new;
		if (mask)
			*out=((int)((char)*out**mask++)/255);
		if (*out >  noise ||
		    *out < -noise ) {
			if (!cnt->conf.realmotion) *out=*new;
			diffs++;
		} else {
			if (!cnt->conf.realmotion) *out=0;
		}
		out++;
		ref++;
		new++;
	}

	return diffs;
}


void alg_region_diff(struct context *cnt, struct feature*feat )
{
    struct images *imgs=&cnt->imgs;
    int diffs=0;
    int noise=cnt->noise;
    char *ref=imgs->ref;
    char *new=imgs->new;
    char *out=imgs->out;

    struct region*reg = feat->reg;

    int x,y,endx,endy,ofs,ofs2;
    char df;
    unsigned int e=0;
    endx = reg->inner.x+reg->inner.width;
    endy = reg->inner.y+reg->inner.height;

#if 1 /* SOBEL_EDGE */
    noise=2;
    for (y=reg->inner.y; y<endy;y++) {
	ofs = cnt->imgs.width*y;
	for (x=reg->inner.x; x<endx;x++) { 
	    //df = abs(new[ofs+x] - new[ofs+x-1]);
	    df = abs(new[ofs+x] - new[ofs+x-1]);
	    //df += abs(new[ofs+x] - new[ofs+x-reg->inner.width]);
	    //df/=2;
	    out[ofs+x] = df;
	    df = df ^ ref[ofs+x];
	    if (df >  noise ) {
		diffs++;
		e ++;
	    }
	}
    }
#endif
#if 0 /* LAPLACE_EDGE */
    noise=2;
    for (y=reg->inner.y; y<endy;y++) {
	ofs = cnt->imgs.width*y;
	for (x=reg->inner.x; x<endx;x++) { 
	    int lap;
	    lap = 2*new[ofs+x]-new[ofs+x-1]-new[ofs+x+1]
		-new[ofs+x-reg->inner.width]-new[ofs+x+reg->inner.width];
	    df= (unsigned char)lap;
	    out[ofs+x] = df;
	    //new[ofs+x] = df;

	    if (df >  noise ) {
		diffs++;
		e ++;
	    }
	}
    }
#endif
#if 1
    //    noise=4;
    ofs2=0;
    for (y=reg->inner.y; y<endy;y++) {
	ofs = cnt->imgs.width*y+reg->inner.x;
	for (x=0; x<reg->inner.width; x++) {
	    df = new[ofs+x] - ref[ofs+x];
	    df = df>0 ? df : 0-df; // abs(df)
	    out[ofs+x] = df;
	    //out[ofs+x] = new[ofs+x];
	    
	    if (df >  noise || df < -noise ) {
		diffs++;
		e += abs(df);
		feat->mask[ ofs2 ] = df | (df>>1 ) | (df>>2);
	    } else {
		feat->mask[ ofs2 ] = 0x00;
	    }
	    out[ofs+x]= feat->mask[ ofs2 ];
	    ofs2++;
	}
    }
#endif
    reg->energy = e;
    /*
    if (reg->max<e) { 
	reg->max=e;

	reg->hystTop = reg->max/2; // 2thirds
	reg->hystBot = reg->max/3; // 2thirds
    }
    if (reg->min>e) {
	reg->min=e;
	reg->hystTop = reg->max/2; // 2thirds
	reg->hystBot = reg->max/3; // 2thirds
    }
    */
}
int x=0;

void diff_histo(struct histo* h)
{
    int i;
    unsigned int *c1,*c2;
    unsigned int  d=0;
    c1=h->histo;
    c2=h->back;
    //    x++;
    for (i=0;i<H_BINS;i++) {      // y component
	if (x==10) { printf("%3d ",*c1);}
	d += abs(*c1++ - *c2++);
	//d += (*c1-*c2)*(*c1 - *c2); c1++;c2++;
    }
    h->dy = d;
    d=0;
    if (x==10) { printf("| ");}
    for (i=0;i<H_BINS;i++) {      // y component
	if (x==10) { printf("%3d ",*c1);}
	d += abs(*c1++ - *c2++);
	//d += (*c1-*c2)*(*c1 - *c2); c1++;c2++;
    }
    h->du = d*4;
    d=0;
    if (x==10) { printf("| ");}
    for (i=0;i<H_BINS;i++) {      // y component
	if (x==10) { printf("%3d ",*c1);}
	d += abs(*c1++ - *c2++);
	//d += (*c1-*c2)*(*c1 - *c2); c1++;c2++;
    }
    if (x==10) { printf("\n"); x=0;}
    h->dv = d*4;
}

int prob_histo(int* histo, int y, int u, int v)
{
  
}

void alg_region_histo(struct context *cnt, struct feature *feat )
{
    struct region*reg = feat->reg;
    unsigned int *bins;
    int i,x,y,endx,endy,ofs,ofsy;
    int npixels;
    struct images *imgs=&cnt->imgs;
    unsigned char *new=imgs->new,*u,*v,*vo,*uo; // u,v are the color planes
    unsigned char *out=imgs->out;
    struct histo* h;

    npixels = reg->inner.width*reg->inner.height;

    if (!feat->data) {
	// first time init
	h=calloc(1,sizeof(struct histo));
	feat->data = h;
	//h->histo = calloc(3*H_BINS, sizeof(unsigned int) );
	//h->back = calloc(3*H_BINS, sizeof(unsigned int) );
	h->ty=500; h->tu=500;h->tv=500;
	h->dy=100; h->du=100;h->dv=100;
    } else {
	h = (struct histo*) feat->data;
    }
    bins = h->histo;
    memset(bins, 0, 3*H_BINS*sizeof(unsigned int));

    endx = reg->inner.x+reg->inner.width;
    endy = reg->inner.y+reg->inner.height;
    
    for (y=reg->inner.y; y<endy;y++) {
	ofs = cnt->imgs.width*y + reg->inner.x;

	for (x=reg->inner.x; x<endx;x++) {
	    /*
	    unsigned char c;
	    c = new[ofs];
	    if (c>192) {c=255;} else {c-=64; c*=2;}
	    new[ofs]=c;
	    out[ofs]=c;
	    */
	    bins[new[ofs++]>>H_SHFT]++;
	}
    }
    // compute color histo:
    if (cnt->imgs.type == VIDEO_PALETTE_YUV420P) {
	int maxy,maxu,maxv;

	u = new+cnt->imgs.width*cnt->imgs.height; // start of color planes
	v = u+cnt->imgs.width*cnt->imgs.height/4;

	uo = out+cnt->imgs.width*cnt->imgs.height; // start of color planes
	vo = uo+cnt->imgs.width*cnt->imgs.height/4;

	endy = (reg->inner.y+reg->inner.height)/2;

	for (y=reg->inner.y/2; y<endy;y++) {
	    ofs = ((cnt->imgs.width/2)*y)+(reg->inner.x/2);
	    
	    for (x=0;x<reg->inner.width/2; x++) {

		int c;
		c = u[ofs];
		if (c>192) {c=255;} else {c-=64; c<<=1;}
		u[ofs]=c;
		//	uo[ofs]=0;
		
		c = v[ofs];
		if (c>192) {c=255;} else {c-=64; c<<=1;}
		v[ofs]=c;
		//vo[ofs]=c;

		bins[H_BINS+   (u[ofs]>>H_SHFT) ]++;
		bins[2*H_BINS+ (v[ofs]>>H_SHFT) ]++;
		ofs++;
	    }
	}

	maxy=maxu=maxv=0; 
	for (i=0;i<H_BINS;i++) { 
	    if ( bins[maxy] < bins[i]) maxy=i;
	    if ( bins[H_BINS+maxu] < bins[H_BINS+i]) maxu=i;
	    if ( bins[2*H_BINS+maxv] < bins[2*H_BINS+i]) maxv=i;
	}
	maxy = bins[maxy];
	maxu = bins[H_BINS+maxu];
	maxv = bins[2*H_BINS+maxv];

#define PAINT_HISTOGRAM 0 /* stay at least 12 pixels away from the top, or beware of cores */
#if PAINT_HISTOGRAM
	for (y=reg->inner.y/2-6; y<reg->inner.y/2-4;y++) {   // mark the background for y
	    ofs = ((cnt->imgs.width)*y+(reg->inner.x))/2;
	    
	    for (i=0;i<H_BINS*2;i++) {
		uo[ofs+i] = 127;
		vo[ofs+i] = 127;
	    }
	}

	for (y=reg->inner.y/2-4; y<reg->inner.y/2-2;y++) {   // mark the background for u plane
	    ofs = ((cnt->imgs.width/2)*y)+(reg->inner.x/2);

	    for (i=0;i<H_BINS*2;i++) {
		uo[ofs+i] = (255>>H_SHFT)*i;
		vo[ofs+i] = 127;
       		uo[ofs+i] = bins[H_BINS+i/4]*256/maxu;
	    }
	}

	for (y=reg->inner.y/2-2; y<reg->inner.y/2;y++) {  // mark the background for v plane
	    ofs = ((cnt->imgs.width/2)*y)+(reg->inner.x/2);

	    for (i=0;i<H_BINS*2;i++) {
		uo[ofs+i] = 127;
		vo[ofs+i] = (255>>H_SHFT)*i;
       		vo[ofs+i+0] = bins[H_BINS+i/4]*256/maxv;
	    }
	}

	for (y=reg->inner.y-12; y<reg->inner.y-8;y++) { // paint bins for y histo
	    ofsy = cnt->imgs.width*y+reg->inner.x;
	    
	    for (i=0;i<H_BINS*4;i+=4) {
		out[ofsy+i+0] = bins[i/4]*256/maxy;
		out[ofsy+i+1] = bins[i/4]*256/maxy;
		out[ofsy+i+2] = bins[i/4]*256/maxy;
		out[ofsy+i+3] = bins[i/4]*256/maxy;
	    }
	}
	for (y=reg->inner.y-8; y<reg->inner.y-4;y++) { // paint bins for u histo
	    ofsy = cnt->imgs.width*y+reg->inner.x;
	    
	    for (i=0;i<H_BINS*4;i+=4) {
       		out[ofsy+i+0] = bins[H_BINS+i/4]*256/maxu;
		out[ofsy+i+1] = bins[H_BINS+i/4]*256/maxu;
		out[ofsy+i+2] = bins[H_BINS+i/4]*256/maxu;
		out[ofsy+i+3] = bins[H_BINS+i/4]*256/maxu;
	    }
	}
	for (y=reg->inner.y-4; y<reg->inner.y;y++) { // paint bins for v histo
	    ofsy = cnt->imgs.width*y+reg->inner.x;
	    
	    for (i=0;i<H_BINS*4;i+=4) {
		out[ofsy+i+0] = bins[2*H_BINS+i/4]*256/maxv;
		out[ofsy+i+1] = bins[2*H_BINS+i/4]*256/maxv;
		out[ofsy+i+2] = bins[2*H_BINS+i/4]*256/maxv;
		out[ofsy+i+3] = bins[2*H_BINS+i/4]*256/maxv;
	    }
	}
#endif
    }
}

/*	The heart of the motion detector
 *	Basicly: ref-new and count the pixels that are over a noise threshold
 */
int alg_diff_simon (struct context *cnt)
{
	struct images *imgs=&cnt->imgs;
	int i, diffs=0;
	char *new=imgs->new;
	char *out=imgs->out;
	static int init=1;

	memcpy(out,new,imgs->size);
	if (init>0) {
	    printf("image type %d\n", cnt->imgs.type );
	    printf("regions defined: %d\n",cnt->nregions);
	    init=0;
	}
	
	/* If the average level of the picture is to low, compensate by 
	 * lowering the noise threshold
	 */
/*	if (cnt->conf.nightcomp) {
		i=imgs->motionsize;
		for (i--; i>=0; i--) {
			level+=(unsigned char)new[i];
		}
		level/=imgs->motionsize;
		if (level < noise*2)
			noise*=2;
	}
*/
	for (i=0; i<cnt->nregions; i++) {	    
	    struct feature*feat = &cnt->regions[i]->features[0];

	    while (feat->type){
		feat->knecht(cnt, feat);
		feat++;
	    }
	}
	return diffs;
}

/*
	Very fast diff function, does not do nightcompensation or mask
	overlaying.
*/
int alg_diff_fast (struct context *cnt)
{
	struct images *imgs=&cnt->imgs;
	int i, diffs=0;
	int noise=cnt->noise;
	char *out=imgs->out;
	char *ref=imgs->ref;
	char *new=imgs->new;

	memset(out, 0, imgs->size);
	i=imgs->motionsize;
	for (; i>0; i-=40) {
		*out=*ref-*new;
		if (*out >  noise ||
		    *out < -noise ) {
			if (!cnt->conf.realmotion) *out=*new;
			diffs++;
		} else {
			if (!cnt->conf.realmotion) *out=0;
		}
		out+=40;
		ref+=40;
		new+=40;
	}
	diffs=diffs*40;
	return diffs;
}

/*
	diff_hybrid uses diff_fast to quickly decide if there is anything worth
	sending to diff_standard.
*/
int alg_diff_hybrid (struct context *cnt)
{
	int diffs;
	
	diffs=alg_diff_fast(cnt);
	if (diffs > cnt->conf.max_changes/2)
		diffs=alg_diff_standard(cnt);
	return diffs;
}

int alg_diff (struct context *cnt)
{
    //return alg_diff_hybrid(cnt);
	return alg_diff_simon(cnt);
}


/* Detect a sudden massive change in the picture.
   It is assumed to be the light being switched on or a camera displacement.
   In any way the user doesn't think it is worth capturing.
 */
int alg_lightswitch (struct context *cnt, int diffs, int dev, int pipe, int mpipe)
{
	struct images *imgs=&cnt->imgs;
	int i, j;
	
	/* is 2/3 of the image right?  */
	if (diffs > imgs->width*imgs->height*2) {
		printf("lightswitch detected\n");
		/* wait for the device to settle and update the
		 * reference frame to prevent the next frames
		 * from being detected
		 * if 1/3 of the frame is still moving asume that the camera
		 * has not yet compensated.
		 */
		i=10;
		while(i && (imgs->new=vid_next(cnt, dev, imgs->new, imgs->width, imgs->height))) {
			if (alg_diff(cnt)
			    > imgs->width*imgs->height)
				i=10;
			j=imgs->width*imgs->height;
			for (j--; j>=0; j--) {
				imgs->ref[j]=imgs->new[j];
			}
			event(EVENT_IMAGE, cnt, imgs->new, &pipe, cnt->currenttime);
			event(EVENT_IMAGEM, cnt, imgs->out, &mpipe, cnt->currenttime);
			i--;
		}
		return 0;
	}
	return diffs;
}

int alg_switchfilter (struct context *cnt, int diffs, struct coord *cent)
{
	int linediff=diffs/cnt->conf.height;
//	struct images *imgs=&cnt->imgs;
	char *out=cnt->imgs.out;
	int y, x, line;
	int lines=0, vertlines=0;
	char tmp[80];

	for (y=0; y<cnt->conf.height; y++) {
		line=0;
		for (x=0; x<cnt->conf.width; x++) {
			if (*(out++)) {
				line++;
			}
		}
		if (line>cnt->conf.width/18) {
			vertlines++;
		}
		if (line>linediff*2) {
			lines++;
		}
	}
//	printf("%d %d %d %d  %d\n", lines, vertlines, linediff, lines-vertlines, diffs);
	if (vertlines>cnt->conf.height/10 && lines<vertlines/3 &&
	    (vertlines>cnt->conf.height/4 || lines-vertlines>lines/2)) {
		if (cnt->conf.drawtext_changes) {
			sprintf(tmp, "%d %d", lines, vertlines);
			draw_text(cnt->imgs.new, 10, 10, cnt->conf.height, cnt->conf.width, tmp, cnt->imgs.type);
		}
		return diffs;
	}
	return 0;
}
