/**
 * @file
 * @brief Wlan feature
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-03-02 13:36:18 +0100 (Die, 02 Mär 2004) $
 * $Revision: 669 $
 *
 * @since    Wed Mar 12 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: Wlan.cpp 669 2004-03-02 12:36:18Z rene-cvs $
 */

// for sprintf
#include <stdio.h>
#include "Wlan.h"

#if _DEBUG_FEATURES
string WlanActiveEssidFeature::toString() const
{
	char buffer[256];
	sprintf(buffer, "WlanActiveEssidFeature: '%s', has code %s", getName().c_str(), serialize().c_str());
	return buffer;
}

string WlanActiveMacAddressFeature::toString() const
{
	char buffer[256];
	sprintf(buffer, "WlanActiveMacAddressFeature: '%s', has code %s", getName().c_str(), serialize().c_str());
	return buffer;
}

string WlanActiveSignalLevelFeature::toString() const
{
	char buffer[256];
	sprintf(buffer, "WlanActiveSignalLevelFeature: %f", getVal());
	return buffer;
}

string WlanNumPeersFeature::toString() const
{
	char buffer[64];
	sprintf(buffer, "WlanNumPeersFeature: %ld", getVal());
	return buffer;
}
#endif


WlanActiveModeFeature::WlanActiveModeFeature()
{
	mode = (FeatureKind)(unsigned) (rand_double() * Last);
}

WlanActiveModeFeature::WlanActiveModeFeature(FeatureKind mode)
{
	this->mode = mode;
}

void WlanActiveModeFeature::moveTowards(Feature *f, double factor)
{
	if (rand_double() < factor) {
		this->mode = ((WlanActiveModeFeature*) f)->mode;
	}
}

double WlanActiveModeFeature::getDistance(Feature *f) const
{
	if (this->mode == ((WlanActiveModeFeature*) f)->mode)
		return 0.0;
	else
		return 1.0;
}

double WlanActiveModeFeature::getPosition() const
{
	return (double) mode / (Last - 1);
}

string WlanActiveModeFeature::serialize() const
{
	char buffer [sizeof(int)*8+1];
	sprintf(buffer, "%d", (int) mode);

	return buffer;
}

void WlanActiveModeFeature::unserialize(string value)
{
	mode = (WlanActiveModeFeature::FeatureKind) atoi(value.c_str());
}

#if _DEBUG_FEATURES
string WlanActiveModeFeature::toString() const
{
	char buffer[54];
	string ret = "WlanActiveMode: ";
	switch(mode)
	{
		case Unknown:
			ret += "Unknown";
			break;
		case Master:
			ret += "Master";
			break;
		case Managed:
			ret += "Managed";
			break;
		case AdHoc:
			ret += "AdHoc";
			break;
		default:
			sprintf(buffer, "Unkown value: %d", mode);
			ret += buffer;
	}
	return ret;
}
#endif


Feature* WlanActiveModeFeature::clone() const
{
	return new WlanActiveModeFeature(mode);
}

void WlanActiveModeFeature::aggregate(aggregatelist )
{
}

WlanPeersFeature::WlanPeersFeature(stringcode* code, long* maxlen) : AbstractStringListFeature(code, maxlen)
{
}

WlanPeersFeature::WlanPeersFeature(stringcode* code, long* maxlen, const vector<WlanPeerInfo> *peers) : AbstractStringListFeature(code, maxlen)
{
	if (peers != NULL) {
		stringvector peerMacs;
		for (unsigned int i=0; i<peers->size(); i++) {
			peerMacs.push_back((*peers)[i].mac);
		}
		listval = getCodeForList(&peerMacs);
	}
	// else listval will be empty anyways
}

#if _DEBUG_FEATURES
string WlanPeersFeature::toString() const
{
	string ret = "WlanPeersFeature: '";
	for (unsigned int i = 0; i < getListValues().size(); i++) {
		ret += getListValues()[i];
		ret += ", ";
	}
	ret += "', has code ";
	ret += serialize().c_str();

	return ret;
}
#endif


WlanFeatureProvider::WlanFeatureProvider(providerparams &params) : FeatureProvider("WLAN", &features)
{
	(void) params;	// to get rid of unused variable warnings

	activePeerValid = peerListValid = false;

	featuremap["ActiveEssid"] = WlanFeatureProvider::ActiveEssid;
	featuremap["ActiveMode"] = WlanFeatureProvider::ActiveMode;
	featuremap["ActiveSignalLevel"] = WlanFeatureProvider::ActiveSignalLevel;
	featuremap["ActiveMacAddress"] = WlanFeatureProvider::ActiveMacAddress;
	featuremap["Peers"] = WlanFeatureProvider::Peers;
	featuremap["NumPeers"] = WlanFeatureProvider::NumPeers;

	features.push_back("ActiveEssid");
	features.push_back("ActiveMode");
	features.push_back("ActiveSignalLevel");
	features.push_back("ActiveMacAddress");
	features.push_back("Peers");
	features.push_back("NumPeers");

	activeEssidFeature_maxlen = 0;
	activeMacAddressFeature_maxlen = 0;
	peersFeature_maxlen = 0;

	/** @todo provide a symbian limits header file */
#ifndef _SYMBIAN
	activeSignalLevelFeature_minval = numeric_limits<double>::infinity();
	activeSignalLevelFeature_maxval = -numeric_limits<double>::infinity();
	numPeersFeature_minval = numeric_limits<long>::max();
	numPeersFeature_maxval = numeric_limits<long>::min();
#else
	activeSignalLevelFeature_minval = KMaxReal;
	activeSignalLevelFeature_maxval = KMinReal;
	numPeersFeature_minval = KMaxTInt;
	numPeersFeature_maxval = KMinTInt;
#endif
}

Feature* WlanFeatureProvider::getSample(string name) const
{
	map<string, FeatureKind>::iterator iter = featuremap.find(name);
	if (iter != featuremap.end()) {
		switch ((*iter).second) {
			case ActiveEssid:
				if (activePeerValid) {
					return new WlanActiveEssidFeature(const_cast<stringcode*> (&activeEssidFeature_code), const_cast<long*> (&activeEssidFeature_maxlen), activePeer.essid);
				}
				break;

			case ActiveMode:
				if (activePeerValid) {
					return new WlanActiveModeFeature(activePeer.mode);
				}
				break;

			case ActiveSignalLevel:
				if (activePeerValid) {
					return new WlanActiveSignalLevelFeature(const_cast<double*> (&activeSignalLevelFeature_minval), const_cast<double*> (&activeSignalLevelFeature_maxval), activePeer.signalLevel);
				}
				break;

			case ActiveMacAddress:
				if (activePeerValid) {
					return new WlanActiveMacAddressFeature(const_cast<stringcode*> (&activeMacAddressFeature_code), const_cast<long*> (&activeMacAddressFeature_maxlen), activePeer.mac);
				}
				break;

			case Peers:
				if (peerListValid) {
					return new WlanPeersFeature(const_cast<stringcode*> (&peersFeature_code), const_cast<long*> (&peersFeature_maxlen), &peerList);
				}
				break;

			case NumPeers:
				if (peerListValid) {
					/* fold same macs */
					map<string, bool> peers;
					for (vector<WlanPeerInfo>::const_iterator iter = peerList.begin(); iter != peerList.end(); iter++) {
						peers[(*iter).mac] = true;
					}
					return new WlanNumPeersFeature(const_cast<long*> (&numPeersFeature_minval), const_cast<long*> (&numPeersFeature_maxval), (long) peers.size());
				}
				break;
		}
	}

	return NULL;
}

Feature* WlanFeatureProvider::getFeature(string name) const
{
	map<string, FeatureKind>::iterator iter = featuremap.find(name);
	if (iter != featuremap.end()) {
		switch ((*iter).second) {
			case ActiveEssid:
				return new WlanActiveEssidFeature(const_cast<stringcode*> (&activeEssidFeature_code), const_cast<long*> (&activeEssidFeature_maxlen));
				break;

			case ActiveMode:
				return new WlanActiveModeFeature();
				break;

			case ActiveSignalLevel:
				return new WlanActiveSignalLevelFeature(const_cast<double*> (&activeSignalLevelFeature_minval), const_cast<double*> (&activeSignalLevelFeature_maxval));
				break;

			case ActiveMacAddress:
				return new WlanActiveMacAddressFeature(const_cast<stringcode*> (&activeMacAddressFeature_code), const_cast<long*> (&activeMacAddressFeature_maxlen));
				break;

			case Peers:
				return new WlanPeersFeature(const_cast<stringcode*> (&peersFeature_code), const_cast<long*> (&peersFeature_maxlen));
				break;

			case NumPeers:
				return new WlanNumPeersFeature(const_cast<long*> (&numPeersFeature_minval), const_cast<long*> (&numPeersFeature_maxval));
				break;
		}
	}

	return NULL;
}
