/**
 * @file
 * @brief Wlan feature
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-03-01 12:15:32 +0100 (Mon, 01 Mär 2004) $
 * $Revision: 664 $
 *
 * @since    Wed Mar 12 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: Wlan.h 664 2004-03-01 11:15:32Z rene $
 */

#ifndef _WLAN_H
#define _WLAN_H

#include "../AbstractString.h"
#include "../Numerical.h"

/** This feature encapsulates the ESSID which is currently set for a WLAN
  * network interface or the ESSID which it is currently bound to (if not set
  * explicitely). It is based on the AbstractStringFeature and uses the
  * Levenshtein distance.
  */
class FEATURE_EXPORT_CLASS WlanActiveEssidFeature : public AbstractStringFeature
{
	friend class WlanFeatureProvider;

	protected:
		/** This constructor should not be used to construct feature
		  * objects. They are only created by the respective feature
		  * provider.
		  */
		WlanActiveEssidFeature(stringcode* code, long* maxlen) : AbstractStringFeature(AbstractStringFeature::Levenshtein, code, maxlen) {}
		/** @todo documentation */
		WlanActiveEssidFeature(stringcode* code, long* maxlen, const string essid) : AbstractStringFeature(AbstractStringFeature::Levenshtein, code, maxlen, essid) {}

	public:
		virtual const string getName() const { return "Wlan.ActiveEssid"; }

#if _DEBUG_FEATURES
		virtual string toString() const;
#endif
};

/** This feature encapsulates the ESSID which is currently set for a WLAN
  * network interface or the ESSID which it is currently bound to (if not set
  * explicitely). It is based on the AbstractStringFeature and uses the
  * Levenshtein distance.
  */
class FEATURE_EXPORT_CLASS WlanActiveMacAddressFeature : public AbstractStringFeature
{
	friend class WlanFeatureProvider;

	protected:
		/** This constructor should not be used to construct feature
		  * objects. They are only created by the respective feature
		  * provider.
		  */
		WlanActiveMacAddressFeature(stringcode* code, long* maxlen) : AbstractStringFeature(AbstractStringFeature::None, code, maxlen) {}

		/** This constructor should not be used to construct feature
		  * objects. They are only created by the respective feature
		  * provider.
		  */
		WlanActiveMacAddressFeature(stringcode* code, long* maxlen, const string mac) : AbstractStringFeature(AbstractStringFeature::None, code, maxlen, mac) {}

	public:
		virtual const string getName() const { return "Wlan.ActiveMacAddress"; }

#if _DEBUG_FEATURES
		/** This is only for testing. */
		virtual string toString() const;
#endif
};

/**
 * @todo documentation
 */
class FEATURE_EXPORT_CLASS WlanActiveModeFeature : public Feature
{
	friend class WlanFeatureProvider;

	public:
		/** @todo documentation */
		typedef enum {Unknown, Master, Managed, AdHoc, Last /* don't use */} FeatureKind;

	protected:
		/** This constructor should not be used to construct feature
		  * objects. They are only created by the respective feature
		  * provider.
		  */
		WlanActiveModeFeature();

		/** This constructor should not be used to construct feature
		  * objects. They are only created by the respective feature
		  * provider.
		  */
		WlanActiveModeFeature(FeatureKind mode);

	public:
		virtual double getPosition() const;
		virtual double getDistance(Feature *f) const;
		virtual void moveTowards(Feature *f, double factor);

		virtual string serialize() const;
		virtual void unserialize(string value);

		virtual Feature* clone() const;
		void aggregate(aggregatelist samples);

		virtual FeatureType getType() const { return nominal; }
		virtual const string getName() const { return "Wlan.ActiveMode"; }

		const FeatureKind getVal() const { return mode; }

#if _DEBUG_FEATURES
		/** This is only for testing. */
		virtual string toString() const;
#endif

	private:
		/** @todo documentation */
		FeatureKind mode;
};

/**
 * @todo documentation
 */
class FEATURE_EXPORT_CLASS WlanActiveSignalLevelFeature : public NumericalContinuousFeature
{
	friend class WlanFeatureProvider;

	protected:
		/** This constructor should not be used to construct feature
		  * objects. They are only created by the respective feature
		  * provider.
		  */
		WlanActiveSignalLevelFeature(double* minval, double* maxval) : NumericalContinuousFeature(minval, maxval) {}

		/** This constructor should not be used to construct feature
		  * objects. They are only created by the respective feature
		  * provider.
		  */
		WlanActiveSignalLevelFeature(double* minval, double* maxval, const double level) : NumericalContinuousFeature(minval, maxval, level) {}

	public:
		virtual const string getName() const { return "Wlan.ActiveSignalLevel"; }

#if _DEBUG_FEATURES
		/** This is only for testing. */
		virtual string toString() const;
#endif
};

/**
 * @todo documentation
 */
struct FEATURE_EXPORT_CLASS WlanPeerInfo {
	//@{
	/** @todo documentation */
	string essid;
	string mac;
	WlanActiveModeFeature::FeatureKind mode;
	double signalLevel;
	//@}
};

/**
 * @todo documentation
 */
class FEATURE_EXPORT_CLASS WlanPeersFeature : public AbstractStringListFeature
{
	friend class WlanFeatureProvider;

	protected:
		/** This constructor should not be used to construct feature
		  * objects. They are only created by the respective feature
		  * provider.
		  */
		WlanPeersFeature(stringcode* code, long* maxlen);

		/** This constructor should not be used to construct feature
		  * objects. They are only created by the respective feature
		  * provider.
		  */
		WlanPeersFeature(stringcode* code, long* maxlen, const vector<WlanPeerInfo> *peers);

	public:
		virtual const string getName() const { return "Wlan.Peers"; }

#if _DEBUG_FEATURES
		/** This is only for testing. */
		virtual string toString() const;
#endif
};

/**
 * @todo documentation
 */
class FEATURE_EXPORT_CLASS WlanNumPeersFeature : public NumericalDiscreteFeature
{
	friend class WlanFeatureProvider;

	protected:
		/** This constructor should not be used to construct feature
		  * objects. They are only created by the respective feature
		  * provider.
		  */
		WlanNumPeersFeature(long* minval, long* maxval) : NumericalDiscreteFeature(minval, maxval) {}

		/** This constructor should not be used to construct feature
		  * objects. They are only created by the respective feature
		  * provider.
		  */
		WlanNumPeersFeature(long* minval, long* maxval, const long numPeers) : NumericalDiscreteFeature(minval, maxval, numPeers) {}

	public:
		virtual const string getName() const { return "Wlan.NumPeers"; }

#if _DEBUG_FEATURES
		/** This is only for testing. */
		virtual string toString() const;
#endif
};


/**
 * @todo documentation
 */
class FEATURE_EXPORT_CLASS WlanFeatureProvider : public FeatureProvider
{
	public:
		/** @todo documentation */
		typedef enum {ActiveEssid, ActiveMode, ActiveSignalLevel, ActiveMacAddress, Peers, NumPeers} FeatureKind;

		/** @todo documentation */
		WlanFeatureProvider(providerparams &params);

		virtual Feature* getFeature(string name) const;
		virtual Feature* getSample(string name) const;
		virtual void nextSample(clock_t checkpoint);

	private:
		//@{
		/** @todo documentation */
		stringvector features;
		// ARGH: hack
		mutable map<string, FeatureKind> featuremap;

		stringcode activeEssidFeature_code;
		long activeEssidFeature_maxlen;
		stringcode activeMacAddressFeature_code;
		long activeMacAddressFeature_maxlen;
		double activeSignalLevelFeature_minval;
		double activeSignalLevelFeature_maxval;
		stringcode peersFeature_code;
		long peersFeature_maxlen;
		long numPeersFeature_minval;
		long numPeersFeature_maxval;
		//@}


	protected:
		//@{
		/** @todo documentation */
		WlanPeerInfo activePeer;
		vector<WlanPeerInfo> peerList;
		bool activePeerValid, peerListValid;
		//@}
};

#endif // _WLAN_H
