/**
 * @file
 * @brief Wlan feature
 *
 * Some parts of this code have been taken from iwconfig.c and iwlist.c of the
 * wireless-tools disribution and were modified accordingly. The original
 * license is:
 *
 * This file is released under the GPL license.
 *     Copyright (c) 1997-2002 Jean Tourrilhes <jt@hpl.hp.com>
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-06-04 13:37:10 +0200 (Fre, 04 Jun 2004) $
 * $Revision: 798 $
 *
 * @since    Wed Mar 19 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: WlanLinux.cpp 798 2004-06-04 11:37:10Z rene $
 */

#include <sys/time.h>
extern "C" {
	#include <iwlib.h>
}

#include "WlanLinux.h"

//@{
/** Default properties */
#define WLAN_INTERFACENAME "eth1"
#define WLAN_SCANDELAY 5
//@}

EXPORT_FEATURE(WlanLinuxFeatureProvider);



// we don't use this here - we have our own class
/** @todo clean up */

void WlanFeatureProvider::nextSample(clock_t checkpoint)
{
	(void) checkpoint;	// to get rid of unused variable warnings
}

//@{
/** just some helper functions that are _really_ specific to Linux wireless extensions */

static inline float decodeSignalLevel(iwqual *qual, iwrange *range, bool has_range) {
	assert(qual != NULL);
	assert(!has_range || range != NULL);
 	if(has_range && (qual->level != 0))
 		/* If the statistics are in dBm */
  		if(qual->level > range->max_qual.level)
			return qual->level - 0x100;
		else
			return qual->level;
	else
		return qual->level;
}

static inline bool setEssid(string newEssid, string ifname) {
	string cmdLine = "/sbin/iwconfig " + ifname + " essid " + newEssid;
	return (system(cmdLine.c_str()) == 0);
}

static inline WlanActiveModeFeature::FeatureKind decodeMode(int modeI)
{
	WlanActiveModeFeature::FeatureKind mode;

	// these codes have been taken from iwlib.c (wireless-tools-25)
	switch (modeI)
	{
		case 1: // ad-hoc
			mode = WlanActiveModeFeature::AdHoc;
			break;
		case 2: // managed
			mode = WlanActiveModeFeature::Managed;
			break;
		case 3: // master
			mode = WlanActiveModeFeature::Master;
			break;
		case 0: // auto
		case 4: // repeater
		case 5: // secondary
		case 6: // monitor
		default:
			mode = WlanActiveModeFeature::Unknown;
	}

	return mode;
}
//@}


WlanLinuxFeatureProvider::WlanLinuxFeatureProvider(featureparams& params) :
	WlanFeatureProvider(params)
{
	string interfaceName;

	if (params.count("delay") == 0)
		this->delay = WLAN_SCANDELAY;
	else
		this->delay = atoi(params["delay"].c_str());

	if (params.count("interface") == 0)
		interfaceName = WLAN_INTERFACENAME;
	else
		interfaceName = params["interface"];

	if (!initSocket(interfaceName))
		fprintf(stderr, "Error: unable to connect to wlan subsystem\n");

	// start the scanning thread
	this->start();
}

WlanLinuxFeatureProvider::~WlanLinuxFeatureProvider() throw()
{
	plog("WlanLinuxFeatureProvider called, waiting for thread to terminate\n");
	// stop the scanning thread and wait for it to finish
	this->cancel();
	this->join();

	closeSocket();
}

void WlanLinuxFeatureProvider::nextSample(clock_t checkpoint)
{
	(void) checkpoint;	// to get rid of unused variable warnings

	activePeerValid = false;

	// these calls should only cause marginal delay
	if (getEssid(activePeer.essid) &&
	    getApAddress(activePeer.mac) &&
	    getApSignalLevel(activePeer.signalLevel) &&
	    getMode(activePeer.mode))
	activePeerValid = true;
}

void WlanLinuxFeatureProvider::run() throw()
{
	// don't allow the list to be read by now
	peerListValid = false;

	while (!this->isCancelled()) {
		peerListValid = getApList(peerList, false);
		for (int i=0; i<delay*10 && !this->isCancelled(); i++)
			this->sleep(100);
	}
}

Feature* WlanLinuxFeatureProvider::getSample(string name) const
{
	Guard g(lockApList);
	return WlanFeatureProvider::getSample(name);
}



bool WlanLinuxFeatureProvider::initSocket(string interfaceName) {
	ifname = interfaceName;
	return ((wlsocket = iw_sockets_open()) >= 0);
}

void WlanLinuxFeatureProvider::closeSocket() {
	close(wlsocket);
}

bool WlanLinuxFeatureProvider::getEssid(string& essid) {
	struct iwreq wrq;
	char cStrEssid[IW_ESSID_MAX_SIZE + 1];
	int essid_on;

	/* Get ESSID */
	memset(cStrEssid, 0, IW_ESSID_MAX_SIZE + 1);
	wrq.u.essid.pointer = (caddr_t) cStrEssid;
	wrq.u.essid.length = IW_ESSID_MAX_SIZE + 1;
	wrq.u.essid.flags = 0;
	if(iw_get_ext(wlsocket, const_cast<char*> (ifname.c_str()), SIOCGIWESSID, &wrq) >= 0) {
		essid_on = wrq.u.data.flags;
		if (essid_on)
			essid = cStrEssid;
		else
			essid = "any";
		return true;
	}
	else {
		fprintf(stderr, "Error on iw_get_ext in getEssid\n");
		return false;
	}
}

bool WlanLinuxFeatureProvider::getMode(WlanActiveModeFeature::FeatureKind& mode) {
	struct iwreq wrq;
	int modeI;

	/* Get operation mode */
	if(iw_get_ext(wlsocket, const_cast<char*> (ifname.c_str()), SIOCGIWMODE, &wrq) >= 0) {
		modeI = wrq.u.mode;
		if((modeI < IW_NUM_OPER_MODE) && (modeI >= 0)) {
			// valid
			mode = decodeMode(modeI);
			return true;
		}
		else
			fprintf(stderr, "Error in getMode: mode in unknown range: %d\n", modeI);
	}
	else
		fprintf(stderr, "Error on iw_get_ext in getMode\n");
	return false;
}


bool WlanLinuxFeatureProvider::getApAddress(string& apAddress) {
	struct iwreq wrq;
	sockaddr ap_addr;
	char buffer[18];

	/* Get AP address */
	if(iw_get_ext(wlsocket, const_cast<char*> (ifname.c_str()), SIOCGIWAP, &wrq) >= 0) {
  		memcpy(&ap_addr, &(wrq.u.ap_addr), sizeof (sockaddr));
  		apAddress = iw_pr_ether(buffer, (unsigned char*) ap_addr.sa_data);
  		return true;
	}
	else {
		fprintf(stderr, "Error on iw_get_ext in getApAddress\n");
		return false;
	}
}

bool WlanLinuxFeatureProvider::getApSignalLevel(double& level) {
	iwstats stats;
	//iwrange range;
	bool has_range = false;
	// I simply don't know why using range causes the method to never return.....
	unsigned char tmp[sizeof(iwrange) * 3]; // this should be enough if the library uses a future struct version

	/* Get ranges */
	has_range = (iw_get_range_info(wlsocket, const_cast<char*> (ifname.c_str()), (iwrange*) &tmp) >= 0);
	/* Get stats */
// this is a bit of a hack, but seems to work - when this define is present, we suspect to have the new API
#ifndef IW_FREQ_FIXED
	// old API
	if(iw_get_stats(wlsocket, const_cast<char*> (ifname.c_str()), &stats) >= 0) {
#else
	// this is the new API
	if(iw_get_stats(wlsocket, const_cast<char*> (ifname.c_str()), &stats, (iwrange*) &tmp, has_range) >= 0) {
#endif
		level = decodeSignalLevel(&(stats.qual), (iwrange*) &tmp, has_range);
		return true;
	}
	else {
		fprintf(stderr, "Error on iw_get_ext in getApSignalLevel\n");
		return false;
	}
}


bool WlanLinuxFeatureProvider::getApList(vector<WlanPeerInfo>& apList, bool changeEssid) {
	struct iwreq		wrq;
	char		buffer[IW_SCAN_MAX_DATA];	/* Results */
	struct timeval	tv;				/* Select timeout */
	int			timeout = 5000000;		/* 5s */
	string essid_old;

	if (changeEssid) {
		if (!getEssid(essid_old))
			return false;

		setEssid("any", ifname);
	}

	/* Init timeout value -> 250ms*/
	tv.tv_sec = 0;
	tv.tv_usec = 250000;

	wrq.u.param.flags = IW_SCAN_DEFAULT;
	wrq.u.param.value = 0;

	/* Initiate Scanning */
	if(iw_set_ext(wlsocket, const_cast<char*> (ifname.c_str()), SIOCSIWSCAN, &wrq) < 0) {
		if(errno != EPERM) {
			fprintf(stderr, "Error on iw_set_ext in getApList\n");
			// this interface does not support scanning --> unable to query
			if (changeEssid)
				setEssid(essid_old, ifname);
				return false;
		}

		/* If we don't have the permission to initiate the scan, we may
 		* still have permission to read left-over results.
 		* But, don't wait !!! */
		/* Not cool, it display for non wireless interfaces... */
		//fprintf(stderr, "%-8.8s  (Could not trigger scanning, just reading left-over results)\n", ifname.c_str());
		tv.tv_usec = 0;
	}
	timeout -= tv.tv_usec;

	while(1) {
		fd_set rfds;
		int last_fd;
		int ret;

		FD_ZERO(&rfds);
		last_fd = -1;

		/* Wait until something happens */
		ret = select(last_fd + 1, &rfds, NULL, NULL, &tv);

		/* Check if there was an error */
		if(ret < 0) {
			if(errno == EAGAIN || errno == EINTR)
				continue;

			if (changeEssid)
				setEssid(essid_old, ifname);
			return false;
		}

		/* Check if there was a timeout */
		if(ret == 0) {
			/* Try to read the results */
			wrq.u.data.pointer = buffer;
			wrq.u.data.flags = 0;
			wrq.u.data.length = sizeof(buffer);
			if(iw_get_ext(wlsocket, const_cast<char*> (ifname.c_str()), SIOCGIWSCAN, &wrq) < 0) {
				/* Check if results not available yet */
				if(errno == EAGAIN) {
					/* Restart timer for only 100ms*/
					tv.tv_sec = 0;
					tv.tv_usec = 100000;
					timeout -= tv.tv_usec;
					if(timeout > 0)
						continue;	/* Try again later */
				}

				/* Bad error */
				/*fprintf(stderr, "%-8.8s  Failed to read scan data : %s\n\n",
				ifname.c_str(), strerror(errno));*/

				if (changeEssid)
					setEssid(essid_old, ifname);
				return false;
			}
			else
				/* We have the results, go to process them */
				break;
		}
	}

	// ok - we have results now, write them to the list
	{ // guard block
		Guard g(lockApList);

		apList.clear();
		if(wrq.u.data.length) {
			struct iw_event iwe;
			struct stream_descr	stream;
			int ap_num = 0;
			int ret;
			//struct iw_range range;
			unsigned char tmp[sizeof(iw_range)*3];
			int has_range;
			char buffer3[18];

			has_range = (iw_get_range_info(wlsocket, const_cast<char*> (ifname.c_str()), (iw_range*) &tmp) >= 0);
			iw_init_event_stream(&stream, buffer, wrq.u.data.length);

			do {
				/* Extract an event and print it */
#if WIRELESS_EXT >= 16
				ret = iw_extract_event_stream(&stream, &iwe, 16);
#else
				ret = iw_extract_event_stream(&stream, &iwe);
#endif
				if(ret > 0) {
					// decode the event now

					switch(iwe.cmd) {

						case SIOCGIWAP:
							apList.push_back(WlanPeerInfo());
							apList[ap_num].mac = iw_pr_ether(buffer3, (unsigned char*) iwe.u.ap_addr.sa_data);
							ap_num++;
							break;

						case SIOCGIWESSID: {
							char essid[IW_ESSID_MAX_SIZE+1];
							if((iwe.u.essid.pointer) && (iwe.u.essid.length))
								memcpy(essid, iwe.u.essid.pointer, iwe.u.essid.length);
							essid[iwe.u.essid.length] = 0;
							if(iwe.u.essid.flags)
								apList[ap_num-1].essid = essid;
							else
								apList[ap_num-1].essid = "any";
						}
						break;

						case IWEVQUAL:
							apList[ap_num-1].signalLevel = decodeSignalLevel(&iwe.u.qual, (iw_range*) &tmp, has_range);
							break;

						case SIOCGIWMODE:
							apList[ap_num-1].mode = decodeMode(iwe.u.mode);
							break;
					}
				}
			} while(ret > 0);

			if (changeEssid)
				setEssid(essid_old, ifname);
			return true;
		}
		else {
			if (changeEssid)
				setEssid(essid_old, ifname);
			// this is still successful - the list is just empty, but valid
			return true;
		}

	} // guard block
}
