/**
 * @file
 * @brief Wlan feature
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-02-12 21:08:39 +0100 (Don, 12 Feb 2004) $
 * $Revision: 612 $
 *
 * @since    Wed Mar 12 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: WlanWindows.cpp 612 2004-02-12 20:08:39Z harald $
 */

#define POLARITY
#ifndef _WIN32_WCE
#define _WIN32_DCOM
#endif

#include "Wlan.h"
#include <wbemcli.h>

//@{
/** WMI handles */

static IWbemLocator *pLoc;
static IWbemServices *pSvc;
//@}

EXPORT_FEATURE(WlanFeatureProvider);

/** Library entry point */
BOOL APIENTRY DllMain(HANDLE hModule, DWORD  fdwReason, LPVOID lpReserved)
{
	switch (fdwReason) { 
		case DLL_PROCESS_ATTACH:
			if (FAILED(CoInitializeEx(0, COINIT_MULTITHREADED))) {
				return FALSE;
			} 

			pLoc = NULL;
			pSvc = NULL;

			// connect to wmi
			if (FAILED(CoCreateInstance(CLSID_WbemLocator, 0, CLSCTX_INPROC_SERVER, IID_IWbemLocator, (LPVOID *) &pLoc))) {
				return FALSE;
			}

			// Connect to the root\default namespace with the current user.
			if (FAILED(pLoc->ConnectServer((L"ROOT\\WMI"), NULL, NULL, 0, NULL, 0, 0, &pSvc))) {
				return FALSE;
			}

			// Set the proxy so that impersonation of the client occurs.
			if (FAILED(CoSetProxyBlanket(pSvc,
						RPC_C_AUTHN_WINNT,
						RPC_C_AUTHZ_NONE,
						NULL,
						RPC_C_AUTHN_LEVEL_CALL,
						RPC_C_IMP_LEVEL_IMPERSONATE,
						NULL,
						EOAC_NONE))) {
				return FALSE;
			}

			library_initialize();
			break;

		case DLL_PROCESS_DETACH:
			library_finalize();

			/* ??? */
			/* if (pSvc != NULL) {
				pSvc->Release();
				pSvc = NULL;
			}
			if (pLoc != NULL) {
				pLoc->Release();
				pLoc = NULL;
			}*/

			CoUninitialize();
			break;
	}

	return TRUE;
}

//@{
/** Windows specific functions */

static void getMode(IWbemClassObject *bssid, WlanPeerInfo *info)
{
	VARIANT var;

	if (bssid->Get(L"Ndis80211InfrastructureMode", 0, &var, NULL, NULL) == WBEM_S_NO_ERROR) {
		IWbemClassObject *inf = (IWbemClassObject *) var.punkVal;
		if (inf->Get(L"Ndis80211NetworkInfrastructure", 0, &var, NULL, NULL) == WBEM_S_NO_ERROR) {
			if (var.iVal) {
				info->mode = WlanActiveModeFeature::Managed;
			} else {
				info->mode = WlanActiveModeFeature::AdHoc;
			}
		}
		inf->Release();
	} else {
		info->mode = WlanActiveModeFeature::Unknown;
	}
}

static void getMac(IWbemClassObject *bssid, WlanPeerInfo *info)
{
	VARIANT var;

	if (bssid->Get(L"Ndis80211MacAddress", 0, &var, NULL, NULL) == WBEM_S_NO_ERROR) {
		long lb, ub;
		int mac[6] = { 0 };
		char str[13];

		SafeArrayGetLBound(var.parray, 1, &lb);
		SafeArrayGetUBound(var.parray, 1, &ub);
		for (; lb <= ub; lb++) {
			SafeArrayGetElement(var.parray, &lb, &mac[lb]);
		}
		sprintf(str, "%02x%02x%02x%02x%02x%02x", mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
		info->mac = str;
		VariantClear(&var);
	} else {
		info->mac.clear();
	}
}

static void getSignal(IWbemClassObject *bssid, WlanPeerInfo *info)
{
	VARIANT var;

	if (bssid->Get(L"Ndis80211ReceivedSignalStrength", 0, &var, NULL, NULL) == WBEM_S_NO_ERROR) {
		info->signalLevel = (double) var.iVal / -100;
	} else {
		info->signalLevel = 0.0;
	}
}

static void getEssid(IWbemClassObject *bssid, WlanPeerInfo *info, bool active)
{
	VARIANT var;

	if (bssid->Get(L"Ndis80211SsId", 0, &var, NULL, NULL) == WBEM_S_NO_ERROR) {
		long lb, ub;
		SafeArrayGetLBound(var.parray, 1, &lb);
		SafeArrayGetUBound(var.parray, 1, &ub);

		char *ssid = (char *) malloc(ub + 2);

		for (; lb <= ub; lb++) {
			SafeArrayGetElement(var.parray, &lb, &ssid[lb]);
		}
		ssid[lb] = '\0';
		VariantClear(&var);

		info->essid = active ? ssid + 4 : ssid;

		free(ssid);
	} else {
		info->essid.clear();
	}
}
//@}


void WlanFeatureProvider::nextSample(clock_t checkpoint)
{
	IEnumWbemClassObject *pList;
	IWbemClassObject *bssidl;
	ULONG arg;

	this->activePeerValid = false;
	this->peerListValid = false;

	if (SUCCEEDED(pSvc->CreateInstanceEnum(L"MSNdis_80211_InfrastructureMode", WBEM_FLAG_RETURN_IMMEDIATELY|WBEM_FLAG_FORWARD_ONLY, NULL, &pList))) {
		if (pList->Next(WBEM_INFINITE, 1, &bssidl, &arg) == WBEM_S_NO_ERROR) {
			getMode(bssidl, &activePeer);
			bssidl->Release();
		}
		pList->Release();
	}

	if (SUCCEEDED(pSvc->CreateInstanceEnum(L"MSNdis_80211_BaseServiceSetIdentifier", WBEM_FLAG_RETURN_IMMEDIATELY|WBEM_FLAG_FORWARD_ONLY, NULL, &pList))) {
		if (pList->Next(WBEM_INFINITE, 1, &bssidl, &arg) == WBEM_S_NO_ERROR) {
			getMac(bssidl, &activePeer);
			bssidl->Release();
		}
		pList->Release();
	}

	if (SUCCEEDED(pSvc->CreateInstanceEnum(L"MSNdis_80211_ReceivedSignalStrength", WBEM_FLAG_RETURN_IMMEDIATELY|WBEM_FLAG_FORWARD_ONLY, NULL, &pList))) {
		if (pList->Next(WBEM_INFINITE, 1, &bssidl, &arg) == WBEM_S_NO_ERROR) {
			getSignal(bssidl, &activePeer);
			bssidl->Release();
		}
		pList->Release();
	}

	if (SUCCEEDED(pSvc->CreateInstanceEnum(L"MSNdis_80211_ServiceSetIdentifier", WBEM_FLAG_RETURN_IMMEDIATELY|WBEM_FLAG_FORWARD_ONLY, NULL, &pList))) {
		if (pList->Next(WBEM_INFINITE, 1, &bssidl, &arg) == WBEM_S_NO_ERROR) {
			getEssid(bssidl, &activePeer, true);
			bssidl->Release();
		}
		pList->Release();
	}

	this->activePeerValid = true;

	this->peerList.clear();

	// get ssi
	if (SUCCEEDED(pSvc->CreateInstanceEnum(L"MSNdis_80211_BSSIList", WBEM_FLAG_RETURN_IMMEDIATELY|WBEM_FLAG_FORWARD_ONLY, NULL, &pList))) {
		while (pList->Next(WBEM_INFINITE, 1, &bssidl, &arg) == WBEM_S_NO_ERROR) {
			if (arg) {
				VARIANT arr;
				long lBound, uBound;

				bssidl->Get(L"Ndis80211BSSIList", 0, &arr, NULL, NULL);
				if (arr.vt != VT_NULL) {
					SafeArrayGetLBound(arr.parray, 1, &lBound);
					SafeArrayGetUBound(arr.parray, 1, &uBound);
					for (;lBound <= uBound; lBound++) {
						WlanPeerInfo info;
						IWbemClassObject *bssid;
			
						SafeArrayGetElement(arr.parray, &lBound, &bssid);

						getMode(bssid, &info);
						getMac(bssid, &info);
						getSignal(bssid, &info);
						getEssid(bssid, &info, false);
				
						bssid->Release();

						peerList.push_back(info);
					}
				}

				VariantClear(&arr);
				bssidl->Release();
				this->peerListValid = true;
			}
		}
		pList->Release();
	}
}