/**
 * @file
 * @brief Active lezi predictor
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-08-29 12:14:03 +0200 (Son, 29 Aug 2004) $
 * $Revision: 905 $
 *
 * @since    Tue Dec 16 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: ActiveLezi.cpp 905 2004-08-29 10:14:03Z rene $
 */

#include "ActiveLezi.h"

EXPORT_PREDICTOR(ActiveLezi);

ActiveLezi::ActiveLezi(predictorparams &params)
{
	// no initialization to do, it is done in the respective constructors
	max_lz_length = 0;

	if (params.count("depth") == 0) {
		maxWindowModelDepth = 50;
	} else {
		maxWindowModelDepth = atoi(params["depth"].c_str());
	}
}

ActiveLezi::~ActiveLezi()
{
}

void ActiveLezi::addContexts(const membershiplist* contexts, time_t time)
{
	(void) contexts;
	(void) time;
	throw 1;
}

void ActiveLezi::addContext(unsigned long contextId, time_t time)
{
	(void) time;

	if (find(seen_context_ids.begin(), seen_context_ids.end(), contextId) == seen_context_ids.end())
		seen_context_ids.push_back(contextId);

	phrase.push_back(contextId);
	// remember the current context for the prediction
	curContext = phrase;
	if (! dictionary.findSequence(&phrase)) {
		dictionary.insertSequence(&phrase, false);
		if (phrase.size() > max_lz_length)
			max_lz_length = phrase.size();
		if (max_lz_length > maxWindowModelDepth)
			max_lz_length = maxWindowModelDepth;
		phrase.clear();
	}

	window.push_back(contextId);
	if (window.size() > max_lz_length)
		(void) window.pop_front();
	// and now go through the window and remember all sequences with the last value and all possible contexts
	list<unsigned long> temp = window;
	while (temp.size() > 0) {
		dictionary.insertSequence(&temp, true);
		(void) temp.pop_front();
	}
}

unsigned long ActiveLezi::getNextContext(const list<unsigned long>* context) const
{
	double maxProbability = 0.0;
	unsigned long maxId = 0;

	for (list<unsigned long>::const_iterator id = seen_context_ids.begin(); id != seen_context_ids.end(); id++) {
		// this is _very_ inefficient
		double prob = dictionary.calculateProbability(context, *id);
		if (prob > maxProbability) {
			maxProbability = prob;
			maxId = *id;
		}
	}
	return maxId;
}

unsigned long ActiveLezi::getNextContext() const
{
	return getNextContext(&curContext);
}

membershiplist ActiveLezi::getNextContexts() const
{
	membershiplist ret;
	for (list<unsigned long>::const_iterator id = seen_context_ids.begin(); id != seen_context_ids.end(); id++)
		ret[*id] = dictionary.calculateProbability(&curContext, *id);

	return ret;
}

contexttrajectory ActiveLezi::getContextTrajectory(unsigned int start, unsigned int end) const
{
	assert(start <= end);
	contexttrajectory ret;
	// this is the (increasing) ocntext we work with
	list<unsigned long> tempContext = curContext;

	for (unsigned int step=0; step<end; step++) {
		/* now, for each prediction step, do the normal forecast, but
		   with our temporary context - just work with that, don't
		   modify the dictionary */
		unsigned long predId = getNextContext(&tempContext);
		// check if this time step is in the requested area
		if (step >= start)
			ret.push_back(predId);

		// and advance the context
		tempContext.push_back(predId);

		/* but take care when we don't know the context in our current
		   dictionary - then we will have to start from the beginning
		   (this mimics the behavior in addContext) */
		// --> 0.879206 accuracy for 30 step prediction with testcases/clusters.txt-test4
/*		if (!dictionary.findSequence(&tempContext))
			tempContext.clear();*/
		// try something else....
		// maybe a substring is in the dictionary
		// --> 0.896524 accuracy for 30 step prediction with testcases/clusters.txt-test4
		while (!dictionary.findSequence(&tempContext))
			(void) tempContext.pop_front();
	}

	return ret;
}

unsigned long ActiveLezi::getContextAt(time_t time) const
{
	(void) time;

	throw 1;
}

membershiplist ActiveLezi::getContextsAt(time_t time) const
{
	(void) time;

	throw 1;
}

string ActiveLezi::serialize() const
{
	throw 1;
}

void ActiveLezi::unserialize(string data)
{
	(void) data;

	throw 1;
}

unsigned int ActiveLezi::getMaxDepth() const
{
	return max_lz_length;
}

#if _DEBUG_PREDICTORS
string ActiveLezi::toString() const
{
	string ret;
	char buf[500];

	ret += "Phrase: '";
	for (list<unsigned long>::const_iterator iter=phrase.begin(); iter!=phrase.end(); iter++) {
		sprintf(buf, "%ld (%c), ", *iter, (char) *iter);
		ret += buf;
	}
	ret += "'\n";
	ret += "Window: '";
	for (list<unsigned long>::const_iterator iter=window.begin(); iter!=window.end(); iter++) {
		sprintf(buf, "%ld (%c), ", *iter, (char) *iter);
		ret += buf;
	}
	ret += "'\n";
	ret += "Current Context: '";
	for (list<unsigned long>::const_iterator iter=curContext.begin(); iter!=curContext.end(); iter++) {
		sprintf(buf, "%ld (%c), ", *iter, (char) *iter);
		ret += buf;
	}
	ret += "'\n";
	sprintf(buf, "Max_LZ_Length: %d\n", max_lz_length);
	ret += buf;

	ret += dictionary.toString();

	return ret;
}
#endif
