/**
 * @file
 * @brief Hidden Markov Model (HMM) predictor
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-08-30 22:03:46 +0200 (Mon, 30 Aug 2004) $
 * $Revision: 911 $
 *
 * @since    Tue Dec 16 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: HMM.cpp 911 2004-08-30 20:03:46Z rene $
 */

#include "HMM.h"

EXPORT_PREDICTOR(HMM);

HMM::HMM(predictorparams &params)
{
	throw 1;
}

// helper function to create a random valued, but normalized to a sum of 1.0, array of the given length
double* createRandNormArray(unsigned int length) 
{
	double* ret = new double[length];
	double sum = 0.0;
	for (unsigned int i=0; i<length; i++) {
		ret[i] = rand_double();
		sum += ret[i];
	}
	for (unsigned int i=0; i<length; i++) 
		ret[i] /= sum;
	return ret;
}

HMM::HMM(unsigned int numHiddenStates, unsigned int numAlphabetSymbols)
{
/*	if (params.count("window_length") == 0) {
		windowLength = 20;
	} else {
		windowLength = atoi(params["window_length"].c_str());
	}
	
	if (params.count("hidden_states") == 0) {
		numHiddenStates = 10;
	} else {
		numHiddenStates = atoi(params["hidden_states"].c_str());
	}
	if (params.count("alphabet_size") == 0) {
		numAlphabetSymbols = 10;
	} else {
		numAlphabetSymbols = atoi(params["alphabet_size"].c_str());
	}*/
	
	hmm = new GHMM_DiscreteModel(numHiddenStates, numAlphabetSymbols);
	
	// it seems that we have to initialize the model "manually" here with random values....
	double* pi = createRandNormArray(numHiddenStates);
	double** a = new double*[numHiddenStates];
	double** b = new double*[numHiddenStates];
	for (unsigned int si=0; si<numHiddenStates; si++) {
		GHMM_State* s = hmm->getState(si);
		s->setInitialProbability(pi[si]);
		// now fill the A matrix
		a[si] = createRandNormArray(numHiddenStates);
		// this is a fully connected model !
		for (unsigned sj=0; sj<numHiddenStates; sj++)
			hmm->setTransition(si, sj, a[si][sj]);
		delete a[si];
		// now the B matrix
		b[si] = createRandNormArray(numAlphabetSymbols);
		for (unsigned oj=0; oj<numAlphabetSymbols; oj++)
			s->setOutputProbability(oj, b[si][oj]);
		delete b[si];
	}
	delete pi;
	delete a;
	delete b;
	
	if (hmm->check())
		printf("Error checking model\n");
}

HMM::~HMM()
{
	if (hmm != NULL)
		delete hmm;
}

void HMM::addContextTrajectory(const vector<unsigned int>& traj, unsigned int from, unsigned int to, unsigned int slidingWindowSize, unsigned int slidingWindowHops)
{
	assert(slidingWindowSize > 0);
	assert(slidingWindowHops > 0);
	printf("Reestimating HMM model with trajectory of size %d, from %d to %d, sliding window size %d, %d hops for the sliding window\n", traj.size(), from, to, slidingWindowSize, slidingWindowHops);
	GHMM_Sequence seq(GHMM_INT, slidingWindowSize);
	GHMM_Sequences seqs(GHMM_INT);
	// slide a window over the whole sequence and use extracts from it for actual training
	for (unsigned int off=from; off<to-slidingWindowSize; off+=slidingWindowHops) {
		for (unsigned int i=0; i<slidingWindowSize; i++) {
			seq.setInt(i, traj[from+off+i]);
			// remember which contexts have been seen
			if (find(seenContexts.begin(), seenContexts.end(), traj[i]) == seenContexts.end())
				seenContexts.push_back(traj[i]);
		}
		seqs.add(&seq);
	}
	printf("Training with %d sequences of length %d:\n", seqs.getNumberOfSequences(), seqs.getLength(0));
/*	int* s = seqs.getIntSequence(0);
	for (unsigned int i=0; i<seqs.getLength(0); i++)
		printf("%d ", s[i]);
	printf("\n");*/

	if (hmm->check())
		printf("Error checking model\n");
/*	hmm->A_print(stdout, "\t", " ", "\n");
	hmm->B_print(stdout, "\t", " ", "\n");
	hmm->fix_print(stdout, "\t", " ", "\n");
	hmm->Pi_print(stdout, "\t", " ", "\n");
	hmm->states_print(stdout);*/
	hmm->print(stdout);
	
	if (hmm->reestimate_baum_welch(&seqs) != 0)
		throw 1;
		
	// remember the last K contexts for the context window
/*	for (unsigned int i=to-windowLength; i<to; i++)
		curContextWindow.push_back(traj[i]);
	curContextWindow.pop_front();*/
}

void HMM::addContexts(const membershiplist* contexts, time_t time)
{
	(void) contexts;
	(void) time;
	throw 1;
}

void HMM::addContext(unsigned long contextId, time_t time)
{
	(void) time;
	(void) contextId;
	throw 1;
}

unsigned long HMM::getNextContext() const
{
	throw 1;
}

// HACK HACK HACK - this should be incorporated into getNextContexts !
vector<unsigned long> HMM::getNextContextsMultiStep(const vector<unsigned int>& window, unsigned int numSteps) const 
{
	GHMM_Sequence seq(GHMM_INT, window.size()+numSteps);
	vector<unsigned long> bestPred(numSteps, 0), curPred(numSteps, 0);
	double bestProb=-1, curProb;
	unsigned int i=0;
	vector<list<unsigned long>::const_iterator> iters(numSteps);
	bool finished=false;
	
	if (seenContexts.size() == 0) {
		// nothing to do - nothing learned yet !
		return bestPred;
	}
	
	for (unsigned int j=0; j<numSteps; j++) 
		iters[j] = seenContexts.begin();

	for (vector<unsigned int>::const_iterator iter=window.begin(); iter!=window.end(); iter++) {
		seq.setInt(i++, *iter);
	}

	// pretty standard backtracking here....
	do {
		// evaluate the current combination of contexts (described by the iterators)
#ifdef _DEBUG_HMM_MULTISTEP
		printf("HMM Multistep prediction: trying sequence (out of %d seen contexts) ", seenContexts.size());
#endif
		for (unsigned int j=0; j<numSteps; j++) {
			curPred[j] = *(iters[j]);
#ifdef _DEBUG_HMM_MULTISTEP
			printf("%ld ", curPred[j]);
#endif
			seq.setInt(window.size()+j, curPred[j]);
		}
		GHMM_Sequences seqs(&seq);
		//GHMM_IntVector* viterbiPath = hmm->viterbi(&seqs, 0, &curProb);
		//delete viterbiPath;
		if (hmm->foba_logp(&seqs, 0, &curProb) != 0) {
			printf("Houston, we have a problem! The HMM says it can't generate the given sequence at all!\n");
		}
#ifdef _DEBUG_HMM_MULTISTEP
		printf(" --> %f", curProb);
#endif
	
		// is this the currently best known ?
		if (curProb > bestProb) {
			bestProb = curProb;
			bestPred = curPred;
#ifdef _DEBUG_HMM_MULTISTEP
			printf(": new best with probability %f\n", bestProb);
#endif
		}
#ifdef _DEBUG_HMM_MULTISTEP
		else
			printf("\n");
#endif
		
		// "count up" - this is guaranteed to work because of the if we are in
		int k=numSteps-1;
		iters[k]++;
		while (iters[k] == seenContexts.end() && k>0) {
			assert(iters[k] == seenContexts.end());
			iters[k] = seenContexts.begin();
			k--;
			assert(k >= 0);
			iters[k]++;
		}
		if (iters[0] == seenContexts.end())
			finished = true;
	} while (!finished);
	
#ifdef _DEBUG_HMM_MULTISTEP
	printf("++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
#endif
	
	return bestPred;
}

unsigned long HMM::getNextContext(const vector<unsigned int>& window) const
{
	membershiplist contexts = getNextContexts(window);
	unsigned long bestContext = 0;
	double bestProb = -numeric_limits<double>::max();
	for (membershiplist::const_iterator c = contexts.begin(); c != contexts.end(); c++) {
		if (c->second > bestProb) {
			bestProb = c->second;
			bestContext = c->first;
		}
	}
/*	printf("Found best context %d with probability %f\n", bestContext, bestProb);*/

	return bestContext;
}

membershiplist HMM::getNextContexts() const
{
	throw 1;
}

membershiplist HMM::getNextContexts(const vector<unsigned int>& window) const
{
	membershiplist ret;

//	GHMM_Sequence seq(GHMM_INT, curContextWindow.size()+1);
	GHMM_Sequence seq(GHMM_INT, window.size()+1);
	unsigned int i=0;
//	for (list<unsigned long>::const_iterator iter=curContextWindow.begin(); iter!=curContextWindow.end(); iter++)
/*	printf("Trying with sequence: ");*/
	for (vector<unsigned int>::const_iterator iter=window.begin(); iter!=window.end(); iter++) {
/*		printf("%d ", *iter);*/
		seq.setInt(i++, *iter);
	}
/*	printf(": ");*/

	for (list<unsigned long>::const_iterator iter=seenContexts.begin(); iter!=seenContexts.end(); iter++) {
		double prob;
//		seq.setInt(curContextWindow.size(), *iter);
		seq.setInt(window.size(), *iter);
		GHMM_Sequences seqs(&seq);
		GHMM_IntVector* viterbiPath = hmm->viterbi(&seqs, 0, &prob);
		delete viterbiPath;
/*		printf("%d->%f ", *iter, prob);*/
		ret[*iter] = prob;
	}
/*	printf("\n");*/
	
	return ret;
}

contexttrajectory HMM::getContextTrajectory(unsigned int start, unsigned int end) const
{
	assert(start <= end);
	contexttrajectory ret;

	return ret;
}

unsigned long HMM::getContextAt(time_t time) const
{
	(void) time;

	throw 1;
}

membershiplist HMM::getContextsAt(time_t time) const
{
	(void) time;

	throw 1;
}

string HMM::serialize() const
{
	throw 1;
}

void HMM::unserialize(string data)
{
	(void) data;

	throw 1;
}

#if _DEBUG_PREDICTORS
string HMM::toString() const
{
	return "";
}
#endif
