/**
 * @file
 * @brief %Predictor interface
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-02-12 21:08:39 +0100 (Don, 12 Feb 2004) $
 * $Revision: 612 $
 *
 * @since    Wed Mar 19 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: Predictor.h 612 2004-02-12 20:08:39Z harald $
 */

#ifndef _PREDICTOR_H
#define _PREDICTOR_H

#ifdef _PREDICTOR_EXPORT
# define _FEATURE_EXPORT
#endif
#ifdef _PREDICTOR_EXPORT_CLASS
# define _FEATURE_EXPORT_CLASS
#endif

#include "../Features/Feature.h"

//@{
/** compiler specific flags to export functions and classes */
#define PREDICTOR_EXPORT FEATURE_EXPORT
#define PREDICTOR_EXPORT_CLASS FEATURE_EXPORT_CLASS
//@}

/**
 * Export a predictor
 *
 * This macro has to be used in global scope and gets substituted with the
 * implementation for exporting a PredictorAlgorithm singleton instance to the
 * framework.
 *
 * @param predictor The Class that implements the PredictorAlgorithm interface
 *                 and should be exported to the framework.
 */
#define EXPORT_PREDICTOR(predictor) \
	/** Singleton implementing the PredictorAlgorithm interface */ \
	static predictor *pred; \
	void INIT_EXPORT library_initialize() { pred = NULL; } \
	void FINI_EXPORT library_finalize() { if (pred != NULL) delete pred; } \
	/** \
     * Returns an instance of a class implementing the PredictorAlgorithm interface \
	 * \
	 * @param params A map of parameters to initialize the instance \
	 * @return A PredictorAlgorithm implementation \
	 */ \
	extern "C" PREDICTOR_EXPORT PredictorAlgorithm* getPredictor(predictorparams &params) { if (pred == NULL) pred = new predictor(params); return pred; }

/**
 * Parameters passed to predictors
 *
 * When initializing a predictor, an arbitrary number of named (string)
 * parameters can be passed from the application (e.g. read from a config file).
 * These key-value tuples can then be used by the predictor as
 * configuration or initialization values. They are completely predictor
 * specific.
 *
 * @see Predictor
 * @see PredictorAlgorithm
 */
typedef parametermap predictorparams;

/** @todo documentation */
typedef vector<unsigned long> contexttrajectory;

/**
 * Predictor Interface
 *
 * @todo documentation
 */
class PredictorAlgorithm {
	public:
		virtual ~PredictorAlgorithm() {}

		// add a single step to the model
		/** @todo documentation */
		virtual void addContexts(const membershiplist* contexts, time_t time) = 0;
		/** @todo documentation */
		virtual void addContext(unsigned long contextId, time_t time) = 0;

		// this is single-step prediction
		/** @todo documentation */
		virtual unsigned long getNextContext() const = 0;
		/** @todo documentation */
		virtual membershiplist getNextContexts() const = 0;

		// this is multi-step prediction
		// the parameters ars start and end (in number of time steps) of the trajectory that should be predicted
		// if no offset is necessary (the trajectory should be predicted from the current context), give 0 as start
		// the returned vector will have a length of (end - start)
		/** @todo documentation */
		virtual contexttrajectory getContextTrajectory(unsigned int start, unsigned int end) const = 0;

		// this is a prediction for specific points in time
		// this will just return 0 if no prediction can be made (because no context is known until now)
		/** @todo documentation */
		virtual unsigned long getContextAt(time_t time) const = 0;
		// this will return an empty list if no prediction can be made at all
		/** @todo documentation */
		virtual membershiplist getContextsAt(time_t time) const = 0;

		/**
		 * Serialize a predictors data to a string
		 *
		 * @return String representation of the predictor data.
		 */
		virtual string serialize() const = 0;

		/**
		 * Unserialize a predictors data from a string
		 *
		 * @param data String representation of the predictor data. 
		 */
		virtual void unserialize(string data) = 0;

#if _DEBUG_PREDICTORS
		/** This is only for testing. */
		virtual string toString() const = 0;
#endif
};

/**
 * Predictor Wrapper
 *
 * The Predictor Class is a wrapper class that loads
 * a dll containing the actual implementation during
 * runtime.
 */
class Predictor : PredictorAlgorithm {
	private:
		/** Implementation */
		PredictorAlgorithm *pa;

#if defined(_WIN32_WCE) || defined (_WINDOWS)
		HMODULE dll;
#else
		/** Library handle */
		void* dlHandle;
#endif

	public:
		/**
		 * Creates the classifier
		 *
		 * @param name The name of the requested classifier implementation.
		 * @param params A map of parameters to initialize the classifier.
		 */
		Predictor(const string &name, const predictorparams &params);
		/**
		 * Destructor
		 */
		virtual ~Predictor();

		virtual void addContexts(const membershiplist* contexts, time_t time) { pa->addContexts(contexts, time); }
		virtual void addContext(unsigned long contextId, time_t time) { pa->addContext(contextId, time); }
		virtual unsigned long getNextContext() const { return pa->getNextContext(); }
		virtual membershiplist getNextContexts() const { return pa->getNextContexts(); }
		virtual unsigned long getContextAt(time_t time) const { return pa->getContextAt(time); }
		virtual membershiplist getContextsAt(time_t time) const { return pa->getContextsAt(time); }
		virtual string serialize() const { return pa->serialize(); }
		virtual void unserialize(string data) { pa->unserialize(data); }
};

/**
 * Type of the getPredictor function
 *
 * This type definition is needed by the Predictor Wrapper to query a
 * PredictorAlgorithm implementation for an instance of its class.
 */
typedef PredictorAlgorithm* (*getPredictor_t) (const predictorparams &params);

#endif
