/**
 * @file
 * @brief A simple %Predictor using averages (median for ordinal time series, most frequent for categorical/nominal) as predicted next values.
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-02-23 12:25:16 +0100 (Mon, 23 Feb 2004) $
 * $Revision: 636 $
 *
 * @since    Wed Mar 19 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: AveragePredictor.cpp 636 2004-02-23 11:25:16Z rene $
 */

#include "AveragePredictor.h"
#include <assert.h>

EXPORT_PREDICTOR(AveragePredictor);

AveragePredictor::AveragePredictor(predictorparams &params)
{
	if (params.count("window") == 0) {
		windowSize = 3;
	} else {
		windowSize = atoi(params["window"].c_str());
	}
}

void AveragePredictor::addContexts(const membershiplist* contexts, time_t time)
{
	(void) contexts;
	(void) time;
	throw 1;
}

void AveragePredictor::addContext(unsigned long contextId, time_t time)
{
	(void) time;
	frequencies[contextId]++;
	window.push_back(contextId);
	if (window.size() > windowSize) {
		frequencies[window.front()]--;
		window.pop_front();
	}
}

unsigned long AveragePredictor::getNextContext() const
{
	unsigned int maxFreq = 0, maxId = 0;
	for (map<unsigned int, unsigned int>::const_iterator iter=frequencies.begin(); iter!=frequencies.end(); iter++) {
		if ((*iter).second > maxFreq) {
			maxFreq = (*iter).second;
			maxId = (*iter).first;
		}
	}

	return maxId;
}

membershiplist AveragePredictor::getNextContexts() const
{
	membershiplist ret;
	for (map<unsigned int, unsigned int>::const_iterator iter=frequencies.begin(); iter!=frequencies.end(); iter++) {
		ret[(*iter).first] = (*iter).second / windowSize;
	}

	return ret;
}

// this is simple - the average will not change when we feed the prediction back for the next step (we know that without actually doing it), so just predict the next step forever
contexttrajectory AveragePredictor::getContextTrajectory(unsigned int start, unsigned int end) const
{
	assert(start <= end);
	unsigned long predId = getNextContext();
	contexttrajectory ret(end-start, predId);

	return ret;
}

unsigned long AveragePredictor::getContextAt(time_t time) const
{
	(void) time;
	throw 1;
}

membershiplist AveragePredictor::getContextsAt(time_t time) const
{
	(void) time;
	throw 1;
}

string AveragePredictor::serialize() const
{
	throw 1;
}

void AveragePredictor::unserialize(string data)
{
	(void) data;
	throw 1;
}

string AveragePredictor::toString() const
{
	string ret;
	char buf[512];
	sprintf(buf, "Current window size is %d\n", windowSize);
	ret += buf;
	ret += "Window: ";
	for (list<unsigned>::const_iterator iter = window.begin(); iter!= window.end(); iter++) {
		sprintf(buf, "%d, ", *iter);
		ret += buf;
	}
	ret += "\n";

	ret += "Frequencies: ";
	for (map<unsigned int, unsigned int>::const_iterator iter=frequencies.begin(); iter!=frequencies.end(); iter++) {
		sprintf(buf, "%d: %d\n", (*iter).first, (*iter).second);
		ret += buf;
	}
	return ret;
}
