/**
 * @file
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-08-29 12:14:03 +0200 (Son, 29 Aug 2004) $
 * $Revision: 905 $
 *
 * @since    Wed Mar 19 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: SingleStepDuration.cpp 905 2004-08-29 10:14:03Z rene $
 */

#include "SingleStepDuration.h"
#include <assert.h>

#define DEFAULT_SMOOTHING_WINDOW_WIDTH 4  ///< @todo documentation
#define DEFAULT_SMOOTHING_FACTOR 0.8      ///< @todo documentation
#define DEFAULT_PEAK_THRESHOLD 0.01       ///< @todo documentation

EXPORT_PREDICTOR(SingleStepDurationPredictor);

SingleStepDurationPredictor::SingleStepDurationPredictor(predictorparams &params)
{
	curDuration = 0;
	curContext = 0;
	firstEntry = true;

	if (params.count("smoothing_window") == 0)
		smoothingWindowWidth = DEFAULT_SMOOTHING_WINDOW_WIDTH;
	else
		/// @todo check parameter !!
		smoothingWindowWidth= atoi(params["smoothing_window"].c_str());
	if (params.count("smoothing_factor") == 0)
		smoothingFactor = DEFAULT_SMOOTHING_FACTOR;
	else
		/// @todo check parameter !!
		smoothingFactor = atof(params["smoothing_factor"].c_str());

	if (params.count("peak_treshold") == 0)
		peakDifference = DEFAULT_PEAK_THRESHOLD;
	else
		/// @todo check parameter !!
		peakDifference = atof(params["peak_threshold"].c_str());
}

void SingleStepDurationPredictor::addContexts(const membershiplist* contexts, time_t time)
{
	(void) contexts;
	(void) time;
	throw 1;
}

void SingleStepDurationPredictor::addContext(unsigned long contextId, time_t time)
{
	(void) time;

	if (firstEntry) {
		curContext = contextId;
		curDuration = 1;
 		frequencies[contextId].entries++;
		firstEntry = false;
	}
	else if (contextId != curContext) {
		// a state transition has now happened - remember the old state duration
		// the initialization of new contextinfo objects is done automatically, but we have to take care of the frequency arrays
		if (frequencies[curContext].transitions.count(contextId) == 0) {
//			printf("Creating frequencies array for %d --> %d\n", curContext, contextId);
			frequencies[curContext].transitions[contextId] = durationfrequencies(startfrequencieslength, 0);
		}
		// check for resizes
		if (frequencies[curContext].transitions[contextId].size() <= curDuration) {
//			printf("Resizing freuencies array for %d --> %d to %d\n", curContext, contextId, curDuration + incrementfrequencieslength);
			frequencies[curContext].transitions[contextId].resize(curDuration + incrementfrequencieslength, 0);
		}
		frequencies[curContext].transitions[contextId][curDuration]++;
		curContext = contextId;
		curDuration = 1;
		// and the new state has been entered
		frequencies[contextId].entries++;
	}
	else
		curDuration++;
}

unsigned long SingleStepDurationPredictor::getNextContext() const
{
	membershiplist contexts = getNextContexts();
	unsigned long bestContext = 0;
	double bestProb = 0.0;
	for (membershiplist::const_iterator c = contexts.begin(); c != contexts.end(); c++) {
		if (c->second > bestProb) {
			bestProb = c->second;
			bestContext = c->first;
		}
	}

	return bestContext;
}

unsigned long SingleStepDurationPredictor::getCurrentContextDuration() const
{
	return curDuration;
}

membershiplist SingleStepDurationPredictor::getNextContexts() const
{
	membershiplist ret;

	unsigned long selfTransitionsLeft = (*frequencies.find(curContext)).second.entries;
	if (frequencies.count(curContext) > 0) {
		for (nextcontexts::const_iterator nextId = (*frequencies.find(curContext)).second.transitions.begin(); nextId != (*frequencies.find(curContext)).second.transitions.end(); nextId++) {
			// subtract from the self-transitions
			for (unsigned int i=0; i<curDuration && i<nextId->second.size(); i++)
				selfTransitionsLeft -= nextId->second[i];
			ret[nextId->first] = ((double) nextId->second[curDuration]) / (*frequencies.find(curContext)).second.entries;
		}
		// also add the self-transition
		ret[curContext] = ((double) selfTransitionsLeft) / (*frequencies.find(curContext)).second.entries;
	}

	return ret;
}

membershiplist_duration SingleStepDurationPredictor::getNextContextsWithDuration(unsigned int offset) const
{
	return getNextContextsWithDuration(curContext, curDuration, offset);
}

// this method is 
membershiplist_duration SingleStepDurationPredictor::getNextContextsWithDuration(unsigned long startContext, unsigned long startDuration, unsigned int offset) const
{
	membershiplist_duration ret;

	/// @todo if this turns out to be too inefficient, cache it in the normal data structures
	nextcontexts_smoothened smoothened;

#ifdef DEBUG_DURATION_PREDICTOR
	printf("D: starting at context %ld, duration %ld, offset %d. Context has been entered %ld times\n", startContext, startDuration, offset, (*frequencies.find(startContext)).first);
#endif
		
	/**
	 * @note in the general case, this can be seen as a hierarchical classification problem for classifiying peaks and vales
	 */
	if (frequencies.count(startContext) == 0) {
#ifdef DEBUG_DURATION_PREDICTOR
		printf("D: Whoops, nothing learned!\n");
#endif		
		return ret;
	}

	// for all possible contexts following startContext....
	for (nextcontexts::const_iterator nextId = (*frequencies.find(startContext)).second.transitions.begin(); nextId != (*frequencies.find(startContext)).second.transitions.end(); nextId++) {
#ifdef DEBUG_DURATION_PREDICTOR
		printf("D: Looking at next context %ld....\n", nextId->first);
		
		printf("D: raw transitions are: ");
		for (size_t l=0; l<nextId->second.size(); l++) 
			printf("%ld ", nextId->second[l]);
		printf("\n");
#endif

				
		// first calculate the smoothened transition frequencies
		smoothened[nextId->first] = vector<double>(nextId->second.size(), 0);
		for (size_t l=0; l<smoothened[nextId->first].size(); l++) {
			// the convolution mask is a simple, exponentially decreasing, symmetric mask
			for (int j= 0 - smoothingWindowWidth; j<=(int) smoothingWindowWidth; j++) {
				if ((int) l+j >= 0 && l+j<nextId->second.size()) {
					smoothened[nextId->first][l] += ((double) nextId->second[l+j] * pow(smoothingFactor, ::abs(j)));
				}
			}
		}
#ifdef DEBUG_DURATION_PREDICTOR
		printf("D: smoothened transitions are: ");
		for (size_t l=0; l<smoothened[nextId->first].size(); l++) 
			printf("%f ", smoothened[nextId->first][l]);
		printf("\n");
#endif
		
/*		// now compute the maximum frequency
		unsigned long maxFreq = 0;
		for (unsigned int i=0; i<smoothened.size(); i++) {
			if (smoothened[nextId->first][i] > maxFreq)
				maxFreq = smoothened[nextId->first][i];
		}*/

#ifdef DEBUG_DURATION_PREDICTOR
		printf("D: Searching for first peak....\n");
#endif		
		// and now find the first peak starting from the current location
		unsigned i = startDuration+offset+1, peakInd=startDuration+offset;
		double peak = smoothened[nextId->first][startDuration];
		bool found = false;
		next_context_description c;
		while (!found && i<smoothened[nextId->first].size()) {
#ifdef DEBUG_DURATION_PREDICTOR
			printf("D: %d: ", i);
#endif			
			if (smoothened[nextId->first][i] > peak) {
				peak = smoothened[nextId->first][i];
				peakInd = i;
#ifdef DEBUG_DURATION_PREDICTOR
				printf("new peak\n");
#endif			
			}
			if (peak - smoothened[nextId->first][i] > (unsigned long) (peakDifference * (*frequencies.find(startContext)).second.entries)) {
				found = true;
				c.duration = peakInd-startDuration-offset;
				c.probability = (double) peak / (*frequencies.find(startContext)).second.entries;
#ifdef DEBUG_DURATION_PREDICTOR
				printf("final peak\n");
#endif			
			}
			i++;
		}
		if (found) {
#ifdef DEBUG_DURATION_PREDICTOR
			printf("D: Yes, found one: at offset %ld with probability %f\n", c.duration, c.probability);
#endif			
			ret[nextId->first] = c;
		} else {
#ifdef DEBUG_DURATION_PREDICTOR
			printf("D: Nope, no peak found!\n");
#endif			
			// we can only predict that it will happen after our horizon....
			//c.duration = smoothened[nextId->first].size() - startDuration-offset;
			// ????
			//c.probability = 0.0;
			// just don't insert anything here
		}
	}

/* this is currently definitely not good - work on that !
	unsigned long sumTotals = 0;
	for (unsigned long i=0; i<startDuration; i++)
		sumTotals += frequencies[startContext].totals[i];
	for (nextcontexts::const_iterator nextId = frequencies[startContext].transitions.begin(); nextId != frequencies[startContext].transitions.end(); nextId++) {
		unsigned long sumFreq = 0, i=0, sumTotalsTemp = sumTotals, bestDur = 0;
		double prob;
		while (i<startDuration && i<nextId->second.size()) {
			sumFreq += nextId->second[i];
			i++;
		}
		prob = ((double) sumFreq) / sumTotalsTemp;
		bestDur = i;
		while (i<nextId->second.size()) {
			sumFreq += nextId->second[i];
			sumTotalsTemp += frequencies[startContext].totals[i];
			if (((double) sumFreq) / sumTotalsTemp < prob) {
				// the probability for this transitions is already declining (with increasing duration), thus return the last duration
				break;
			}
			else {
				prob = ((double) sumFreq) / sumTotalsTemp;
				bestDur = i;
			}
			i++;
		}
		// if we haven't left the loop until the break, then the maximum known duration is currently the best for this transition
		next_context_description c;
		c.duration = bestDur;
		c.probability = prob;
		ret[nextId->first] = c;
	}*/

	return ret;
}

contexttrajectory SingleStepDurationPredictor::getContextTrajectory(unsigned int start, unsigned int end) const
{
	contexttrajectory ret;

	unsigned int step = 0;

	while (step < end) {
		// this is the first, very simple implementation
		// just take the one with the maximum probability
		/// @todo this has to be enhanced
		membershiplist_duration dur = getNextContextsWithDuration(step);
		double maxProb = 0;
/*		unsigned long maxId = 0;
		for (membershiplist_duration::const_iterator nextId = dur.begin(); nextId != dur.end(); nextId++) {
			if (nextId->second.probability > maxProb) {
				maxProb = nextId->second.probability;
				maxId = nextId->first;
			}
		}*/
		if (maxProb > 0) {
			// and simply take that
/*			for (unsigned int i=0; i<dur[maxId].duration; i++)
				if (step >= start)
					ret.push_back(maxId);
			step += dur[maxId].duration;*/
		}
		else {
			if (step < start)
				step = start;
			// if we did not longer get any prediction out of that, predict nonsense
			for (unsigned int i=0; i<end-step; i++)
				ret.push_back(0);
			step=end;
		}
	}

	return ret;
}

unsigned long SingleStepDurationPredictor::getContextAt(time_t time) const
{
	(void) time;

	throw 1;
}

membershiplist SingleStepDurationPredictor::getContextsAt(time_t time) const
{
	(void) time;

	throw 1;
}

string SingleStepDurationPredictor::serialize() const
{
	throw 1;
}

void SingleStepDurationPredictor::unserialize(string data)
{
	(void) data;

	throw 1;
}

#if _DEBUG_PREDICTORS
string SingleStepDurationPredictor::toString() const
{
	string ret;
	char buf[64];
/*	for (map<unsigned long, contextinfo>::const_iterator iter = frequencies.begin(); iter != frequencies.end(); iter++) {
		sprintf(buf, "Context %d: %d times entered\n", (*iter).first, (*iter).second.entries);
		ret += buf;
		for (nextcontexts::const_iterator iter2 = (*iter).second.transitions.begin(); iter2 != (*iter).second.transitions.end(); iter2++) {
			sprintf(buf, "%d; ", (*iter2).first);
			ret += buf;
			for (durationfrequencies::const_iterator iter3 = (*iter2).second.begin(); iter3 != (*iter2).second.end(); iter3++) {
				sprintf(buf, "%d;", (*iter3));
				ret += buf;
			}
			ret += "\n";
		}
	}
	return ret;*/

	for (map<unsigned long, contextinfo>::const_iterator iter = frequencies.begin(); iter != frequencies.end(); iter++) {
		sprintf(buf, "Context %ld: %ld times entered\n", (*iter).first, (*iter).second.entries);
		ret += buf;

		unsigned int maxlength=0;
		for (nextcontexts::const_iterator iter2 = (*iter).second.transitions.begin(); iter2 != (*iter).second.transitions.end(); iter2++)
			if ((*iter2).second.size() > maxlength)
				maxlength = (*iter2).second.size();

		for (nextcontexts::const_iterator iter2 = (*iter).second.transitions.begin(); iter2 != (*iter).second.transitions.end(); iter2++) {
			sprintf(buf, "%ld;", (*iter2).first);
			ret += buf;
		}
		ret += "\n\n";

		for (unsigned int i=0; i<maxlength; i++) {
			for (nextcontexts::const_iterator iter2 = (*iter).second.transitions.begin(); iter2 != (*iter).second.transitions.end(); iter2++) {
				if ((*iter2).second.size() > i) {
					sprintf(buf, "%ld;", (*iter2).second[i]);
					ret += buf;
				}
				else
					ret += ";";
			}
			ret += "\n";
		}
		ret += "\n\n";
	}
	return ret;
}
#endif
