/**
 * @file
 * @brief %Predictor interface
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-08-31 22:21:04 +0200 (Die, 31 Aug 2004) $
 * $Revision: 919 $
 *
 * @since    Wed Mar 19 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: Statistics.cpp 919 2004-08-31 20:21:04Z rene $
 */

#include "Statistics.h"
#include <assert.h>

Statistics::Statistics(unsigned int maxAcfOrder, bool useExactCalculation): maxAcfOrder(maxAcfOrder), acfWindow(maxAcfOrder, 0)
{
	num = 0;
	sum = 0.0;
	squaresum = 0.0;
	numTurningPoints = numPositiveDifferences = 0;

	acfSums = new double[maxAcfOrder];
	assert(acfSums != NULL);
	if (useExactCalculation) {
		acfLagSums = new double[maxAcfOrder];
		assert(acfLagSums != NULL);
	}
	else
		acfLagSums = NULL;
	for (unsigned int i=0; i<maxAcfOrder; i++) {
		acfSums[i] = 0.0;
		if (acfLagSums != NULL)
			acfLagSums[i] = 0.0;
	}
}

Statistics::~Statistics()
{
	if (acfSums != NULL)
		delete acfSums;
	if (acfLagSums != NULL)
		delete acfLagSums;
}

void Statistics::nextSample(double value)
{
	num++;
	sum += value;
	squaresum += value*value;

	// a sliding window (length = order of the autocorrelation function)
	assert(acfWindow.size() == maxAcfOrder);
	acfWindow.push_back(value);
	(void) acfWindow.pop_front();
	assert(acfWindow.size() == maxAcfOrder);
	int i=maxAcfOrder-1;
	for (list<double>::const_iterator iter = acfWindow.begin(); iter != acfWindow.end(); iter++) {
		assert(i >= 0);
		if ((unsigned) i < num)
			acfSums[i] += value*(*iter);
		i--;
	}
	if (acfLagSums != NULL) {
		for (unsigned int i=0; i<=num && i<maxAcfOrder; i++)
			acfLagSums[i] += value;
		assert(acfLagSums[0] == sum);
	}

	// count the turning points
	if ((lastVal2 < lastVal1 && lastVal1 > value) ||
	    (lastVal2 > lastVal1 && lastVal1 < value))
		numTurningPoints++;
	// count the number of times that the differenced series is positive
	if (value > lastVal1)
		numPositiveDifferences++;

	// for iid tests
	lastVal2 = lastVal1;
	lastVal1 = value;
}

unsigned long Statistics::getNum() const
{
	return num;
}

double Statistics::getMean() const
{
	if (num > 0)
		return sum / num;
	else
		return 0.0;
}

double Statistics::getVariance() const
{
	if (num > 1) {
		double mean = getMean();
		return (squaresum - 2*mean*sum + num*mean*mean) / (num-1);
	}
	else
		return 0.0;
}

double Statistics::getAcf(unsigned int lag) const
{
	if (lag > maxAcfOrder)
		return 0.0;

	if (num > 1) {
		double mean = getMean();
		double var = getVariance();
		if (acfLagSums == NULL)
			return (acfSums[lag] - 2*mean*sum + num*mean*mean) / num / var;
		else
			return (acfSums[lag] - 2*mean*acfLagSums[lag] + (num-lag)*mean*mean) / num / var;
	}
	else
		return 0.0;
}

void Statistics::testIncrementalAcf(unsigned int maxAcfOrder, double* values, unsigned int len, double accuracy)
{
	Statistics stat(maxAcfOrder);

	for (unsigned int i=0; i<len; i++)
		stat.nextSample(values[i]);

	for (unsigned int i=0; i<maxAcfOrder; i++) {
		double sumtest=0.0, sumtest1=0.0, sumtest2=0.0, sumtest3=0.0, sumtest4=0.0, sumtest5=0.0, sumtest6=0.0;
		for (unsigned int j=0; j<len; j++) {
			if (j >= i) {
				sumtest += (values[j] - stat.getMean())*(values[j-i] - stat.getMean());
				sumtest1 += values[j]*values[j-i];
				sumtest2 += stat.getMean()*values[j-i];
				sumtest3 += stat.getMean()*values[j];
				sumtest4 += stat.getMean()*stat.getMean();
				sumtest5 += values[j-i];
				sumtest6 += values[j];
			}
		}
		printf("%03d: %f ", i, stat.getAcf(i));
		if ((sumtest / stat.getNum() / stat.getVariance() - stat.getAcf(i)) < accuracy) printf("OK\n");
		else {
		    printf("NOT OK (should be %f / %f / %f)\n", sumtest / stat.getNum() / stat.getVariance(),
			(sumtest1 - sumtest2 - sumtest3 + sumtest4) / stat.getNum() / stat.getVariance(),
			(sumtest1 - stat.getMean()*sumtest5 - stat.getMean()*sumtest6 + (stat.getNum()-i)*stat.getMean()*stat.getMean()) / stat.getNum() / stat.getVariance());
		    printf("%f %f\n", sumtest1, stat.acfSums[i]);
		    printf("%f %f\n", sumtest5, stat.acfLagSums[i]);
		    printf("%f %f\n", sumtest6, stat.acfLagSums[i]);
		}

	}
}

bool Statistics::isIidAcf() const
{
	double threshold = 1.96 / sqrt(num);
	unsigned int numOutsideBounds = 0;
	for (unsigned int i=1; i<maxAcfOrder; i++) {
		// count the number of values that fall outside the bound
		if (getAcf(i) > threshold)
			numOutsideBounds++;
		// but reject the test when one value is far outside the bounds
		// currently we define far by 2 times, but that is arbitrary
		if (getAcf(i) > 2*threshold)
			return false;
	}

	// the test passed if at least 95% of all values are within the bounds
	return ((float) numOutsideBounds < (float) maxAcfOrder * 0.05);
}

bool Statistics::isIidTurningPoint() const
{
	// mean and variance of the expected number of turning points, if the sequence was iid
	double expectedNum = 2 * (num-2)/3;
	double var = (16*num - 29)/90;

	// check with a probability of 95%
	return (fabs(numTurningPoints - expectedNum) / var > 1.96);
}

bool Statistics::isIidDifferenceSign() const
{
	// mean and variance of the expected number, if the sequence was iid
	double expectedNum = (num-1)/2;
	double var = (num+1)/12;

	// check with a probability of 95%
	return (fabs(numPositiveDifferences - expectedNum) / var > 1.96);
}

bool Statistics::isIid() const
{
	return isIidAcf() && isIidTurningPoint() && isIidDifferenceSign();
}

unsigned int Statistics::getMaModelOrder() const
{
	double threshold = 1.96 / sqrt(num);
	unsigned int maxLag = 0;
	for (unsigned int i=1; i<maxAcfOrder; i++) {
		if (getAcf(i) > threshold)
			maxLag = i;
	}
	return maxLag;
}
