/**
 * @file
 * @brief Statistics of time series
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-03-02 16:56:08 +0100 (Die, 02 Mär 2004) $
 * $Revision: 679 $
 *
 * @since    Wed Mar 19 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: Statistics.h 679 2004-03-02 15:56:08Z rene $
 */

#ifndef _STATISTICS_H
#define _STATISTICS_H

#include "../util/defines.h"

/** This class can be used to compute statistics on a _single_, numerical continuous variable.
 *  All computations are performance incrementally with runtime performance in mind. */
class Statistics
{
	private:
		// these are all helper variables for incrementally computing the statistical values
		/** The number of samples until now. */
		unsigned long num;
		/** The sum of all values until now. */
		double sum;
		/** The square sum of all values until now. */
		double squaresum;
		/** The maximum lag of the autocorrelation function. */
		const unsigned int maxAcfOrder;
		/** All sums for the autocorrelation function values, from 0 lag to maximum lag. */
		double* acfSums;
		/** More helper sums for the autocorrelation function: the sums lagging behind. */
		double* acfLagSums;
		/** The sliding window for computing the autocorrelation function. */
		list<double> acfWindow;
		/** The number of turning points in the series.
		  * @see isIidTurningPoint
		  */
		unsigned long numTurningPoints;
		/** The number of times the differenced series is positive.
		  * @see isIidDifferenceSign
		  */
		unsigned long numPositiveDifferences;
		/** Two two second last values, needed for the turning point and difference sign tests. */
		double lastVal1, lastVal2;

	public:
		/** If useExactCalculation is true, the autocorrellation values
		 *  will be exact. If set to false, an approximation is used,
		 *  which should be good enough until lag at about 10, and the
		 *  object will only use 2/3 of the memory on the heap
		 *  (2*sizeof(double)*maxAcfOrder instead of
		 *   3*sizeof(double)*maxAcfOrder).
		 *  Please be aware that even with useExactCalculation, the ACF
		 *  will still be off a bit, but at least be exact to an
		 *  accuracy of 10^-3 until lag 128.
		 *
		 *  Setting maxAcfOrder to 0 will disable computation of the
		 *  ACF and thus reduce the runtime complexity of nextSample
		 *  significantly.
		 */
		Statistics(unsigned int maxAcfOrder, bool useExactCalculation = true);
		~Statistics();

		/** Update the internal helper variables with a new sample. */
		void nextSample(double value);

		/** The number of samples. */
		unsigned long getNum() const;

		/** The (sample) mean of all sample values. */
		double getMean() const;

		/** The (sample) variance of all sample values. */
		double getVariance() const;

		/** The (sample) autocorrelation of all samples values at the
		  * given lag.
		  *
		  * This function thus represents the (sample) autocorrelation
		  * function (ACF). Please see "Introduction to Time Series and
		  * Forecasting", page 59 for details.
		  */
		double getAcf(unsigned int lag) const;

		/** This can be used for validating the computed ACF values. */
		static void testIncrementalAcf(unsigned int maxAcfOrder, double* values, unsigned int len, double accuracy);

		/** Test the time series if it can be regarded as <i>iid</i>
		  * noise - independent and identically distributed random
		  * variables.
		  *
		  * This method uses the sample autocorrelation function
		  * (whose values are available with getAcf) and checks if all
		  * values are within the bounds +-1.96/sqrt(getNum()).
		  * This corresponds to a probability of 95% that the sequence
		  * is iid according to this test.
		  *
		  * @return <i>true</i> if all values are within the bounds,
		  *         <i>false</i> otherwise.
		  */
		bool isIidAcf() const;

		/** Test the time series if it can be regarded as <i>iid</i>
		  * noise - independent and identically distributed random
		  * variables.
		  *
		  * This method uses the turning point test, described in
		  * "Introduction to Time Series and Forecasting", page 36f.
		  *
		  * @return <i>true</i> if the sequence is iid according to the
		  *         test, <i>false</i> otherwise.
		  */
		bool isIidTurningPoint() const;

		/** Test the time series if it can be regarded as <i>iid</i>
		  * noise - independent and identically distributed random
		  * variables.
		  *
		  * This method uses the difference sign test, described in
		  * "Introduction to Time Series and Forecasting", page 37f.
		  *
		  * If this test returns with <i>false</i>, it indicates the
		  * presence of an increasing (or decreasing) trend in the data.
		  *
		  * @return <i>true</i> if the sequence is iid according to the
		  *         test, <i>false</i> otherwise.
		  */
		bool isIidDifferenceSign() const;

		/** Test the time series if it can be regarded as <i>iid</i>
		  * noise - independent and identically distributed random
		  * variables.
		  *
		  * This test combines all implemented specific tests.
		  *
		  * @return <i>true</i> if the sequence is iid according to the
		  *         all tests (none returned <i>false</i>), <i>false</i>
		  *         otherwise.
		  */
		bool isIid() const;

		/** Try to estimate the order q of an MA(q) model, under the
		  * proposition that the time series indeed can be expressed
		  * by an MA model.
		  *
		  * This method basically returns the highest lag for which
		  * the sample ACF value (computed with getAcf(lag) is greater
		  * than +-1.96/sqrt(getNum()).
		  * It is an implementation of the method described in
		  * "Introduction to Time Series and Forecasting", page 94.
		  *
		  * @return The estimated order of an MA model, at maximum
		  *         the value of maxAcfOrder specified in the
		  *         constructor (which can also be <i>0</i> if the
		  *         ACF calculation was disabled in that way).
		  * @see getAcf
		  */
		unsigned int getMaModelOrder() const;

};

#endif
