#define WINDOWS_LEAN_AND_MEAN

#include <windows.h>
#include <delayimp.h>

#include "clustering_dispatch.h"
#include "libmatlbm.h"
#include "string.h"

#include "../gng/GNG.h"
#include <map>

static std::map<int, ClusteringAlgorithm *> instance;
static unsigned long instance_counter = 0;
static unsigned long reference_counter = 0;

static mxArray *init(char *algorithm);
static void free(int handle);
static mxArray *add_sample(int handle, mxArray *sample);
static mxArray *get_cluster(int handle);
static mxArray *get_connections(int handle);

extern "C" _mexLocalFunctionTable _local_function_table_clustering_dispatch = {0, NULL};

extern "C" void InitializeModule_clustering_dispatch(void)
{
}

extern "C" void TerminateModule_clustering_dispatch(void)
{
}

extern "C" void mlxClustering_dispatch(int nlhs, mxArray * plhs[], int nrhs, mxArray * prhs[])
{
    char buffer[16];
	unsigned handle;

	if ((nrhs < 1) || !mxIsChar(prhs[0])) {
		mexErrMsgTxt("Run-time Error: The function \"clustering_dispatch\" has to be called with a command.");
	} else {
		mxGetString(prhs[0], buffer, sizeof(buffer));
		/* init */
		if (!_stricmp(buffer, "init")) {
			if ((nlhs != 1) || (nrhs != 2)) {
				mexErrMsgTxt("Run-time Error: The function \"clustering_dispatch\" was with the wrong count of parameters for \"init\" mode.");
			} else {
				char algo[64];
				
				mxGetString(prhs[1], algo, sizeof(algo));
				plhs[0] = init(algo);
			}
		} else {
			if ((nrhs < 2) || !mxIsDouble(prhs[1])) {
				mexErrMsgTxt("Run-time Error: The function \"clustering_dispatch\" was called without handle.");
			} else {
				handle = (unsigned) *mxGetPr(prhs[1]);
				/* free */
				if (!_stricmp(buffer, "free")) {
					if (nlhs || (nrhs != 2)) {
						mexErrMsgTxt("Run-time Error: The function \"clustering_dispatch\" was with the wrong count of parameters for \"free\" mode.");
					} else {
						free(handle);
					}

				/* add_sample */
				} else if (!_stricmp(buffer, "add_sample")) {
					if (nlhs != 1 || (nrhs != 3)) {
						mexErrMsgTxt("Run-time Error: The function \"clustering_dispatch\" was with the wrong count of parameters for \"add_sample\" mode.");
					} else {
						if (!mxIsDouble(prhs[2]) || (mxGetNumberOfDimensions(prhs[2]) != 2) || (*(mxGetDimensions(prhs[2])) != 1)) {
							mexErrMsgTxt("Run-time Error: Third Parameter has to be a single dimensional array of doubles.");
						} else {
							plhs[0] = add_sample(handle, prhs[2]);
						}
					}

				/* get_cluster */
				} else if (!_stricmp(buffer, "get_cluster")) {
					if ((nlhs != 1) || (nrhs != 2)) {
						mexErrMsgTxt("Run-time Error: The function \"clustering_dispatch\" was with the wrong count of parameters for \"get_cluster\" mode.");
					} else {
						plhs[0] = get_cluster(handle);
					}

				/* get_connections */
				} else if (!_stricmp(buffer, "get_connections")) {
					if ((nlhs != 1) || (nrhs != 2)) {
						mexErrMsgTxt("Run-time Error: The function \"clustering_dispatch\" was with the wrong count of parameters for \"get_connections\" mode.");
					} else {
						plhs[0] = get_connections(handle);
					}
				
				/* *unknown* */
				} else {
					mexErrMsgTxt("Run-time Error: The function \"clustering_dispatch\" was called with an illegal command.");
				}
			}
		}
	}

	if (!reference_counter) {
		GNG_UNLOAD();
	}
}

static mxArray *init(char *algorithm)
{
	ClusteringAlgorithm *algo = NULL;
	mxArray *ret = mxCreateDoubleMatrix(1, 1, mxREAL);
	double *val = mxGetPr(ret);
	
	if (!_stricmp(algorithm, "gng")) {
		algo = new GNG();
		reference_counter++;
	} else {
		mexErrMsgTxt("Run-time Error: Requested algorithm is not defined.");
	}

	*val = instance_counter;
	instance[instance_counter++] = algo;

	return ret;
}

static void free(int handle)
{
	ClusteringAlgorithm *algo = instance[handle];

	if (algo == NULL) {
		mexErrMsgTxt("Run-time Error: Instance already freed.");
	} else {
		delete algo;
		instance[handle] = NULL;
		reference_counter--;
	}
}

static mxArray *add_sample(int handle, mxArray *sample)
{
	mxArray *ret = mxCreateDoubleMatrix(1, 1, mxREAL);
	double *val = mxGetPr(ret);

	if (instance[handle] == NULL) {
		mexErrMsgTxt("Run-time Error: Instance already freed.");
		*val = 0;
	} else {
		position *pos = new position();
		int const *dims = mxGetDimensions(sample);
		double *elem = mxGetPr(sample);

		for (int i=0; i < *dims; i++) {
			pos->push_back(elem[i]);
		}

		*val = (double) instance[handle]->addSample(pos);
	}

	return ret;
}

static mxArray *get_cluster(int handle)
{
	mxArray *ret = NULL;

	if (instance[handle] == NULL) {
		mexErrMsgTxt("Run-time Error: Instance already freed.");
	} else {
		const char *fields[] = {
			"position",
			"parameter"
		};

		cluster_table *table = instance[handle]->getCluster();
		ret = mxCreateStructMatrix(1, table->size(), 2, fields);

		for (unsigned i=0; i < table->size(); i++) {
			mxArray *arrs;
			double *elem;
			unsigned size;

			/* params */
			position *pos = &((*table)[i].pos);
			size = pos->size();
			arrs = mxCreateDoubleMatrix(1, size, mxREAL);
			mxSetFieldByNumber(ret, i, 0, arrs);
			elem = mxGetPr(arrs);
			for (unsigned j=0; j < size; j++) {
				elem[j] = (*pos)[j];
			}

			/* pos */
			parameter *param = &((*table)[i].param);
			size = param->size();
			arrs = mxCreateDoubleMatrix(1, size, mxREAL);
			mxSetFieldByNumber(ret, i, 1, arrs);
			elem = mxGetPr(arrs);
			for (unsigned j=0; j < size; j++) {
				elem[j] = (*param)[j];
			}
		}

		delete table;
	}

	return ret;
}

static mxArray *get_connections(int handle)
{
	mxArray *ret = NULL;

	if (instance[handle] == NULL) {
		mexErrMsgTxt("Run-time Error: Instance already freed.");
	} else {
		double *a, *b;
		connection_table *table = instance[handle]->getConnections();
		unsigned size = table->size();

		ret = mxCreateDoubleMatrix(size, 2, mxREAL);
		a = mxGetPr(ret);
		b = a + size;

		for (unsigned i=0; i < size; i++) {
			*a++ = (double) (*table)[i].a;
			*b++ = (double) (*table)[i].b;
		}

		delete table;
	}

	return ret;
}