/**
 * @file
 * @brief main
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-08-29 15:06:22 +0200 (Son, 29 Aug 2004) $
 * $Revision: 906 $
 *
 * @since    Mon Mar 10 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: main-classification.cpp 906 2004-08-29 13:06:22Z rene $
 */

#include "Classifiers/gng/GNG.h"
#include "Features/Time/TimeFeature.h"

#include "ConfigReader/ConfigReader.h"
#include "FeatureContainer/ReplayFeatureContainer.h"
#include "Classifiers/Classifier.h"

#ifdef _WINDOWS
#define PATH "\\"
#endif

#ifdef _LINUX
#define PATH "./"
#endif

#ifdef _WIN32_WCE
#define PATH "\\"
#endif

#define CLASSIFICATION_RUNS 1 ///< amount of training iterations
#define HOT 0.995             ///< start temperature for sa
#define COLD 0.0              ///< end temperature for sa
#define MINIMIZE_ERROR 0      ///< try to minimize classification error by finding the best classifier parameters
#define MINIMIZE_VARIANCE 0   ///< try to minimize classification error by finding the best classifier parameters

//@{
/** @todo documentation */
typedef map<unsigned, list<double> > cluster_list;
typedef map<unsigned, cluster_list> meta_list;

static string classes_log;
static string algo_name;
static map<string, string> algo_params;
//@}

/**
 * Evaluate quality of training
 *
 * @param fc ReplayFeatureContainer reference
 * @param classes_log Path to logfile
 */
static void evaluate(ReplayFeatureContainer *fc, string classes_log = "")
{
	FILE* member_file = NULL;
	meta_list samples;
	bool frozen = false;
	bool hasSamples = true;
	double temp = HOT;
        unsigned int epoch=1;

	Classifier *ca = new Classifier(algo_name, algo_params);

	// known good solution
	double total_s2 = 0.00223;
	double total_error = 0.0336;
	double total_minimize = 0.0;

	double total_s2_best = numeric_limits<double>::infinity();
	double total_error_best = numeric_limits<double>::infinity();
	double total_minimize_best = numeric_limits<double>::infinity();

#if MINIMIZE_ERROR || MINIMIZE_VARIANCE
	// params to tune
	unsigned int lambda;
	double minimal_deletion_age;
	double deletion_threshold;
	double deletion_threshold_lq;
	unsigned long edge_age_max;
	double T_short;
	double T_long;
	double T_youth;
	double T_insert_threshold;
	double adaptation_threshold;
	double insertion_tolerance;
	double lr_winner;
	double lr_neighbor;
	double minimum_lr;
	double lr_insert_threshold;
#endif

	plog("\n--------------- Reloading file for evaluation ------------------\n");

	while (!frozen) {

#if MINIMIZE_ERROR || MINIMIZE_VARIANCE
# ifdef _GNG_H
		GNG* gng = ((GNG *) ((Classifier *) ca)->getCa());

		if ((rand_double() < (temp / 10.0)) && ((total_error_best / total_error) > (1.0 - temp))) { // accept worse solutions with max 0.1
			plog("accept worse solution with %lf", (temp / 10.0));

			total_error_best = total_error;
			total_s2_best = total_s2;
		}

		total_minimize = 1.0
#  if MINIMIZE_ERROR
		* total_error
#  endif
#  if MINIMIZE_VARIANCE
		* total_s2
#  endif
		;

		if (total_minimize <= total_minimize_best || total_minimize_best == numeric_limits<double>::infinity())
		{
			lambda = gng->lambda;
			minimal_deletion_age = gng->minimal_deletion_age;
			deletion_threshold = gng->deletion_threshold;
			deletion_threshold_lq = gng->deletion_threshold_lq;
			edge_age_max = gng->edge_age_max;
			T_short = gng->T_short;
			T_long = gng->T_long;
			T_youth = gng->T_youth;
			T_insert_threshold = gng->T_insert_threshold;
			adaptation_threshold = gng->adaptation_threshold;
			insertion_tolerance = gng->insertion_tolerance;
			lr_winner = gng->lr_winner;
			lr_neighbor = gng->lr_neighbor;
			minimum_lr = gng->minimum_lr;
			lr_insert_threshold = gng->lr_insert_threshold;

			plog("\nbest match so far\n");

			total_s2_best = total_s2;
			total_error_best = total_error;
			total_minimize_best = total_minimize;
		}

		plog("\nCurrent best solution (avg: %.17lf, s2: %.17lf)", total_error_best, total_s2_best);
		plog("\nGNG[%u, %.17lf, %.17lf, %.17lf, %u, %.17lf, %.17lf, %.17lf, %.17lf, %.17lf, %.17lf, %.17lf, %.17lf, %.17lf, %.17lf]\n",
			lambda,
			minimal_deletion_age,
			deletion_threshold,
			deletion_threshold_lq,
			edge_age_max,
			T_short,
			T_long,
			T_youth,
			T_insert_threshold,
			adaptation_threshold,
			insertion_tolerance,
			lr_winner,
			lr_neighbor,
			minimum_lr,
			lr_insert_threshold);

		delete ca;
		ca = new Classifier(algo_name, algo_params);
		ca->init(fc);

		// reset params
		gng->lambda = lambda;
		gng->minimal_deletion_age = minimal_deletion_age;
		gng->deletion_threshold = deletion_threshold;
		gng->deletion_threshold_lq = deletion_threshold_lq;
		gng->edge_age_max = edge_age_max;
		gng->T_short = T_short;
		gng->T_long = T_long;
		gng->T_youth = T_youth;
		gng->T_insert_threshold = T_insert_threshold;
		gng->adaptation_threshold = adaptation_threshold;
		gng->insertion_tolerance = insertion_tolerance;
		gng->lr_winner = lr_winner;
		gng->lr_neighbor = lr_neighbor;
		gng->minimum_lr = minimum_lr;
		gng->lr_insert_threshold = lr_insert_threshold;

		// seed the random number generator to get better results
		srand(clock());

		// anneal
		double sign = (rand_double() < 0.5) ? 1.0 : -0.5; // sign
		int param = (int) (rand_double() * 15); // select param to tune

/** Annealing operator */
#define tune(param) while (gng->param == param) { gng->param = param + param * sign * rand_double(); }
#define bound(param, min, max) if(gng->param < min) gng->param=min; if(gng->param > max) gng->param=max;

		switch (param) {
			case 0: // lambda
				tune(lambda);
				break;

			case 1: // minimal_deletion_age
				tune(minimal_deletion_age);
				break;

			case 2: // deletion_threshold
				tune(deletion_threshold);
				break;

			case 3: // deletion_threshold_lq
				tune(deletion_threshold_lq);
				break;

			case 4: // edge_age_max
				tune(edge_age_max);
				break;

			case 5: // T_short
				tune(T_short);
				break;

			case 6: // T_long
				tune(T_long);
				break;

			case 7: // T_youth
				tune(T_youth);
				break;

			case 8: // T_insert_threshold
				tune(T_insert_threshold);
				break;

			case 9: // adaptation_threshold
				tune(adaptation_threshold);
				bound(adaptation_threshold, 0, 1);
				break;

			case 10: // lambda
				tune(insertion_tolerance);
				bound(insertion_tolerance, 0, 1);
				break;

			case 11: // lr_winner
				tune(lr_winner);
				break;

			case 12: // lr_neighbor
				tune(lr_neighbor);
				break;

			case 13: // minimum_lr
				tune(minimum_lr);
				break;

			case 14: // lambda
				tune(lr_insert_threshold);
				bound(lr_insert_threshold, 0, 1);
				break;

		}

		plog("\nGNG[%u, %.17lf, %.17lf, %.17lf, %u, %.17lf, %.17lf, %.17lf, %.17lf, %.17lf, %.17lf, %.17lf, %.17lf, %.17lf, %.17lf]\n",
			gng->lambda,
			gng->minimal_deletion_age,
			gng->deletion_threshold,
			gng->deletion_threshold_lq,
			gng->edge_age_max,
			gng->T_short,
			gng->T_long,
			gng->T_youth,
			gng->T_insert_threshold,
			gng->adaptation_threshold,
			gng->insertion_tolerance,
			gng->lr_winner,
			gng->lr_neighbor,
			gng->minimum_lr,
			gng->lr_insert_threshold);

		// cool down
		temp *= temp;
		plog("temperature: %.17lf\n", temp);

		if (temp < COLD) {
			frozen = true;
		}

		// reset pseudorandom generator to be a bit more deterministic here
		srand(1);

		fc->resetFile();
		hasSamples = true;
		while (hasSamples) {
			try {
				// circumvent disk serialization
				fc->nextSample();
				const featurevector *sample = fc->getSampleVector();
				ca->nextSample(sample);
			} catch(...) {
				hasSamples = false;
			}
		}
		
		plog("\n\n");

		plog("Saving GNG data structure for epoch %d\n", epoch);
		FILE* gngfile;
		char gngfilename[256];
                sprintf(gngfilename, PATH"gngdata.txt-epoch%04ld", epoch);
       		gngfile = fopen(gngfilename, "w");
       		assert(gngfile != NULL);
       		string gngdata = ca->serialize();
       		fputs(gngdata.c_str(), gngfile);
       		fclose(gngfile);

                member_file = NULL;
		if (classes_log != "") {
		    plog("Creating classes log for epoch %d\n", epoch);
		    char memberfilename[256];
		    sprintf(memberfilename, "%s-epoch%04d", classes_log.c_str(), epoch);
			member_file = fopen(memberfilename, "w");
			assert(member_file != NULL);
		}
# else
		frozen = true;
# endif
#else
		frozen = true;

		if (classes_log != "") {
			member_file = fopen(classes_log.c_str(), "w");
			assert(member_file != NULL);
		}

		ca->init(fc);

		FILE* gngfile;
		if ((gngfile = fopen(PATH"gngdata.txt", "r")) != NULL) {
			plog("Found GNG data file, unserializing GNG structure and skipping training\n");
			fclose(gngfile);
		} else {
			plog("Saving GNG data structure after randomized initialization\n");
			FILE* gnginitfile = fopen(PATH"gngdata-init.txt", "w");
			assert(gnginitfile != NULL);
			string gnginitdata = ca->serialize();
			fputs(gnginitdata.c_str(), gnginitfile);
			fclose(gnginitfile);

			int num = 0;
			time_t start;
			time(&start);
			for (int i=1; i<= CLASSIFICATION_RUNS; i++) {
				plog("\n--------------- Reloading file for training (%d) --------------\n", i);

				fc->resetFile();
				hasSamples = true;
				while (hasSamples) {
					try {
						num++;

						if (num % 1000 == 0) {
							plog(".");
						}
						if (num % 10000 == 0) {
							plog(" %i\n", num);
						}

						// circumvent disk serialization
						fc->nextSample();
						const featurevector *sample = fc->getSampleVector();
						ca->nextSample(sample);
					} catch(...) {
						hasSamples = false;
					}
				}
			}
			time_t end;
			time(&end);

			int diff = (int) (end - start);
			plog("\ntraining took %i seconds\n", diff);

			plog("Processed %i samples\n", num);
			plog("Saving GNG data structure\n");
			gngfile = fopen(PATH"gngdata.txt", "w");
			assert(gngfile != NULL);
			string gngdata = ca->serialize();
			fputs(gngdata.c_str(), gngfile);
			fclose(gngfile);
		}
#endif

		fc->resetFile();
		samples.clear();

		int num2 = 0;
		hasSamples = true;
		while (hasSamples) {
			try {
				fc->nextSample();

				const featurevector *s = fc->getSampleVector();
				unsigned cluster_id = 0, meta_cluster_id = 0;
				double error = 0.0;

	#ifdef _TIMEFEATURE_H
				TimeFeature* time = NULL;
				for (featurevector::const_iterator iter1 = s->begin(); time == NULL && iter1 != s->end(); iter1++) {
					if ((*iter1)->getName().find("Time.") == 0) {
						time = (TimeFeature*) *iter1;
					}
				}
	#endif

	#ifdef _GNG_H
				GNG* gng = ((GNG *) ((Classifier *) ca)->getCa());
				error = gng->getWinnerDistance(s, cluster_id, meta_cluster_id);
	#endif

				samples[meta_cluster_id][cluster_id].push_back(error);

				if (member_file != NULL) {
	#ifdef _TIMEFEATURE_H
					if (time != NULL) {
						fprintf(member_file, "%d;", time->getVal());
					}
	#endif
					membershiplist ml = ca->getClusterMembership(s);
					for (membershiplist::iterator iter3 = ml.begin(); iter3 != ml.end(); iter3++) {
						fprintf(member_file, "%.17f;", (*iter3).second);
					}
					fprintf(member_file, "%d;\n", meta_cluster_id);
				}

				num2++;

				if (num2 % 1000 == 0) {
					plog(".");
				}
				if (num2 % 10000 == 0) {
					plog(" %i\n", num2);
				}
			} catch(...) {
				hasSamples = false;
			}
		}

		if (member_file != NULL) {
			fclose(member_file);
		}

		/* output statistics */

		plog("\n ------------- -------- ---------- ---------- ---------- ----------\n");
		plog("| cluster     | hits   | avg      | s2       | 0.95  *) | 0.98  *) |\n");
		plog("|-------------+--------+----------+----------+----------+----------|\n");

		int skip;
		unsigned total_hits = 0;
		double total_conf95 = 0.0;
		double total_conf98 = 0.0;
		list<double> total_list;

		total_error = 0.0;
		total_s2 = 0.0;

		for (meta_list::iterator iter4 = samples.begin(); iter4 != samples.end(); iter4++) {
			unsigned meta_hits = 0;
			double meta_s2 = 0.0;
			double meta_error = 0.0;
			double meta_conf95 = 0.0;
			double meta_conf98 = 0.0;
			list<double> meta_list;

			for (cluster_list::iterator iter5 = (*iter4).second.begin(); iter5 != (*iter4).second.end(); iter5++) {
				unsigned cluster_hits = (unsigned) (*iter5).second.size();
				double cluster_s2 = 0.0;
				double cluster_error = 0.0;
				double cluster_conf95 = 0.0;
				double cluster_conf98 = 0.0;

				(*iter5).second.sort();
				(*iter5).second.reverse();

				list<double>::iterator iter6 = (*iter5).second.begin();

				skip = (int) ((double) cluster_hits * 0.02 + 0.5); // skip 2%

				while ((skip > 0) && (iter6 != (*iter5).second.end())) {
					cluster_error += (*iter6);
					cluster_s2 += (*iter6) * (*iter6);
					meta_list.push_back(*iter6);

					iter6++;
					skip--;
				}

				for (int i = 0;(i < 2) && (iter6 != (*iter5).second.end()); i++) {
					cluster_conf98 += (*iter6);
					cluster_error += (*iter6);
					cluster_s2 += (*iter6) * (*iter6);
					meta_list.push_back(*iter6);

					iter6++;
				}

				skip = (int) ((double) cluster_hits * 0.05 + -1.5); // skip 3% + 2% (95%)

				while ((skip > 0) && (iter6 != (*iter5).second.end())) {
					cluster_error += (*iter6);
					cluster_s2 += (*iter6) * (*iter6);
					meta_list.push_back(*iter6);

					iter6++;
					skip--;
				}

				for (int i = 0;(i < 2) && (iter6 != (*iter5).second.end()); i++) {
					cluster_conf95 += (*iter6);
					cluster_error += (*iter6);
					cluster_s2 += (*iter6) * (*iter6);
					meta_list.push_back(*iter6);

					iter6++;
				}

				while (iter6 != (*iter5).second.end()) {
					cluster_error += (*iter6);
					cluster_s2 += (*iter6) * (*iter6);
					meta_list.push_back(*iter6);

					iter6++;
				}

				meta_hits += cluster_hits;
				meta_error += cluster_error;
				meta_s2 += cluster_s2;

				cluster_error /= cluster_hits;
				cluster_s2 /= cluster_hits - 1;

				plog("| cluster %3i | %6u | %.6lf | %.6lf | %.6lf | %.6lf |\n", (*iter5).first, cluster_hits, cluster_error, cluster_s2, cluster_conf95, cluster_conf98);
			}

			total_hits += meta_hits;
			total_error += meta_error;
			total_s2 += meta_s2;

			meta_list.sort();
			meta_list.reverse();

			list<double>::iterator iter7 = meta_list.begin();

			skip = (int) ((double) meta_hits * 0.02 + 0.5); // skip 2%

			while ((skip > 0) && (iter7 != meta_list.end())) {
				total_list.push_back(*iter7);

				iter7++;
				skip--;
			}

			for (int i = 0;(i < 2) && (iter7 != meta_list.end()); i++) {
				meta_conf98 += (*iter7);
				total_list.push_back(*iter7);

				iter7++;
			}

			skip = (int) ((double) meta_hits * 0.05 + -1.5); // skip 3% + 2% (95%)

			while ((skip > 0) && (iter7 != meta_list.end())) {
				total_list.push_back(*iter7);

				iter7++;
				skip--;
			}

			for (int i = 0;(i < 2) && (iter7 != meta_list.end()); i++) {
				meta_conf95 += (*iter7);
				total_list.push_back(*iter7);

				iter7++;
			}

			while (iter7 != meta_list.end()) {
				total_list.push_back(*iter7);

				iter7++;
			}

			meta_error /= meta_hits;
			meta_s2 /= meta_hits - 1;

			plog("|-------------+--------+----------+----------+----------+----------|\n");
			plog("| meta    %3i | %6u | %.6lf | %.6lf | %.6lf | %.6lf |\n", (*iter4).first, meta_hits, meta_error, meta_s2, meta_conf95, meta_conf98);
			plog("|=============|========|==========|==========|==========|==========|\n");
			plog("|             |        |          |          |          |          |\n");
		}

		total_list.sort();
		total_list.reverse();

		list<double>::iterator iter8 = total_list.begin();

		skip = (int) ((double) total_hits * 0.02 + 0.5); // skip 2%

		while ((skip > 0) && (iter8 != total_list.end())) {
			iter8++;
			skip--;
		}

		for (int i = 0;(i < 2) && (iter8 != total_list.end()); i++) {
			total_conf98 += (*iter8);

			iter8++;
		}

		skip = (int) ((double) total_hits * 0.05 + -1.5); // skip 3% + 2% (95%)

		while ((skip > 0) && (iter8 != total_list.end())) {
			iter8++;
			skip--;
		}

		for (int i = 0;(i < 2) && (iter8 != total_list.end()); i++) {
			total_conf95 += (*iter8);

			iter8++;
		}

		total_error /= total_hits;
		total_s2 /= total_hits - 1;

		plog("|=============|========|==========|==========|==========|==========|\n");
		plog("| total       | %6u | %.6lf | %.6lf | %.6lf | %.6lf |\n", total_hits, total_error, total_s2, total_conf95, total_conf98);
		plog(" ------------- -------- ---------- ---------- ---------- ----------\n");
                plog("(at epoch %04d)\n", epoch);
		plog("\n*) worst case estimation\n\n");

		//plog("%u of %u clusters with hits\n%u of %u metaclusters without hits\n\n", total_list.size(), gng->getNumberNodes(), samples.size(), gng->getMetaClusters().size());
                epoch++;
	}
}

#ifdef _WIN32_WCE
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow)
{
	MSG msg;
#else
/** Main */
int main(int argc, char* argv[])
{
	(void) argc;
	(void) argv;
#endif

//	_CrtSetBreakAlloc(562);

	open_log(PATH"debug_logfile.txt");

	plog("Reading Configuration\n");

	ConfigReader *r = new ConfigReader(PATH"config.xml");
	map<string, string> globals = (*r->getGlobals());

	assert(globals["persist"] != "");
	assert(globals["features_log"] != "");
	assert(globals["classes_log"] != "");

	plog("Creating ReplayFeatureContainer\n");
	ReplayFeatureContainer* fc = new ReplayFeatureContainer(r);
	plog("Features log file has a size of %i bytes\n", fc->getFilesize());

	const pair<string, classifierparams> *foo = r->getClassifier();
	algo_name = (*foo).first;
	algo_params = (*foo).second;

	plog("Creating Classifier %s\n", algo_name.c_str());

	classes_log = globals["classes_log"];

	delete r;

	evaluate(fc, classes_log);

	delete fc;

	plog("Terminating main program now\n");
	close_log();

	DUMP_LEAKS();

	return 0;
}
