#include "intelligence.h"

#include "../ConfigReader/ConfigReader.h"
#include "../Classifiers/Classifier.h"
#include "../FeatureContainer/LoggingFeatureContainer.h"
#include ".\intelligence.h"

#include "../Features/Time/TimeFeature.h"

/** Registry root */
#define ROOT "SOFTWARE\\JKU\\Intelligence"

ULONG CIntelligence::refCount = 0;
ITypeInfo *CIntelligence::typeInfo = NULL;

CIntelligence::CIntelligence(void)
{
	HKEY key;
	DWORD len;
	unsigned char config[256];
	unsigned char log[256];

	// com specific
    if (!refCount) {
		ITypeLib* typeLib;

		if (FAILED(LoadRegTypeLib(LIBID_Intelligence, 1, 0, 0, &typeLib))) {
			throw;
		}

		if (FAILED(typeLib->GetTypeInfoOfGuid(IID_IIntelligence, &typeInfo))) {
			throw;
		}

		typeLib->Release();
	}

	AddRef();

	// intelligence specific
	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, ROOT, 0, KEY_READ, &key) != ERROR_SUCCESS) {
		return;
	}

	len = sizeof(config);
	if (RegQueryValueEx(key, "config", NULL, NULL, config, &len) != ERROR_SUCCESS) {
		return;
	}

	len = sizeof(log);
	if (RegQueryValueEx(key, "log", NULL, NULL, log, &len) != ERROR_SUCCESS) {
		return;
	}

	// before doing anything dangerous, open the logfile
	open_log((char *) log);

	plog("Parsing configuration\n");
	ConfigReader *r = new ConfigReader((char *) config);
	map<string, string> *globals = (map<string, string> *) r->getGlobals();

	classes = (*globals)["classes_log"].c_str();
	plog("Configuration parsed\n");

	plog("Loading features\n");
	f = new LoggingFeatureContainer(r);
	plog("Features loaded\n");

	plog("Creating classifier\n");
	const pair<string, classifierparams> *foo = r->getClassifier();
	ca = new Classifier((*foo).first, (*foo).second);
	ca->init(f);
	plog("Classifier %s loaded\n", (*foo).first.c_str());

	delete r;

	// init
	ca->nextSample();
	samples = f->getSampleVector();
}

CIntelligence::~CIntelligence(void)
{
	plog("Terminating intelligence\n");
	delete ca;
	plog("Intelligence terminated\n");

	close_log();
	if (typeInfo != NULL) {
		typeInfo->Release();
	}
}

void CIntelligence::nextSample()
{
	FILE *log = NULL;
	double membership = -1.0;

	ca->nextSample();
	samples = f->getSampleVector();

	membershiplist list = ca->getClusterMembership(samples);

	log = fopen(classes, "a");

#ifdef _TIMEFEATURE_H
	for (featurevector::const_iterator iter1 = samples->begin(); time == NULL && (iter1 != samples->end()); iter1++) {
		if ((*iter1)->getName().find("Time.") == 0) {
			TimeFeature* time = (TimeFeature*) *iter1;

			if (log != NULL) {
				fprintf(log, "%d;", time->getVal());
			}
		}
	}
#endif

	activeId = -1;

	for (membershiplist::iterator iter2 = list.begin(); iter2 != list.end(); iter2++) {
		if (membership < (*iter2).second) {
			activeId = (*iter2).first;
			membership = (*iter2).second;

			if (log != NULL) {
				fprintf(log, "%.17f;", activeId);
			}
		}
	}

	if (log != NULL) {
		fprintf(log, "%d;\n", activeId);
		fclose(log);
	}

#if _DEBUG_FEATURES
	for (featurevector::const_iterator iter = samples->begin(); iter != samples->end(); iter++) {
		if ((*iter) != NULL) {
			plog("%s\n", (*iter)->toString().c_str());
		} else {
			plog("NULL\n");
		}
	}
#endif
}

bool CIntelligence::wait(bool &keepRunning)
{
	for (int j=0; j<30 && keepRunning; j++) {
		Sleep(1000);
	}

	return keepRunning;
}

void CIntelligence::calcActiveId()
{
	double membership = -1.0;
	membershiplist list = ca->getClusterMembership(samples);

	activeId = -1;

	for (membershiplist::iterator iter = list.begin(); iter != list.end(); iter++) {
		if (membership < (*iter).second) {
			activeId = (*iter).first;
			membership = (*iter).second;
		}
	}
}

STDMETHODIMP CIntelligence::QueryInterface(REFIID riid, LPVOID *ppv)
{
	if (IsBadWritePtr(ppv, sizeof(LPVOID))) {
        return E_INVALIDARG;
	}

    *ppv = NULL;

	if (riid == IID_IUnknown || riid == IID_IIntelligence || riid == IID_IDispatch || riid == IID_ISupportErrorInfo) {
		*ppv = (LPVOID) this;
		AddRef();

		return NOERROR;
	} else if (riid == IID_IErrorInfo) {
		return GetErrorInfo(0,(IErrorInfo **) ppv);
	} else {
		*ppv = NULL;
		return E_NOINTERFACE;
	}
}

STDMETHODIMP_(ULONG) CIntelligence::AddRef(void)
{
	InterlockedIncrement((LPLONG) &refCount);

	return refCount;
}

STDMETHODIMP_(ULONG) CIntelligence::Release(void)
{
	long count = InterlockedDecrement((LPLONG) &refCount);

	if (count == 0) {
		PostQuitMessage(0);
		delete this;
	}

	return count;
}

STDMETHODIMP CIntelligence::GetTypeInfoCount(UINT* pctinfo)
{	
	*pctinfo = 1; 
	return S_OK;
}

STDMETHODIMP CIntelligence::GetTypeInfo(UINT itinfo, LCID lcid, ITypeInfo** pptinfo)
{
	*pptinfo = typeInfo;
	typeInfo->AddRef();
	return S_OK;
}

STDMETHODIMP CIntelligence::GetIDsOfNames(REFIID riid, LPOLESTR* rgszNames, UINT cNames, LCID lcid, DISPID* rgdispid)
{
	return typeInfo->GetIDsOfNames(rgszNames, cNames, rgdispid);
}

STDMETHODIMP CIntelligence::Invoke(DISPID dispidMember, REFIID riid, 
		LCID lcid, WORD wFlags, DISPPARAMS* pdispparams, VARIANT* pvarResult,
		EXCEPINFO* pexcepinfo, UINT* puArgErr)
{
	return typeInfo->Invoke(this, dispidMember, wFlags, pdispparams, pvarResult, pexcepinfo, puArgErr);
}

STDMETHODIMP CIntelligence::InterfaceSupportsErrorInfo(REFIID riid)
{
	if (riid == IID_IIntelligence) {
		return S_OK;
	} else {
		return S_FALSE;
	}
}

STDMETHODIMP CIntelligence::getActiveId(long* id)
{
	*id = activeId;
	return S_OK;
}
