/**
 * @file
 * @brief Reverse Tunnel Client Service
 *
 * @version $Revision: 629 $
 * @date $Date: 2004-02-22 00:01:47 +0100 (Son, 22 Feb 2004) $
 * @author Harald Radi <h.radi@nme.at>
 */

/* $Id: Service.cpp 629 2004-02-21 23:01:47Z harald $ */

#include "service.h"

#include <tchar.h>
#include <stdio.h>

// static variables
Service* Service::This = NULL;

Service::Service(LPTSTR name, LPTSTR desc)
{
	// Set the default service name and version
	EventSource = NULL;

	ServiceName = name;
	ServiceDesc = desc;

	// set up the initial service status 
	ServiceStatus = NULL;
	Status.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
	Status.dwCurrentState = SERVICE_STOPPED;
	Status.dwControlsAccepted = SERVICE_ACCEPT_STOP;
	Status.dwWin32ExitCode = 0;
	Status.dwServiceSpecificExitCode = 0;
	Status.dwCheckPoint = 0;
	Status.dwWaitHint = 0;
	isRunning = false;
}

Service::~Service()
{
	DebugMsg("Service::~Service()");
	if (EventSource) {
		::DeregisterEventSource(EventSource);
	}
}

////////////////////////////////////////////////////////////////////////////////////////
// Default command line argument parsing

// Returns TRUE if it found an arg it recognised, FALSE if not
// Note: processing some arguments causes output to stdout to be generated.
bool Service::parseStandardArgs(int argc, char* argv[])
{
	// See if we have any command line args we recognise
	if ((argc <= 1) || (argv[1][0] != '-')) return false;

	switch (argv[1][1]) {
		case 'h':
			printHelp();
			return true;

		case 'v':
			printf("\n%s\n\nBuild:\t\t%s\n", ServiceDesc, __TIMESTAMP__);
			printVersion();
			printf("The service is %s installed\n", isInstalled() ? "currently" : "not");
			return true;

		case 'i':
			// Request to install.
			if (isInstalled()) {
				printf("%s is already installed\n", ServiceName);
			} else {
				// Try and install the copy that's running
				if (install()) {
					printf("%s installed\n", ServiceName);
				} else {
					printf("%s failed to install. Error %d\n", ServiceName, GetLastError());
				}
			}
			return true;

		case 'u':
			// Request to uninstall.
			if (!isInstalled()) {
				printf("%s is not installed\n", ServiceName);
			} else {
				// Try and remove the copy that's installed
				if (uninstall()) {
					// Get the executable file path
					char szFilePath[_MAX_PATH];
					::GetModuleFileName(NULL, szFilePath, sizeof(szFilePath));
					printf("%s removed. (You must delete the file (%s) yourself.)\n",
						ServiceName, szFilePath);
				} else {
					printf("Could not remove %s. Error %d\n", ServiceName, GetLastError());
				}
			}
			return true;

		case 'c':
			int l = -1;
			if ((argc > 3) && (argv[2][0] == '-') && (argv[2][1] == 'l')) {
                l = atoi(argv[3]);
			}

			SetConsoleCtrlHandler((PHANDLER_ROUTINE) CtrlHandler, TRUE);
			execConsole(l);
			return true;
	}

	// Don't recognise the args
	return false;
}

void Service::printHelp()
{
	printf("usage:\t\n%s <option>\n\noptions:\n\t-i\t\tinstall service\n\t-u\t\tuninstall service\n\t-c [-l <n>]\tconsole mode [with log level n (0..4)]\n\t-v\t\tversion information\n\t-h\t\tthis help\n", ServiceName);
}

////////////////////////////////////////////////////////////////////////////////////////
// Install/uninstall routines

// Test if the service is currently installed
bool Service::isInstalled()
{
	bool result = false;

	// Open the Service Control Manager
	SC_HANDLE hSCM = ::OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);

	if (hSCM) {
		// Try to open the service
		SC_HANDLE hService = ::OpenService(hSCM, ServiceName, SERVICE_QUERY_CONFIG);
		if (hService) {
			::CloseServiceHandle(hService);
			result = true;
		}

		::CloseServiceHandle(hSCM);
	}

	return result;
}

bool Service::install()
{
	if (!onInstall()) return false;

	// Open the Service Control Manager
	SC_HANDLE hSCM = ::OpenSCManager(NULL, // local machine
		NULL, // ServicesActive database
		SC_MANAGER_ALL_ACCESS); // full access
	if (!hSCM) return false;

	// Get the executable file path
	char szFilePath[_MAX_PATH];
	::GetModuleFileName(NULL, szFilePath, sizeof(szFilePath));

	// Create the service
	SC_HANDLE hService = ::CreateService(hSCM,
		ServiceName,
		ServiceDesc,
		SERVICE_ALL_ACCESS,
		SERVICE_WIN32_OWN_PROCESS,
		SERVICE_AUTO_START,
		SERVICE_ERROR_NORMAL,
		szFilePath,
		NULL,
		NULL,
		NULL,
		NULL,
		NULL);
	if (!hService) {
		::CloseServiceHandle(hSCM);
		return false;
	}

	// make registry entries to support logging messages
	// Add the source name as a subkey under the Application
	// key in the EventLog service portion of the registry.
	char szKey[256];
	HKEY hKey = NULL;
	strcpy(szKey, "SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\");
	strcat(szKey, ServiceName);
	if (::RegCreateKey(HKEY_LOCAL_MACHINE, szKey, &hKey) != ERROR_SUCCESS) {
		::CloseServiceHandle(hService);
		::CloseServiceHandle(hSCM);
		return false;
	}

	// Add the Event ID message-file name to the 'EventMessageFile' subkey.
	::RegSetValueEx(hKey,
		"EventMessageFile",
		0,
		REG_EXPAND_SZ, 
		(CONST BYTE*)szFilePath,
		(DWORD) strlen(szFilePath) + 1);     

	// Set the supported types flags.
	DWORD dwData = EVENTLOG_ERROR_TYPE | EVENTLOG_WARNING_TYPE | EVENTLOG_INFORMATION_TYPE;
	::RegSetValueEx(hKey,
		"TypesSupported",
		0,
		REG_DWORD,
		(CONST BYTE*)&dwData,
		sizeof(DWORD));
	::RegCloseKey(hKey);

	logEvent(EVENTLOG_INFORMATION_TYPE, EVMSG_INSTALLED, ServiceName);

	// tidy up
	::CloseServiceHandle(hService);
	::CloseServiceHandle(hSCM);
	return true;
}

bool Service::uninstall()
{
	if (!onUninstall()) return false;

	// Open the Service Control Manager
	SC_HANDLE hSCM = ::OpenSCManager(NULL, // local machine
		NULL, // ServicesActive database
		SC_MANAGER_ALL_ACCESS); // full access
	if (!hSCM) return false;

	bool result = false;
	SC_HANDLE hService = ::OpenService(hSCM,
		ServiceName,
		DELETE);
	if (hService) {
		if (::DeleteService(hService)) {
			logEvent(EVENTLOG_INFORMATION_TYPE, EVMSG_REMOVED, ServiceName);
			result = true;
		} else {
			logEvent(EVENTLOG_ERROR_TYPE, EVMSG_NOTREMOVED, ServiceName);
		}
		::CloseServiceHandle(hService);
	}

	::CloseServiceHandle(hSCM);
	return result;
}

///////////////////////////////////////////////////////////////////////////////////////
// Logging functions

// This function makes an entry into the application event log
void Service::logEvent(WORD wType, DWORD dwID,
					   const char* pszS1,
					   const char* pszS2,
					   const char* pszS3)
{
	const char* ps[3];
	ps[0] = pszS1;
	ps[1] = pszS2;
	ps[2] = pszS3;

	int iStr = 0;
	for (int i = 0; i < 3; i++) {
		if (ps[i] != NULL) iStr++;
	}

	// Check the event source has been registered and if
	// not then register it now
	if (EventSource == NULL) {
		EventSource = ::RegisterEventSource(NULL,  // local machine
			ServiceName); // source name
	}

	if (EventSource) {
		::ReportEvent(EventSource,
			wType,
			0,
			dwID,
			NULL, // sid
			iStr,
			0,
			ps,
			NULL);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////
// Service startup and registration

bool Service::startService()
{
	BOOL b = FALSE;

	if (isInstalled()) {
		SERVICE_TABLE_ENTRY st[] = {
			{ServiceName, ServiceMain},
			{NULL, NULL}
		};

		DebugMsg("Calling StartServiceCtrlDispatcher()");
		b = ::StartServiceCtrlDispatcher(st);
		DebugMsg("Returned from StartServiceCtrlDispatcher()");
	}

	if (!b) {
		printHelp();
		return false;
	}

	return true;
}

int Service::exitService()
{
	int ret = Status.dwWin32ExitCode;
	delete This;

	return ret;
}

// static member function (callback)
void Service::ServiceMain(DWORD dwArgc, LPTSTR* lpszArgv)
{
	This->DebugMsg("Entering Service::ServiceMain()");
	// Register the control request handler
	This->Status.dwCurrentState = SERVICE_START_PENDING;
	This->ServiceStatus = RegisterServiceCtrlHandler(This->ServiceName, Handler);
	if (This->ServiceStatus == NULL) {
		This->logEvent(EVENTLOG_ERROR_TYPE, EVMSG_CTRLHANDLERNOTINSTALLED);
		return;
	}

	// Start the initialisation
	if (This->initialize()) {
		// Do the real work. 
		// When the Run function returns, the service has stopped.
		This->isRunning = true;
		This->Status.dwWin32ExitCode = 0;
		This->Status.dwCheckPoint = 0;
		This->Status.dwWaitHint = 0;
		This->run();
	}

	// Tell the service manager we are stopped
	This->setStatus(SERVICE_STOPPED);

	This->DebugMsg("Leaving Service::ServiceMain()");
}

///////////////////////////////////////////////////////////////////////////////////////////
// status functions

void Service::setStatus(DWORD dwState)
{
	DebugMsg("Service::SetStatus(%lu, %lu)", ServiceStatus, dwState);
	Status.dwCurrentState = dwState;
	::SetServiceStatus(ServiceStatus, &Status);
}

///////////////////////////////////////////////////////////////////////////////////////////
// Service initialization

bool Service::initialize()
{
	DebugMsg("Entering Service::Initialize()");

	// Start the initialization
	setStatus(SERVICE_START_PENDING);

	// Perform the actual initialization
	bool result = onInit(); 

	// Set final state
	Status.dwWin32ExitCode = GetLastError();
	Status.dwCheckPoint = 0;
	Status.dwWaitHint = 0;
	if (!result) {
		logEvent(EVENTLOG_ERROR_TYPE, EVMSG_FAILEDINIT);
		setStatus(SERVICE_STOPPED);
		return false;    
	}

	logEvent(EVENTLOG_INFORMATION_TYPE, EVMSG_STARTED);
	setStatus(SERVICE_RUNNING);

	DebugMsg("Leaving Service::Initialize()");
	return true;
}

BOOL Service::CtrlHandler(DWORD fdwCtrlType)
{
    switch (fdwCtrlType) {
        // Handle the CTRL+C signal.
        case CTRL_C_EVENT:
            This->onStop();
            return TRUE;

        // CTRL+CLOSE: confirm that the user wants to exit.
        case CTRL_CLOSE_EVENT:
            return TRUE;

        // Pass other signals to the next handler.
        case CTRL_BREAK_EVENT:
        case CTRL_LOGOFF_EVENT:
        case CTRL_SHUTDOWN_EVENT:

        default:
            return FALSE;
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////
// main function to do the real work of the service

// This function performs the main work of the service. 
// When this function returns the service has stopped.
void Service::run()
{
	DebugMsg("Entering Service::Run()");

	while (isRunning) {
		DebugMsg("Sleeping...");
		Sleep(5000);
	}

	// nothing more to do
	DebugMsg("Leaving Service::Run()");
}

//////////////////////////////////////////////////////////////////////////////////////
// Control request handlers

// static member function (callback) to handle commands from the
// service control manager
void Service::Handler(DWORD dwOpcode)
{
	This->DebugMsg("Service::Handler(%lu)", dwOpcode);
	switch (dwOpcode) {
	case SERVICE_CONTROL_STOP: // 1
		This->setStatus(SERVICE_STOP_PENDING);
		This->onStop();
		This->logEvent(EVENTLOG_INFORMATION_TYPE, EVMSG_STOPPED);
		break;

	case SERVICE_CONTROL_PAUSE: // 2
		This->onPause();
		break;

	case SERVICE_CONTROL_CONTINUE: // 3
		This->onContinue();
		break;

	case SERVICE_CONTROL_INTERROGATE: // 4
		This->onInterrogate();
		break;

	case SERVICE_CONTROL_SHUTDOWN: // 5
		This->onShutdown();
		break;

	default:
		if (dwOpcode >= SERVICE_CONTROL_USER) {
			if (!This->onUserControl(dwOpcode)) {
				This->logEvent(EVENTLOG_ERROR_TYPE, EVMSG_BADREQUEST);
			}
		} else {
			This->logEvent(EVENTLOG_ERROR_TYPE, EVMSG_BADREQUEST);
		}
		break;
	}

	// Report current status
	This->DebugMsg("Updating status (%lu, %lu)",
		This->ServiceStatus,
		This->Status.dwCurrentState);
	::SetServiceStatus(This->ServiceStatus, &This->Status);

	if (dwOpcode == SERVICE_CONTROL_STOP) {
		This->isRunning = false;
	}
}

// Called when the service is installed
bool Service::onInstall()
{
	DebugMsg("Service::onInstall()");
	return true;
}

// Called when the service is uninstalled
bool Service::onUninstall()
{
	DebugMsg("Service::onUninstall()");
	return true;
}

// Called when the service is first initialized
bool Service::onInit()
{
	DebugMsg("Service::onInit()");
	return true;
}

// Called when the service control manager wants to stop the service
void Service::onStop()
{
	DebugMsg("Service::onStop()");
}

// called when the service is interrogated
void Service::onInterrogate()
{
	DebugMsg("Service::onInterrogate()");
}

// called when the service is paused
void Service::onPause()
{
	DebugMsg("Service::onPause()");
}

// called when the service is continued
void Service::onContinue()
{
	DebugMsg("Service::onContinue()");
}

// called when the service is shut down
void Service::onShutdown()
{
	DebugMsg("Service::onShutdown()");
}

// called when the service gets a user control message
bool Service::onUserControl(DWORD dwOpcode)
{
	DebugMsg("Service::onUserControl(%8.8lXH)", dwOpcode);
	return false; // say not handled
}

string Service::ToString()
{
	return ServiceName;
}

////////////////////////////////////////////////////////////////////////////////////////////
// Debugging support

void Service::DebugMsg(const char* pszFormat, ...)
{
	char buf[1024];
	sprintf(buf, "[%s](%lu): ", This->ServiceName, GetCurrentThreadId());
	va_list arglist;
	va_start(arglist, pszFormat);
	vsprintf(&buf[strlen(buf)], pszFormat, arglist);
	va_end(arglist);
	strcat(buf, "\n");
	OutputDebugString(buf);
}
