/**
 * @file
 * @brief Abstract thread function declaration
 *
 * This is a simple, really cross-platform thread class with support for
 * - Linux
 * - Win32
 * - WinCE
 *
 * It doesn't do much, just offer methods to start a native thread (no
 * cancellation !) and some synchronization. Why has it been written when
 * there are so many other thread toolkits out there with a lot of 
 * functionality (e.g. the very nice ZThreads package) ? Because none of them
 * was able to get a thread running in Linux, Win32 _and_ WinCE without major
 * porting efforts (major >= 2 days). In short, this class grew out of 
 * frustration !
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-02-23 12:24:40 +0100 (Mon, 23 Feb 2004) $
 * $Revision: 635 $
 *
 * @since    Tue May 13 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: Thread.h 635 2004-02-23 11:24:40Z rene $
 */

#ifndef _THREAD_H
#define _THREAD_H

#include "defines.h"

#if defined(_WINDOWS) || defined(_WIN32_WCE)
# define THREAD DWORD WINAPI
#elif defined(_SYMBIAN)
# define THREAD TInt
#else
/** Attributes for thread main callback function */
# define THREAD void*
#endif

/**
 * Platform independent thread class
 */
class Thread
{
	protected:
		/**
		 * %Thread handle
		 */
		class ThreadHandle
		{
			friend class Thread;

			private:
				/** handle */
				void* nativeHandle;

			public:
				ThreadHandle();
				~ThreadHandle();
		};

		/**
		 * Lock
		 */
		class Lock
		{
			private:
				/** handle */
				void* nativeLock;

			public:
				/** Constructor */
				Lock();
				/** Destructor */
				~Lock();

				/** Lock */
				void lock();
				/** Unlock */
				void unlock();
		};

		/**
		 * Guard
		 */
		class Guard {
			private:
				/** lock instance */
				Lock& lock;
				
			public:
				/**
				 * Constructor
				 *
				 * @param lock Lock instance
				 */
				Guard(Lock& lock);
				/** Destructor */
				~Guard();
		};

	private:
		//@{
		/** Thread properties */
		bool cancelled;
		bool started;
		ThreadHandle handle;
		Lock lockShouldTerminate;
		//@}

		/**
		 * Worker function
		 *
		 * @param userData Thread object
		 * @return Status code
		 */
		static THREAD runProc(void *userData);

	public:
		/** Constructor */
		Thread();
		/** Destructor */
		virtual ~Thread();

		/**
		 * sleep function
		 *
		 * @param milliseconds Timeout in milliseconds
		 */
		static void sleep(unsigned milliseconds);

		/** Start thread execution */
		void start();
		/** End thread execution */
		void cancel();
		/** Join thread */
		void join();
		/** Terminate thread execution */
		void terminate();


	protected:
		/**
		 * Thread main worker function
		 */
		virtual void run() = 0;

		/**
		 * @return <var>true</var> if the thread is cancelled, <var>false</var>
		 *         if the thread is still running
		 */
		bool isCancelled();
};


#endif
