/**
 * @file
 * @brief Debug log implementation
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-02-12 21:08:39 +0100 (Don, 12 Feb 2004) $
 * $Revision: 612 $
 *
 * @since    Sat Aug 30 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: debug.cpp 612 2004-02-12 20:08:39Z harald $
 */

#include "defines.h"
#include "assert.h"

#include <stdio.h>
#include <stdarg.h>

#if (!defined(_WIN32_WCE) && !defined(SYMBIAN))
# include <iostream>
#elif defined(_WIN32_WCE)
# include <arConsolePPC.h>
# include <arOfStream.h>
#endif

#ifdef _LINUX
# include <dlfcn.h>
#endif

/** Function type definitions for logging */
typedef void (*log_t) (const char *format, va_list args);
/** Function type definitions for initializing logging */
typedef void (*init_log_t) (log_t log);

static bool con;              /**< Flag, signaling whether output should be printed to the console or not */
static FILE* logfile = NULL;  /**< Logfile file handle */

#ifdef _WIN32_WCE
static arConsolePPC *pConsole;
#endif

#ifdef _SYMBIAN
_LIT(KTxtConsole, "Proca Console");
_LIT(KTxtPressAnyKey, "[press any key]");
_LIT(KAnyString, "%s");
LOCAL_D CConsoleBase* pConsole;
#endif

//@{
/** Implementations */
static void __logfoo(const char *format, va_list args);
static void __log(const char *format, va_list args);
//@}

/** Globally used logging function set by the init_log function */
log_t log_cb = __logfoo;

void _open_log(const char* path, bool console)
{
	log_cb = __log;
	con = console;
	if (path != NULL) {
		logfile = fopen(path, "w");
		assert(logfile != NULL);
	}

#ifdef _WIN32_WCE
	if (con) {
		pConsole = (arConsolePPC*)(arConsolePPC::create("Test Intelligence"));
		arConsole::setFocus();
	}
#endif

#ifdef _SYMBIAN
	if (con) {
		pConsole = Console::NewL(KTxtConsole, TSize(KConsFullScreen, KConsFullScreen));
		CleanupStack::PushL(pConsole);
	}
#endif
}

void _close_log()
{
	if (logfile != NULL)
		fclose(logfile);

#ifdef _WIN32_WCE
	if (con) {
		arConsole::destroy();
	}
#endif

#ifdef _SYMBIAN
	if (con) {
		pConsole->Printf(KTxtPressAnyKey);
		pConsole->Getch(); // just ignore the character, but wait
		CleanupStack::PopAndDestroy(); // close console
	}
#endif
}

/* initialize logging for dynamically loaded libs */
#ifdef WIN32
# define DEBUG_EXPORT __declspec(dllexport)

void _dll_init_log(void *dll)
{
	init_log_t init;
	if ((init = (init_log_t) GetProcAddress((HMODULE) dll, TEXT("init_log"))) != NULL) {
		init(__log);
	}
}
#else
# define DEBUG_EXPORT

void _dll_init_log(void *dlHandle)
{
	init_log_t init;
	if (((init = (init_log_t) dlsym(dlHandle, "init_log")) != NULL) && (dlerror() == 0)) {
		init(__log);
	}
}
#endif

/**
 * Exported funtion to initialize logging
 *
 * @param log Logging function
 */
extern "C" DEBUG_EXPORT void init_log(log_t log)
{
	log_cb = log;
}

void _log(const char *format, ...)
{
	va_list args;
	va_start(args, format);
	log_cb(format, args);
	va_end(args);
}

static void __log(const char *format, va_list args)
{
	char buf[4096];
#ifndef _SYMBIAN
	vsnprintf(buf, sizeof(buf), format, args);
#else
	// DANGEROUS !
	vsprintf(buf, format, args);
#endif

	if (con) {
#ifndef _SYMBIAN
		cout << buf << flush;
#else
		TPtrC8 b((const TText8*) buf, strlen(buf));
#if !defined(_UNICODE)
		pConsole->Printf(b);
#else 
		// this hack has been taken from http://www3.symbian.com/faq.nsf/51681023dd5b127980256a64005ac61c/3a5431bc2ac06ed580256a570051ba40?OpenDocument&Highlight=2,console
		TInt remaining = b.Length();
		const TText8* cp = b.Ptr(); // get pointer
		while (remaining) {
			TBuf16<256> tbc;
			TInt len = Min(remaining, 256);
			tbc.Copy(TPtrC8(cp, len)); // convert in 256 byte chunks
			pConsole->Write(tbc);
			cp += len;
			remaining -= len;
		}
#endif // unicode
#endif // symbian
	}

	if (logfile != NULL) {
		fprintf(logfile, buf);
		fflush(logfile);
	}
}

#ifdef _SYMBIAN
int getConsoleChar() {
	if (con)
		return pConsole->Getch();
	else
		return -1;
}
#endif

void _logfoo(const char *, ...) {}
static void __logfoo(const char *, va_list) {}
