/**
 * @file
 * @brief Global defines
 *
 * This file contains often used defines and includes and
 * contains macros to have a platform and compiler independend
 * way of accessing basic functions.
 * @note STL includes should only be done in this file to better
 *       cope with the different STL implementations.
 *
 * &copy; 2003-2004 by Rene Mayrhofer, Harald Radi
 *
 * $Date: 2004-03-02 13:41:03 +0100 (Die, 02 Mär 2004) $
 * $Revision: 674 $
 *
 * @since    Sun Jul 13 2003
 * @author   Rene Mayrhofer <rene@mayrhofer.eu.org>
 * @author   Harald Radi    <harald.radi@nme.at>
 *
 * <hr><strong>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * </strong><hr>
 *
 * $Id: defines.h 674 2004-03-02 12:41:03Z rene-cvs $
 */

#ifndef _DEFINES_H
#define _DEFINES_H

/// Prevent the MIN() and MAX() macros from being defined
#define NOMINMAX

// first of all, try to figure out for which platform the code is compiled
#if (defined(_WIN32) && !defined(_WINDOWS))
# define _WINDOWS
#endif

//#if (!defined(_WINDOWS) && !defined(WIN32) && !defined(_WIN32_WCE) && !defined(_SYMBIAN)) || defined(__linux__)
#ifdef __linux__
# define _LINUX
#endif

/// Hint the debug heap to dump all memory blocks not freed so far
#define DUMP_LEAKS()

/**
 * Get a random value
 *
 * This macro returns a random value within the interval [0; 1[
 *
 * @return Random value
 */
# define rand_double() ((double) rand() / (RAND_MAX + 0.00001))

#ifdef _LINUX
# define INIT_EXPORT __attribute__ ((constructor))
# define FINI_EXPORT __attribute__ ((destructor))
#else
/** Compiler attributes for library initialization */
# define INIT_EXPORT
/** Compiler attributes for library cleanup */
# define FINI_EXPORT
#endif

#ifdef _WINDOWS
# ifdef _DEBUG
#  undef DUMP_LEAKS
#  define DUMP_LEAKS() _CrtDumpMemoryLeaks()
#  define _CRTDBG_MAP_ALLOC
#  include <malloc.h>
#  include <stdlib.h>
#  ifndef _WIN32_WCE
#   include <crtdbg.h>
#  endif
# endif // _DEBUG

# define bit_vector vector<bool>
# define tzset _tzset
# define vsnprintf _vsnprintf

# define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers
# include <windows.h>
# include <winbase.h>
#endif // _WINDOWS

#ifdef _WIN32_WCE
# include <windows.h>
# include <winbase.h>
# define strdup _strdup
# define ltoa _ltoa
# define itoa _itoa
# define ftoa _ftoa

// and with the stupid VisualC++ Embedded 3.0 compiler, we can't even use exceptions
# define try
# define throw wce_assert<TCHAR>(false, TEXT(__FILE__), __LINE__, L"Exception thrown");
# define catch(x) if (false)
# define finally
#endif // _WIN32_WCE

#ifdef _SYMBIAN
# include <e32base.h>
# include <e32cons.h>
# include <e32std.h>
# include <e32math.h>

# define try
# define throw User::Exit(1);
# define catch(x) if (false)
# define finally

//# define assert(x) __ASSERT_DEBUG(x, Panic(_L"Assertion", 0))
#endif

/**
 * @name Debugging Defines
 *
 * Macros for enabling and disabling debug output
 */
//@{
#define _DEBUG_CLASSIFIERS 1
#define _DEBUG_FEATURES 1
#define _DEBUG_PREDICTORS 1
//@}

/**
 * @name Optimisation Defines
 *
 * Macros for enabling and disabling various optimizations
 */
//@{
#define NO_OPTIMIZATIONS 0                ///< turn off all optimizations (for debugging)

#define OPTIMIZATIONS_SPLAYTREE 0         ///< turn on splaytree optimization
#define OPTIMIZATIONS_CACHE_SIMILARITY 1  ///< turn on cache similarity optimization
#define OPTIMIZATIONS_SPLIT_MERGE 0       ///< turn on split/merge optimization
//@}

#include "debug.h"

// stl includes, once for all to have better control over the different stl implementations
#include <algorithm>
#include <list>
#include <map>
#include <string>
#include <vector>

// no namespace support in the Symbian compiler (old gcc)
#ifndef _SYMBIAN
# include <math.h>
# include <limits>

using namespace std;
#endif

/**
 * @name Global Typedefs
 */
//@{
/** A simple ordered list of strings */
typedef vector<string> stringvector;

/** A map to represent configuration file entries */
typedef map<string, string> parametermap;

//@}

#endif // _DEFINES_H
